(function webpackUniversalModuleDefinition(root, factory) {
	if(typeof exports === 'object' && typeof module === 'object')
		module.exports = factory();
	else if(typeof define === 'function' && define.amd)
		define([], factory);
	else if(typeof exports === 'object')
		exports["whatsharp-core"] = factory();
	else
		root["whatsharp-core"] = factory();
})(global, function() {
return /******/ (() => { // webpackBootstrap
/******/ 	"use strict";
/******/ 	var __webpack_modules__ = ({

/***/ "./src/compiler/FunctionCompiler.ts":
/*!******************************************!*\
  !*** ./src/compiler/FunctionCompiler.ts ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.FunctionCompiler = void 0;
const source_tree_1 = __webpack_require__(/*! ./source-tree */ "./src/compiler/source-tree.ts");
const wa_nodes_1 = __webpack_require__(/*! ../wa-ast/wa-nodes */ "./src/wa-ast/wa-nodes.ts");
const WatSharpCompiler_1 = __webpack_require__(/*! ./WatSharpCompiler */ "./src/compiler/WatSharpCompiler.ts");
const FunctionBuilder_1 = __webpack_require__(/*! ../wa-ast/FunctionBuilder */ "./src/wa-ast/FunctionBuilder.ts");
const expression_resolver_1 = __webpack_require__(/*! ./expression-resolver */ "./src/compiler/expression-resolver.ts");
const wat_optimizer_1 = __webpack_require__(/*! ./wat-optimizer */ "./src/compiler/wat-optimizer.ts");
const wat_helpers_1 = __webpack_require__(/*! ./wat-helpers */ "./src/compiler/wat-helpers.ts");
/**
 * This class is responsible for compiling a function body
 */
class FunctionCompiler {
    /**
     * Initializes a function compiler instance
     * @param wsCompiler WAT# compiler instance
     * @param func Function to compile
     */
    constructor(wsCompiler, func) {
        this.wsCompiler = wsCompiler;
        this.func = func;
        // --- Local parameters and variabled of the function
        this._locals = new Map();
        // --- The result value of the function
        this._resultType = null;
        // --- Temporary locals assigned to the function
        this._tempLocals = new Set();
        // --- The current loop depth
        this._loopDepth = 0;
        // --- Do not emit
        this._suspendCount = 0;
    }
    /**
     * Gets the locals of this function
     */
    get locals() {
        return this._locals;
    }
    /**
     * Gets the builder of this function
     */
    get builder() {
        return this._builder;
    }
    /**
     * Gets the body of the compiled function
     */
    getInstructionBody() {
        return this._builder.body;
    }
    /**
     * Adds a trace message
     * @param traceFactory Factory function to generate trace message
     */
    addTrace(traceFactory) {
        this.wsCompiler.addTrace(traceFactory);
    }
    /**
     * Processes the body of the function
     */
    process() {
        // --- Prepare the function signature
        this.processSignature();
        // --- Compile the function body
        this.func.body.forEach((stmt) => this.compileStatement(stmt, this._builder.body));
        // --- Optimize the emitted WA code of the function body
        wat_optimizer_1.optimizeWat(this._builder.body);
        // --- Remove unused locals
        wat_optimizer_1.optimizeLocalUsage(this._builder);
        // --- Create trace information for test purposes
        this._builder.body.forEach((ins) => {
            if (ins.type !== "Comment") {
                const parts = this.wsCompiler.waTree
                    .renderInstructionNode(ins)
                    .split("\n")
                    .map((item) => item.trim());
                parts.forEach((p) => {
                    this.addTrace(() => ["inject", this.func.funcId, p]);
                });
            }
        });
    }
    /**
     * Processes the function parameters and result type
     */
    processSignature() {
        // --- Map the result type
        this._resultType = WatSharpCompiler_1.mapFunctionParameterType(this.func.resultType);
        // --- Map parameters to locals
        const waPars = [];
        this.func.params.forEach((param) => {
            if (this._locals.has(param.name)) {
                this.reportError("W140", this.func);
            }
            else {
                // --- Fix the type specification
                this.wsCompiler.resolveDependencies(param.spec);
                const resolvedTypeSpec = this.resolveNamedTypes(param.spec);
                if (resolvedTypeSpec) {
                    param.spec = resolvedTypeSpec;
                }
                const paramType = WatSharpCompiler_1.mapFunctionParameterType(param.spec);
                const paramName = WatSharpCompiler_1.createParameterName(param.name);
                this.locals.set(param.name, {
                    fromParameter: true,
                    name: paramName,
                    type: param.spec,
                    waType: paramType,
                });
                waPars.push({
                    id: paramName,
                    type: paramType,
                });
            }
        });
        // --- Create the function builder
        this._builder = this.wsCompiler.waTree.func(WatSharpCompiler_1.createGlobalName(this.func.name), waPars, this._resultType);
    }
    /**
     * Processes the specified function statement
     * @param stmt Statement to process
     * @param body Body to put the statements in
     */
    compileStatement(stmt, body) {
        switch (stmt.type) {
            case "LocalVariable":
                this.compileLocalDeclaration(stmt, body);
                return;
            case "Assignment":
                this.compileAssignment(stmt, body);
                return;
            case "Break":
                this.compileBreak(body);
                return;
            case "Continue":
                this.compileContinue(body);
                return;
            case "Do":
                this.compileDoWhileLoop(stmt, body);
                return;
            case "If":
                this.compileIf(stmt, body);
                return;
            case "LocalFunctionInvocation":
                this.compileLocalFunctionInvocation(stmt, body);
                return;
            case "Return":
                this.compileReturn(stmt, body);
                return;
            case "While":
                this.compileWhileLoop(stmt, body);
                return;
        }
    }
    /**
     * Processes a local variable declaration
     * @param body Body to put the statements in
     */
    compileLocalDeclaration(localVar, body) {
        if (this._locals.has(localVar.name)) {
            this.reportError("W140", this.func);
            return;
        }
        const compiler = this.wsCompiler;
        // --- Fix the type specification
        compiler.resolveDependencies(localVar.spec);
        const resolvedTypeSpec = this.resolveNamedTypes(localVar.spec);
        if (resolvedTypeSpec) {
            localVar.spec = resolvedTypeSpec;
        }
        // --- Start compilation
        const localName = WatSharpCompiler_1.createLocalName(localVar.name);
        let initExpr = null;
        if (localVar.initExpr) {
            initExpr = this.processExpression(localVar.initExpr, body);
            if (initExpr) {
                this.castForStorage(localVar.spec, initExpr.exprType, body, initExpr === null || initExpr === void 0 ? void 0 : initExpr.expr.value);
                this.inject(FunctionBuilder_1.localSet(localName), body);
            }
        }
        const paramType = localVar.spec.type === "Pointer"
            ? wa_nodes_1.WaType.i32
            : WatSharpCompiler_1.waTypeMappings[localVar.spec.underlying];
        this.locals.set(localVar.name, {
            fromParameter: false,
            name: localName,
            type: localVar.spec,
            waType: paramType,
        });
        const local = this._builder.addLocal(localName, paramType);
        this.addTrace(() => [
            "local",
            0,
            this.wsCompiler.waTree.renderLocal(local),
        ]);
    }
    /**
     * Resolves named types within the local type declaration
     * @param spec Type specification to resolve
     * @returns Resolved name type specification
     */
    resolveNamedTypes(spec) {
        switch (spec.type) {
            case "Pointer":
            case "Array":
                const resolved = this.resolveNamedTypes(spec.spec);
                if (resolved) {
                    spec.spec = resolved;
                }
                break;
            case "Struct":
                for (let i = 0; i < spec.fields.length; i++) {
                    const resolved = this.resolveNamedTypes(spec.fields[i].spec);
                    if (resolved) {
                        spec.fields[i].spec = resolved;
                    }
                }
                break;
            case "NamedType":
                const decl = this.wsCompiler.declarations.get(spec.name);
                if (decl) {
                    if (decl.type === "TypeDeclaration") {
                        return decl.spec;
                    }
                }
                break;
        }
        return null;
    }
    /**
     * Processes the specified assignment
     * @param body Body to put the statements in
     */
    compileAssignment(asgn, body) {
        // --- Check the left value type
        let leftSide = "var";
        let leftType;
        let idAddress = false;
        let varName;
        let isParam = false;
        if (asgn.lval.type == "Identifier") {
            // --- A single identifier may be a global, a local, or a memory variable
            varName = asgn.lval.name;
            const resolvedId = this.resolveIdentifier(asgn.lval);
            if (!resolvedId) {
                return;
            }
            idAddress = true;
            if (resolvedId.global) {
                // --- Left side is a global
                leftSide = "global";
                leftType = {
                    type: "Intrinsic",
                    underlying: resolvedId.global.underlyingType,
                };
            }
            else if (resolvedId.local) {
                // --- Left side is a local
                leftSide = "local";
                leftType = resolvedId.local.type;
                isParam = resolvedId.local.fromParameter;
            }
            else if (resolvedId.data) {
                this.reportError("W169", asgn);
                return;
            }
            else {
                // --- Left side is a variable with a const address
                leftType = resolvedId.var.spec;
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, resolvedId.var.address), body);
                if (asgn.asgn !== "=" && asgn.asgn !== ":=") {
                    // --- We need to put this address to the stack twice
                    this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, resolvedId.var.address), body);
                }
            }
        }
        else {
            // --- Left side is a compound left value expression
            const resolvedAddr = this.calculateAddressOf(asgn.lval, body);
            if (!resolvedAddr) {
                return;
            }
            leftType = resolvedAddr.spec;
        }
        // --- Handle copy assignment separately
        if (asgn.asgn === ":=") {
            this.compileCopyAssignmentTail(asgn, leftType, body);
            return;
        }
        // --- Now, we accept only intrinsic types and pointers as left values
        if (leftType.type !== "Intrinsic" && leftType.type !== "Pointer") {
            this.reportError("W161", asgn);
            return;
        }
        // --- If it is an addressed variable, calculate the address
        let tmpAddrVar = "";
        if (leftSide === "var") {
            // --- Do we need double access to this address?
            if (!idAddress && asgn.asgn !== "=") {
                tmpAddrVar = this.createTempLocal(wa_nodes_1.WaType.i32, "asgn");
                this.inject(FunctionBuilder_1.localTee(tmpAddrVar), body);
                this.inject(FunctionBuilder_1.localGet(tmpAddrVar), body);
                // --- At this point we have the variable access on the top of the stack
            }
        }
        // --- We need to get the value
        const leftIntrinsic = leftType.type === "Intrinsic" ? leftType : i32Desc;
        if (asgn.asgn !== "=") {
            switch (leftSide) {
                case "global":
                    this.inject(FunctionBuilder_1.globalGet(WatSharpCompiler_1.createGlobalName(varName)), body);
                    break;
                case "local":
                    this.inject(FunctionBuilder_1.localGet(isParam ? WatSharpCompiler_1.createParameterName(varName) : WatSharpCompiler_1.createLocalName(varName)), body);
                    break;
                case "var":
                    this.compileIntrinsicVariableGet(leftIntrinsic, body);
                    break;
            }
        }
        // --- Now, evaluate the right-side expression
        const rightType = this.compileExpression(asgn.expr, body);
        // --- Valid type?
        if (!rightType) {
            return;
        }
        // --- Match intrinsic left type with intrinsic right type
        if (leftType.type === "Intrinsic" && rightType.type !== "Intrinsic") {
            this.reportError("W141", asgn);
            return;
        }
        // --- Match pointer left type with pointer or integral right type
        if (leftType.type === "Pointer" &&
            rightType.type !== "Pointer" &&
            (rightType.type !== "Intrinsic" || rightType.underlying.startsWith("f"))) {
            this.reportError("W141", asgn);
            return;
        }
        // --- Instrinsic types for the right side
        const rightIntrinsic = rightType.type === "Intrinsic" ? rightType : i32Desc;
        // --- Is it a compound assignment?
        if (asgn.asgn !== "=") {
            // --- Check if the operation should be signed
            const isSigned = leftIntrinsic.underlying.startsWith("i") ||
                rightIntrinsic.underlying.startsWith("i");
            // --- Calculate operation type
            let resultType = i32Desc;
            if (leftIntrinsic.underlying === "f32" &&
                rightIntrinsic.underlying === "f32") {
                resultType = f32Desc;
            }
            else if (leftIntrinsic.underlying.startsWith("f") ||
                rightIntrinsic.underlying.startsWith("f")) {
                resultType = f64Desc;
            }
            else if (leftIntrinsic.underlying.endsWith("64") ||
                rightIntrinsic.underlying.endsWith("64")) {
                resultType = i64Desc;
            }
            // --- Compile the operands and cast them to the appropriate type
            this.castIntrinsicToIntrinsic(resultType, rightIntrinsic, body);
            const waType = WatSharpCompiler_1.waTypeMappings[resultType.underlying];
            // --- Compile the assignment operation
            switch (asgn.asgn) {
                case "+=":
                case "-=":
                    if (leftType.type === "Pointer" && rightType.type === "Intrinsic") {
                        // --- Pointer arithmetic, use the size of the target type
                        this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, this.wsCompiler.getSizeof(leftType.spec)), body);
                        // --- Calculate the result
                        this.inject(FunctionBuilder_1.mul(wa_nodes_1.WaType.i32), body);
                    }
                    if (asgn.asgn === "+=") {
                        this.inject(FunctionBuilder_1.add(waType), body);
                    }
                    else {
                        this.inject(FunctionBuilder_1.sub(waType), body);
                    }
                    break;
                case "*=":
                    this.inject(FunctionBuilder_1.mul(waType), body);
                    break;
                case "/=":
                    this.inject(FunctionBuilder_1.div(waType, isSigned), body);
                    break;
                case "%=":
                    if (resultType.underlying.startsWith("f")) {
                        this.reportError("W145", asgn, "remainder (%)");
                        return;
                    }
                    this.inject(FunctionBuilder_1.rem(waType, isSigned), body);
                    break;
                case "&=":
                    if (resultType.underlying.startsWith("f")) {
                        this.reportError("W145", asgn, "bitwise AND");
                        return;
                    }
                    this.inject(FunctionBuilder_1.and(waType), body);
                    break;
                case "|=":
                    if (resultType.underlying.startsWith("f")) {
                        this.reportError("W145", asgn, "bitwise OR");
                        return;
                    }
                    this.inject(FunctionBuilder_1.or(waType), body);
                    break;
                case "^=":
                    if (resultType.underlying.startsWith("f")) {
                        this.reportError("W145", asgn, "bitwise XOR");
                        return;
                    }
                    this.inject(FunctionBuilder_1.xor(waType), body);
                    break;
                case "<<=":
                    if (resultType.underlying.startsWith("f")) {
                        this.reportError("W145", asgn, "shift left");
                        return;
                    }
                    this.inject(FunctionBuilder_1.shl(waType), body);
                    break;
                case ">>=":
                    if (resultType.underlying.startsWith("f")) {
                        this.reportError("W145", asgn, "signed shift right");
                        return;
                    }
                    this.inject(FunctionBuilder_1.shr(waType, true), body);
                    break;
                case ">>>=":
                    if (resultType.underlying.startsWith("f")) {
                        this.reportError("W145", asgn, "shift right");
                        return;
                    }
                    this.inject(FunctionBuilder_1.shr(waType, false), body);
                    break;
            }
        }
        // --- Cast the result
        this.castIntrinsicToIntrinsic(leftIntrinsic, rightIntrinsic, body);
        // --- Store the resuls
        switch (leftSide) {
            case "global":
                this.inject(FunctionBuilder_1.globalSet(WatSharpCompiler_1.createGlobalName(varName)), body);
                break;
            case "local":
                this.inject(FunctionBuilder_1.localSet(isParam ? WatSharpCompiler_1.createParameterName(varName) : WatSharpCompiler_1.createLocalName(varName)), body);
                break;
            case "var":
                this.compileIntrinsicVariableSet(leftIntrinsic, body);
                break;
        }
    }
    /**
     * Compiles the tail of a copy assignment
     * @param asgn Assignment to compile
     * @param leftType Left value type
     * @param body Body to put the statements in
     */
    compileCopyAssignmentTail(asgn, leftType, body) {
        // --- The left value address is already on the stack
        // --- Test the left value type
        if (leftType.type !== "Array" && leftType.type !== "Struct") {
            this.reportError("W166", asgn);
            return;
        }
        // --- Check if the left side has a constand address
        let leftAddrCons;
        let leftTmpVar;
        if (body.length > 0 && body[body.length - 1].type === "ConstVal") {
            // --- Constant address
            leftAddrCons = body[body.length - 1];
            body.splice(body.length - 1, 1);
        }
        else {
            // --- Compound address
            leftTmpVar = this.createTempLocal(wa_nodes_1.WaType.i32, "lcpy");
            this.inject(FunctionBuilder_1.localSet(leftTmpVar), body);
        }
        // --- Get the right side value
        const rightType = this.compileExpression(asgn.expr, body);
        if (!rightType) {
            return;
        }
        // --- Check right side type
        if (rightType.type !== "Intrinsic" ||
            rightType.underlying.startsWith("f")) {
            this.reportError("W167", asgn);
            return;
        }
        // --- Check if the left side has a constand address
        let rightAddrCons;
        let rightTmpVar;
        if (body.length > 0 && body[body.length - 1].type === "ConstVal") {
            // --- Constant address
            rightAddrCons = body[body.length - 1];
            body.splice(body.length - 1, 1);
        }
        else {
            // --- Compound address
            rightTmpVar = this.createTempLocal(wa_nodes_1.WaType.i32, "rcpy");
            this.inject(FunctionBuilder_1.localSet(rightTmpVar), body);
        }
        // --- Compile the entiry structure, provided its length is less than 256 bytes
        let copyLength = this.wsCompiler.getSizeof(leftType);
        if (copyLength < 0 || copyLength > 256) {
            this.reportError("W168", asgn);
            return;
        }
        // --- Generate copy instructions
        let offset = 0;
        while (copyLength >= 8) {
            if (leftAddrCons) {
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, leftAddrCons.value), body);
            }
            else {
                this.inject(FunctionBuilder_1.localGet(leftTmpVar), body);
            }
            if (rightAddrCons) {
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, rightAddrCons.value), body);
            }
            else {
                this.inject(FunctionBuilder_1.localGet(rightTmpVar), body);
            }
            this.inject(FunctionBuilder_1.load(wa_nodes_1.WaType.i64, undefined, offset), body);
            this.inject(FunctionBuilder_1.store(wa_nodes_1.WaType.i64, undefined, offset), body);
            offset += 8;
            copyLength -= 8;
        }
        while (copyLength >= 4) {
            if (leftAddrCons) {
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, leftAddrCons.value), body);
            }
            else {
                this.inject(FunctionBuilder_1.localGet(leftTmpVar), body);
            }
            if (rightAddrCons) {
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, rightAddrCons.value), body);
            }
            else {
                this.inject(FunctionBuilder_1.localGet(rightTmpVar), body);
            }
            this.inject(FunctionBuilder_1.load(wa_nodes_1.WaType.i32, undefined, offset), body);
            this.inject(FunctionBuilder_1.store(wa_nodes_1.WaType.i32, undefined, offset), body);
            offset += 4;
            copyLength -= 4;
        }
        while (copyLength >= 2) {
            if (leftAddrCons) {
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, leftAddrCons.value), body);
            }
            else {
                this.inject(FunctionBuilder_1.localGet(leftTmpVar), body);
            }
            if (rightAddrCons) {
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, rightAddrCons.value), body);
            }
            else {
                this.inject(FunctionBuilder_1.localGet(rightTmpVar), body);
            }
            this.inject(FunctionBuilder_1.load(wa_nodes_1.WaType.i32, wa_nodes_1.WaBitSpec.Bit16, offset), body);
            this.inject(FunctionBuilder_1.store(wa_nodes_1.WaType.i32, wa_nodes_1.WaBitSpec.Bit16, offset), body);
            offset += 2;
            copyLength -= 2;
        }
        while (copyLength > 0) {
            if (leftAddrCons) {
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, leftAddrCons.value), body);
            }
            else {
                this.inject(FunctionBuilder_1.localGet(leftTmpVar), body);
            }
            if (rightAddrCons) {
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, rightAddrCons.value), body);
            }
            else {
                this.inject(FunctionBuilder_1.localGet(rightTmpVar), body);
            }
            this.inject(FunctionBuilder_1.load(wa_nodes_1.WaType.i32, wa_nodes_1.WaBitSpec.Bit8, offset), body);
            this.inject(FunctionBuilder_1.store(wa_nodes_1.WaType.i32, wa_nodes_1.WaBitSpec.Bit8, offset), body);
            offset += 1;
            copyLength -= 1;
        }
    }
    /**
     * Processes a break statement
     * @param body Body to put the statements in
     */
    compileBreak(body) {
        this.inject(FunctionBuilder_1.branch(WatSharpCompiler_1.createBreakLabel(this._loopDepth)), body);
    }
    /**
     * Processes a continue statement
     * @param body Body to put the statements in
     */
    compileContinue(body) {
        this.inject(FunctionBuilder_1.branch(WatSharpCompiler_1.createLoopLabel(this._loopDepth)), body);
    }
    /**
     * Processes a do..while loop
     * @param body Body to put the statements in
     */
    compileDoWhileLoop(dLoop, body) {
        const hasBreak = dLoop.loopBody.some((st) => this.hasBreak(st));
        // --- New loop depth
        this._loopDepth++;
        // --- This contains the body of the loop
        const loopBody = [];
        // --- Compilet the main loop
        const loopLabel = WatSharpCompiler_1.createLoopLabel(this._loopDepth);
        dLoop.loopBody.forEach((stmt) => this.compileStatement(stmt, loopBody));
        // --- Compile the test
        if (!this.compileExpression(dLoop.test, loopBody)) {
            return;
        }
        // --- Combine the test and the body into a loop
        const loopToInject = FunctionBuilder_1.loop(loopLabel, [...loopBody, FunctionBuilder_1.branchIf(loopLabel)]);
        if (hasBreak) {
            // --- Encapsulate the loop into a break block
            const breakBlock = FunctionBuilder_1.block(WatSharpCompiler_1.createBreakLabel(this._loopDepth), [
                loopToInject,
            ]);
            this.inject(breakBlock, body);
        }
        else {
            // --- Inject the loop
            this.inject(loopToInject, body);
        }
        // --- Back to previous loop back
        this._loopDepth--;
    }
    /**
     * Processes a while loop
     * @param body Body to put the statements in
     */
    compileWhileLoop(wLoop, body) {
        const hasBreak = wLoop.loopBody.some((st) => this.hasBreak(st));
        // --- New loop depth
        this._loopDepth++;
        // --- This contains the body of the loop
        const loopBody = [];
        // --- Compile the test
        if (!this.compileExpression(wLoop.test, loopBody)) {
            return;
        }
        // --- Compile the loop body
        const loopMain = [];
        const loopLabel = WatSharpCompiler_1.createLoopLabel(this._loopDepth);
        wLoop.loopBody.forEach((stmt) => this.compileStatement(stmt, loopMain));
        this.inject(FunctionBuilder_1.branch(loopLabel), loopMain);
        // --- Combine the test and the body into a loop
        const loopToInject = FunctionBuilder_1.loop(loopLabel, [...loopBody, FunctionBuilder_1.ifBlock(loopMain)]);
        if (hasBreak) {
            // --- Encapsulate the loop into a break block
            const breakBlock = FunctionBuilder_1.block(WatSharpCompiler_1.createBreakLabel(this._loopDepth), [
                loopToInject,
            ]);
            this.inject(breakBlock, body);
        }
        else {
            // --- Inject the loop
            this.inject(loopToInject, body);
        }
        // --- Back to previous loop back
        this._loopDepth--;
    }
    /**
     * Tests if the specified statement is a break or contains a break
     * @param stmt Statement to check
     * @returns True, if a break found; otherwise, false
     */
    hasBreak(stmt) {
        switch (stmt.type) {
            case "Break":
                return true;
            case "If":
                const hasIt = stmt.consequent.some((st) => this.hasBreak(st));
                if (hasIt) {
                    return true;
                }
                if (stmt.alternate) {
                    return stmt.alternate.some((st) => this.hasBreak(st));
                }
                break;
        }
        return false;
    }
    /**
     * Processes an if statement
     * @param body Body to put the statements in
     */
    compileIf(ifStmt, body) {
        // --- Test expression
        if (!this.compileExpression(ifStmt.test, body)) {
            return;
        }
        // --- Consequent branch
        const consequentBody = [];
        ifStmt.consequent.forEach((stmt) => this.compileStatement(stmt, consequentBody));
        // --- Alternate branch
        let alternateBody;
        if (ifStmt.alternate) {
            alternateBody = [];
            ifStmt.alternate.forEach((stmt) => this.compileStatement(stmt, alternateBody));
        }
        // --- Combined if statement
        this.inject(FunctionBuilder_1.ifBlock(consequentBody, alternateBody), body);
    }
    /**
     * Processes a local function invocation
     * @param body Body to put the statements in
     */
    compileLocalFunctionInvocation(invocation, body) {
        const resultTypeSpec = this.compileFunctionInvocation(invocation.invoked, body);
        if (!resultTypeSpec) {
            return;
        }
        // --- Remove the result of a non-void function
        if (resultTypeSpec.type !== "Void") {
            this.inject(FunctionBuilder_1.drop(), body);
        }
    }
    /**
     * Processes a return statement
     * @param body Body to put the statements in
     */
    compileReturn(retStmt, body) {
        // --- Check for return argument
        if (!this.func.resultType && retStmt.expr) {
            this.reportError("W156", retStmt);
            return;
        }
        if (this.func.resultType && !retStmt.expr) {
            this.reportError("W157", retStmt);
            return;
        }
        // --- Ok, we can compile this return
        if (retStmt.expr) {
            // --- Compile the expression
            const returnType = this.compileExpression(retStmt.expr, body);
            if (returnType === null) {
                return null;
            }
            // --- Cast it to the return type
            this.castForStorage(this.func.resultType, returnType, body);
        }
        // --- Issue the return
        this.inject(FunctionBuilder_1.ret(), body);
    }
    // ==========================================================================
    // Expression processing
    /**
     * Processes the specified expression
     * @param body Body to put the statements in
     */
    processExpression(expr, body) {
        this.addTrace(() => ["pExpr", 0, expression_resolver_1.renderExpression(expr)]);
        const simplified = this.simplifyExpression(expr);
        this.addTrace(() => ["pExpr", 1, expression_resolver_1.renderExpression(simplified)]);
        const exprType = this.compileExpression(simplified, body);
        if (!exprType) {
            return null;
        }
        return {
            expr: simplified,
            exprType,
        };
    }
    /**
     * Simplifies the expression
     * @param expr Expression to simplify
     */
    simplifyExpression(expr) {
        expr = this.removeTrivialLiteralsFromBinaryOps(expr);
        expr = this.orderLiteralsToRight(expr);
        expr = this.refoldBinaryOps(expr);
        expr = this.processLiterals(expr);
        return expr;
    }
    /**
     *
     * @param expr
     * @param action
     */
    visitExpression(expr, action) {
        switch (expr.type) {
            case "BinaryExpression":
                const left = this.visitExpression(expr.left, action);
                if (left !== expr.left) {
                    expr.left = left;
                }
                const right = this.visitExpression(expr.right, action);
                if (right !== expr.right) {
                    expr.right = right;
                }
                break;
            case "BuiltInFunctionInvocation":
            case "FunctionInvocation":
                for (let i = 0; i < expr.arguments.length; i++) {
                    const arg = this.visitExpression(expr.arguments[i], action);
                    if (arg !== expr.arguments[i]) {
                        expr.arguments[i] = arg;
                    }
                }
                break;
            case "ConditionalExpression":
                const condition = this.visitExpression(expr.condition, action);
                if (condition !== expr.condition) {
                    expr.condition = condition;
                }
                const consequent = this.visitExpression(expr.consequent, action);
                if (consequent !== expr.consequent) {
                    expr.consequent = consequent;
                }
                const alternate = this.visitExpression(expr.alternate, action);
                if (alternate !== expr.alternate) {
                    expr.condition = condition;
                }
                break;
            case "ItemAccess":
                const array = this.visitExpression(expr.array, action);
                if (array !== expr.array) {
                    expr.array = array;
                }
                const index = this.visitExpression(expr.index, action);
                if (index !== expr.index) {
                    expr.index = index;
                }
                break;
            case "MemberAccess":
                const obj = this.visitExpression(expr.object, action);
                if (obj !== expr.object) {
                    expr.object = obj;
                }
                break;
            case "TypeCast":
            case "UnaryExpression":
                const operand = this.visitExpression(expr.operand, action);
                if (operand !== expr.operand) {
                    expr.operand = operand;
                }
                break;
        }
        return action(expr);
    }
    /**
     * Flips constant values to the right for
     * commutative binary ops
     * @param expr
     */
    orderLiteralsToRight(expr) {
        return this.visitExpression(expr, (e) => {
            const commExpr = isCommutativeOp(e);
            if (!commExpr) {
                return e;
            }
            if (commExpr.left.type === "Literal" &&
                commExpr.right.type !== "Literal") {
                const tmp = commExpr.left;
                commExpr.left = commExpr.right;
                commExpr.right = tmp;
            }
            return commExpr;
        });
    }
    /**
     * Removes literals from expressions, which do not change the result of an operation
     * @param expr Expression to manage
     */
    removeTrivialLiteralsFromBinaryOps(expr) {
        return this.visitExpression(expr, (e) => {
            if (e.type !== "BinaryExpression") {
                return e;
            }
            switch (e.operator) {
                case "+":
                case "|":
                case "^":
                    if (e.left.type === "Literal" && e.left.value === 0) {
                        return e.right;
                    }
                    if (e.right.type === "Literal" && e.right.value === 0) {
                        return e.left;
                    }
                    break;
                case ">>":
                case ">>>":
                case "<<":
                    if (e.right.type === "Literal" && e.right.value === 0) {
                        return e.left;
                    }
                    break;
                case "-":
                    if (e.left.type === "Literal" && e.left.value === 0) {
                        return {
                            type: "UnaryExpression",
                            operator: "-",
                            operand: e.right,
                        };
                    }
                    if (e.right.type === "Literal" && e.right.value === 0) {
                        return e.left;
                    }
                    break;
                case "*":
                    if (e.right.type === "Literal" && e.right.value === 1) {
                        return e.left;
                    }
                    if (e.left.type === "Literal" && e.left.value === 1) {
                        return e.right;
                    }
                    break;
                case "/":
                    if (e.right.type === "Literal" && e.right.value === 1) {
                        return e.left;
                    }
                    break;
                case "%":
                    if (e.right.type === "Literal" && e.right.value === 1) {
                        return createLiteral(0);
                    }
                    break;
                case "&":
                    if ((e.left.type === "Literal" && e.left.value === 0) ||
                        (e.right.type === "Literal" && e.right.value === 0)) {
                        return createLiteral(0);
                    }
                    break;
            }
            return e;
        });
    }
    /**
     * Processes literal values; evaluates constant expressions
     * @param expr Expression to manage
     */
    processLiterals(expr) {
        const compiler = this.wsCompiler;
        return this.visitExpression(expr, (e) => calculate(e));
        function calculate(e) {
            switch (e.type) {
                case "ConditionalExpression":
                    if (e.condition.type === "Literal" &&
                        e.consequent.type === "Literal" &&
                        e.alternate.type === "Literal") {
                        return createLiteral(e.condition.value ? e.consequent.value : e.alternate.value);
                    }
                    break;
                case "BinaryExpression":
                    if (e.left.type === "Literal" && e.right.type === "Literal") {
                        return createLiteral(expression_resolver_1.applyBinaryOperation(e.operator, e.left.value, e.right.value));
                    }
                    break;
                case "UnaryExpression":
                    if (e.operand.type === "Literal") {
                        return createLiteral(expression_resolver_1.applyUnaryOperation(e.operator, e.operand.value));
                    }
                    break;
                case "BuiltInFunctionInvocation":
                    const nonLiteralArgs = e.arguments.filter((a) => a.type !== "Literal");
                    if (nonLiteralArgs.length === 0) {
                        return createLiteral(expression_resolver_1.applyBuiltInFunction(e.name, e.arguments.map((a) => a.value)));
                    }
                    break;
                case "TypeCast":
                    if (e.operand.type === "Literal") {
                        try {
                            const cast = expression_resolver_1.applyTypeCast(e.name, e.operand.value);
                            return createLiteral(cast, e.name === "u64");
                        }
                        catch (_a) {
                            // --- Intentionally ignored
                        }
                    }
                    break;
                case "SizeOfExpression":
                    compiler.resolveDependencies(e.spec);
                    return createLiteral(compiler.getSizeof(e.spec));
                case "Identifier":
                    const decl = compiler.declarations.get(e.name);
                    if ((decl === null || decl === void 0 ? void 0 : decl.type) === "ConstDeclaration") {
                        return createLiteral(decl.value);
                    }
                    break;
            }
            return e;
        }
    }
    /**
     * Refolds expressions to process literals
     * @param expr
     */
    refoldBinaryOps(expr) {
        return this.visitExpression(expr, (e) => refold(e));
        function refold(expr) {
            if (expr.type !== "BinaryExpression" ||
                expr.right.type !== "Literal" ||
                expr.left.type !== "BinaryExpression" ||
                expr.left.right.type !== "Literal") {
                return expr;
            }
            // --- The expression is like (expr binop2 literal2) binop1 literal1)
            const binop1 = expr.operator;
            const literal1 = expr.right.value;
            const binop2 = expr.left.operator;
            const literal2 = expr.left.right.value;
            switch (binop2) {
                case "+":
                    if (binop1 === "+") {
                        return foldLiteralIntoBinary(expr.left, add(literal2, literal1));
                    }
                    else if (binop1 === "-") {
                        return foldLiteralIntoBinary(expr.left, add(literal2, -literal1));
                    }
                case "-":
                    if (binop1 === "+") {
                        return foldLiteralIntoBinary(expr.left, add(literal2, -literal1));
                    }
                    else if (binop1 === "-") {
                        return foldLiteralIntoBinary(expr.left, add(literal2, literal1));
                    }
            }
            return expr;
        }
        function foldLiteralIntoBinary(binExpr, value) {
            return {
                type: "BinaryExpression",
                operator: binExpr.operator,
                left: binExpr.left,
                right: createLiteral(value),
            };
        }
        function add(left, right) {
            return typeof left === "number" && typeof right === "number"
                ? left + right
                : BigInt(left) + BigInt(right);
        }
    }
    // ==========================================================================
    // Expression compilation
    compileExpressionWithNoEmit(expr, body) {
        this._suspendCount++;
        try {
            return this.compileExpression(expr, body);
        }
        finally {
            this._suspendCount--;
        }
    }
    /**
     * Compiles the specified expression
     * @param expr Expression to compile
     * @param body Body to inject the code into
     * @returns Type specification of the result
     */
    compileExpression(expr, body) {
        switch (expr.type) {
            case "Literal":
                return this.compileLiteral(expr, body);
            case "Identifier":
                return this.compileIdentifier(expr, body);
            case "UnaryExpression":
                return this.compileUnaryExpression(expr, body);
            case "BinaryExpression":
                return this.compileBinaryExpression(expr, body);
            case "ConditionalExpression":
                return this.compileConditionalExpression(expr, body);
            case "TypeCast":
                return this.compileTypeCast(expr, body);
            case "MemberAccess":
            case "ItemAccess":
            case "DereferenceExpression":
                return this.compileIndirectAccess(expr, body);
            case "BuiltInFunctionInvocation":
                return this.compileBuiltinFunctionInvocation(expr, body);
            case "FunctionInvocation":
                return this.compileFunctionInvocation(expr, body);
        }
        return i32Desc;
    }
    /**
     * Compiles a literal
     * @param lit Literal to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileLiteral(lit, body) {
        let instr;
        let typeSpec;
        if (typeof lit.value === "number") {
            if (Number.isInteger(lit.value)) {
                const shrunkVal = Number(BigInt.asIntN(32, BigInt(lit.value)));
                instr = FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, shrunkVal);
                typeSpec = i32Desc;
            }
            else {
                instr = FunctionBuilder_1.constVal(wa_nodes_1.WaType.f64, lit.value);
                typeSpec = f64Desc;
            }
        }
        else {
            const shrunkVal = BigInt.asIntN(64, BigInt(lit.value));
            instr = FunctionBuilder_1.constVal(wa_nodes_1.WaType.i64, shrunkVal);
            typeSpec = i64Desc;
        }
        this.inject(instr, body);
        return typeSpec;
    }
    /**
     * Compiles an identifier
     * @param id Identifier to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileIdentifier(id, body) {
        // --- Check for a constant value
        const local = this._locals.get(id.name);
        if (!local) {
            // --- Ok, this name in not for a local
            const decl = this.wsCompiler.declarations.get(id.name);
            if (decl && decl.type === "ConstDeclaration") {
                return this.compileLiteral({
                    type: "Literal",
                    value: decl.value,
                }, body);
            }
        }
        // --- Go on with a non-constant identifier
        const resolvedId = this.resolveIdentifier(id);
        if (!resolvedId) {
            return null;
        }
        if (resolvedId.local) {
            this.inject(FunctionBuilder_1.localGet(resolvedId.local.name), body);
            return resolvedId.local.type;
        }
        if (resolvedId.global) {
            this.inject(FunctionBuilder_1.globalGet(WatSharpCompiler_1.createGlobalName(resolvedId.global.name)), body);
            return {
                type: "Intrinsic",
                underlying: resolvedId.global.underlyingType,
            };
        }
        if (resolvedId.var) {
            const typeSpec = resolvedId.var.spec;
            if (typeSpec.type !== "Intrinsic" && typeSpec.type !== "Pointer") {
                this.reportError("W143", id);
                return null;
            }
            this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, resolvedId.var.address), body);
            this.compileIntrinsicVariableGet(typeSpec.type === "Intrinsic" ? typeSpec : i32Desc, body);
            return typeSpec;
        }
        if (resolvedId.data) {
            const typeSpec = resolvedId.data.spec
                .spec;
            this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, resolvedId.data.address), body);
            this.compileIntrinsicVariableGet(typeSpec, body);
            return typeSpec;
        }
    }
    /**
     * Compiles getter to a variable with the specified type
     * @param typeSpec Variabel type
     * @param body Body to put the statements in
     */
    compileIntrinsicVariableGet(typeSpec, body) {
        const waType = WatSharpCompiler_1.waTypeMappings[typeSpec.underlying];
        switch (typeSpec.underlying) {
            case "bool":
                this.inject(FunctionBuilder_1.load(waType, wa_nodes_1.WaBitSpec.Bit8), body);
                break;
            case "f32":
            case "f64":
                this.inject(FunctionBuilder_1.load(waType), body);
                break;
            case "i8":
            case "u8":
                this.inject(FunctionBuilder_1.load(waType, wa_nodes_1.WaBitSpec.Bit8, undefined, undefined, typeSpec.underlying === "i8"), body);
                break;
            case "i16":
            case "u16":
                this.inject(FunctionBuilder_1.load(waType, wa_nodes_1.WaBitSpec.Bit16, undefined, undefined, typeSpec.underlying === "i16"), body);
                break;
            case "i32":
            case "u32":
                this.inject(FunctionBuilder_1.load(waType, wa_nodes_1.WaBitSpec.Bit32, undefined, undefined, typeSpec.underlying === "i32"), body);
                break;
            case "i64":
            case "u64":
                this.inject(FunctionBuilder_1.load(waType, undefined, undefined, undefined, typeSpec.underlying === "i64"), body);
        }
    }
    /**
     * Compiles setter to a variable with the specified type
     * @param typeSpec Variabel type
     * @param body Body to put the statements in
     */
    compileIntrinsicVariableSet(typeSpec, body) {
        const waType = WatSharpCompiler_1.waTypeMappings[typeSpec.underlying];
        switch (typeSpec.underlying) {
            case "f32":
            case "f64":
                this.inject(FunctionBuilder_1.store(waType), body);
                break;
            case "bool":
            case "i8":
            case "u8":
                this.inject(FunctionBuilder_1.store(waType, wa_nodes_1.WaBitSpec.Bit8), body);
                break;
            case "i16":
            case "u16":
                this.inject(FunctionBuilder_1.store(waType, wa_nodes_1.WaBitSpec.Bit16), body);
                break;
            case "i32":
            case "u32":
                this.inject(FunctionBuilder_1.store(waType, wa_nodes_1.WaBitSpec.Bit32), body);
                break;
            case "i64":
            case "u64":
                this.inject(FunctionBuilder_1.store(waType), body);
        }
    }
    /**
     * Compiles a unary expression
     * @param unary Expression to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileUnaryExpression(unary, body) {
        var _a;
        switch (unary.operator) {
            case "+": {
                // --- Compile the operand
                const operandType = this.compileExpression(unary.operand, body);
                if (operandType === null) {
                    return null;
                }
                // --- Allow intrinsic types only
                if (operandType.type !== "Intrinsic") {
                    this.reportError("W144", unary, "unary +");
                    return null;
                }
                // --- "+" means type cast to i32
                this.castIntrinsicToIntrinsic(i32Desc, operandType, body, (_a = unary.operand) === null || _a === void 0 ? void 0 : _a.value);
                return i32Desc;
            }
            case "-": {
                // --- Compile the operand
                const operandType = this.compileExpression(unary.operand, body);
                if (operandType === null) {
                    return null;
                }
                // --- Allow intrinsic types only
                if (operandType.type !== "Intrinsic") {
                    this.reportError("W144", unary, "unary -");
                    return null;
                }
                // --- "-" --> -1 * operand
                const waType = WatSharpCompiler_1.waTypeMappings[operandType.underlying];
                this.inject(FunctionBuilder_1.constVal(waType, -1), body);
                this.inject(FunctionBuilder_1.mul(waType), body);
                return operandType;
            }
            case "!":
            case "~": {
                // --- Compile the operand
                const operandType = this.compileExpression(unary.operand, body);
                if (operandType === null) {
                    return null;
                }
                // --- Allow intrinsic types only
                if (operandType.type !== "Intrinsic") {
                    this.reportError("W145", unary, "logical NOT");
                    return null;
                }
                const waType = WatSharpCompiler_1.waTypeMappings[operandType.underlying];
                // --- Allow integer types only
                if (waType === wa_nodes_1.WaType.f32 || waType === wa_nodes_1.WaType.f64) {
                    this.reportError("W145", unary, "logical NOT");
                    return null;
                }
                if (unary.operator === "!") {
                    // --- "!" --> eqz
                    this.inject(FunctionBuilder_1.eqz(waType), body);
                    return i32Desc;
                }
                else {
                    // --- "~" --> xor with all bits 1
                    this.inject(FunctionBuilder_1.constVal(waType, WatSharpCompiler_1.bitwiseNotMasks[operandType.underlying]), body);
                    this.inject(FunctionBuilder_1.xor(waType), body);
                    return operandType;
                }
            }
            case "&": {
                const address = this.calculateAddressOf(unary.operand, body);
                if (address === null) {
                    return null;
                }
                return i32Desc;
            }
        }
    }
    /**
     * Compiles a binary expression
     * @param binary Expression to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileBinaryExpression(binary, body) {
        // --- Compile the left and right operands to obtain result types
        const left = this.compileExpressionWithNoEmit(binary.left, body);
        if (left === null) {
            return null;
        }
        const right = this.compileExpressionWithNoEmit(binary.right, body);
        if (right === null) {
            return null;
        }
        // --- Special case: pointer arithmetic
        if (left.type === "Pointer" &&
            (right.type === "Intrinsic" || right.type === "Pointer")) {
            return this.compilePointerBinaryExpression(binary.left, left, binary.right, right, binary.operator, body);
        }
        // --- Make sure both operands are intrinsic
        if (left.type !== "Intrinsic" || right.type !== "Intrinsic") {
            this.reportError("W144", binary, `binary ${binary.operator}`);
            return null;
        }
        // --- Check if the operation should be signed
        const isSigned = left.underlying.startsWith("i") || right.underlying.startsWith("i");
        // --- Calculate operation type
        const op = binary.operator;
        let operationType = i32Desc;
        if (left.underlying.startsWith("f") || right.underlying.startsWith("f")) {
            operationType = f64Desc;
        }
        else if (left.underlying.endsWith("64") ||
            right.underlying.endsWith("64")) {
            operationType = i64Desc;
        }
        // --- Calculate result type
        let resultType = i32Desc;
        if (op !== "==" &&
            op !== "!=" &&
            op !== "<" &&
            op !== "<=" &&
            op !== ">" &&
            op !== ">=") {
            resultType = operationType;
        }
        // --- Compile the operands and cast them to the appropriate type
        this.compileExpression(binary.left, body);
        this.castIntrinsicToIntrinsic(operationType, left, body);
        this.compileExpression(binary.right, body);
        this.castIntrinsicToIntrinsic(operationType, right, body);
        const waType = WatSharpCompiler_1.waTypeMappings[operationType.underlying];
        // --- Process operations
        switch (binary.operator) {
            case "+":
                this.inject(FunctionBuilder_1.add(waType), body);
                break;
            case "-":
                this.inject(FunctionBuilder_1.sub(waType), body);
                break;
            case "*":
                this.inject(FunctionBuilder_1.mul(waType), body);
                break;
            case "/":
                this.inject(FunctionBuilder_1.div(waType, isSigned), body);
                break;
            case "%":
                if (operationType.underlying.startsWith("f")) {
                    this.reportError("W145", binary, "remainder (%)");
                    return null;
                }
                this.inject(FunctionBuilder_1.rem(waType, isSigned), body);
                break;
            case "&":
                if (operationType.underlying.startsWith("f")) {
                    this.reportError("W145", binary, "bitwise AND");
                    return null;
                }
                this.inject(FunctionBuilder_1.and(waType), body);
                break;
            case "|":
                if (operationType.underlying.startsWith("f")) {
                    this.reportError("W145", binary, "bitwise OR");
                    return null;
                }
                this.inject(FunctionBuilder_1.or(waType), body);
                break;
            case "^":
                if (operationType.underlying.startsWith("f")) {
                    this.reportError("W145", binary, "bitwise XOR");
                    return null;
                }
                this.inject(FunctionBuilder_1.xor(waType), body);
                break;
            case "<<":
                if (operationType.underlying.startsWith("f")) {
                    this.reportError("W145", binary, "shift left");
                    return null;
                }
                this.inject(FunctionBuilder_1.shl(waType), body);
                break;
            case ">>":
                if (operationType.underlying.startsWith("f")) {
                    this.reportError("W145", binary, "signed shift right");
                    return null;
                }
                this.inject(FunctionBuilder_1.shr(waType, true), body);
                break;
            case ">>>":
                if (operationType.underlying.startsWith("f")) {
                    this.reportError("W145", binary, "shift right");
                    return null;
                }
                this.inject(FunctionBuilder_1.shr(waType, false), body);
                break;
            case "==":
                this.inject(FunctionBuilder_1.eq(waType), body);
                break;
            case "!=":
                this.inject(FunctionBuilder_1.ne(waType), body);
                break;
            case "<":
                this.inject(FunctionBuilder_1.lt(waType, isSigned), body);
                break;
            case "<=":
                this.inject(FunctionBuilder_1.le(waType, isSigned), body);
                break;
            case ">":
                this.inject(FunctionBuilder_1.gt(waType, isSigned), body);
                break;
            case ">=":
                this.inject(FunctionBuilder_1.ge(waType, isSigned), body);
                break;
        }
        return resultType;
    }
    /**
     * Compiles a pointer binary operation
     * @param left Left operand
     * @param leftType Left operand type
     * @param right Right operand
     * @param rightType Right operand type
     * @param op Operation
     * @param body
     */
    compilePointerBinaryExpression(left, leftType, right, rightType, op, body) {
        // --- Check operands
        switch (op) {
            case "+":
            case "-":
                if (rightType.type !== "Intrinsic" ||
                    rightType.underlying.startsWith("f")) {
                    this.reportError("W165", right);
                    return null;
                }
                break;
            case "==":
            case "!=":
            case "<":
            case "<=":
            case ">":
            case ">=":
                if (rightType.type !== "Pointer") {
                    this.reportError("W171", right);
                    return null;
                }
                break;
            default:
                this.reportError("W164", left);
                return null;
        }
        // --- Get the left pointer value
        this.compileExpression(left, body);
        if (rightType.type === "Intrinsic") {
            // --- Pointer addition or subtraction
            // --- Get the increment value
            this.compileExpression(right, body);
            this.castIntrinsicToIntrinsic(i32Desc, rightType, body);
            // --- Get the multiplier value
            this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, this.wsCompiler.getSizeof(leftType.spec)), body);
            // --- Calculate the result
            this.inject(FunctionBuilder_1.mul(wa_nodes_1.WaType.i32), body);
            this.inject(op === "+" ? FunctionBuilder_1.add(wa_nodes_1.WaType.i32) : FunctionBuilder_1.sub(wa_nodes_1.WaType.i32), body);
            return leftType;
        }
        else {
            // --- Pointer comparison
            // --- Get the increment value
            this.compileExpression(right, body);
            switch (op) {
                case "==":
                    this.inject(FunctionBuilder_1.eq(wa_nodes_1.WaType.i32), body);
                    break;
                case "!=":
                    this.inject(FunctionBuilder_1.ne(wa_nodes_1.WaType.i32), body);
                    break;
                case "<":
                    this.inject(FunctionBuilder_1.lt(wa_nodes_1.WaType.i32), body);
                    break;
                case "<=":
                    this.inject(FunctionBuilder_1.le(wa_nodes_1.WaType.i32), body);
                    break;
                case ">":
                    this.inject(FunctionBuilder_1.gt(wa_nodes_1.WaType.i32), body);
                    break;
                case ">=":
                    this.inject(FunctionBuilder_1.ge(wa_nodes_1.WaType.i32), body);
                    break;
            }
            return i32Desc;
        }
    }
    /**
     * Compiles a conditional expression
     * @param conditional Expression to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileConditionalExpression(conditional, body) {
        // --- Compile the condition, consequent, and alternate values
        const condition = this.compileExpressionWithNoEmit(conditional.condition, body);
        if (condition === null) {
            return null;
        }
        const consequent = this.compileExpressionWithNoEmit(conditional.consequent, body);
        if (consequent === null) {
            return null;
        }
        const alternate = this.compileExpressionWithNoEmit(conditional.alternate, body);
        if (alternate === null) {
            return null;
        }
        // --- Make sure both operands are intrinsic
        if (condition.type !== "Intrinsic" ||
            consequent.type !== "Intrinsic" ||
            alternate.type !== "Intrinsic") {
            this.reportError("W144", conditional, "conditional");
            return null;
        }
        // --- Calculate operation type
        let resultType = i32Desc;
        if (consequent.underlying.startsWith("f") ||
            alternate.underlying.startsWith("f")) {
            resultType = f64Desc;
        }
        else if (alternate.underlying.endsWith("64") ||
            alternate.underlying.endsWith("64")) {
            resultType = i64Desc;
        }
        // --- Compile the operands and cast them to the appropriate type
        this.compileExpression(conditional.condition, body);
        this.castIntrinsicToIntrinsic(i32Desc, condition, body);
        // --- Consequent branch
        const consequentBody = [];
        this.compileExpression(conditional.consequent, consequentBody);
        this.castIntrinsicToIntrinsic(resultType, consequent, consequentBody);
        // --- Alternate branch
        const alternateBody = [];
        this.compileExpression(conditional.alternate, alternateBody);
        this.castIntrinsicToIntrinsic(resultType, alternate, alternateBody);
        // --- Combined if statement
        this.inject(FunctionBuilder_1.ifBlock(consequentBody, alternateBody, WatSharpCompiler_1.waTypeMappings[resultType.underlying]), body);
        // --- Done
        return resultType;
    }
    /**
     * Compiles a type cast
     * @param cast Expression to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileTypeCast(cast, body) {
        const operand = this.compileExpressionWithNoEmit(cast.operand, body);
        if (operand === null) {
            return null;
        }
        if (operand.type !== "Intrinsic") {
            this.reportError("W144", cast, `${cast.name}()`);
            return null;
        }
        const resultType = {
            type: "Intrinsic",
            underlying: cast.name,
        };
        this.compileExpression(cast.operand, body);
        this.castIntrinsicToIntrinsic(resultType, operand, body);
        return resultType;
    }
    /**
     * Compiles an indirect value access
     * @param expr Expression to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileIndirectAccess(expr, body) {
        const varAddr = this.calculateAddressOf(expr, body);
        if (varAddr == null) {
            return null;
        }
        let typeSpec = varAddr.spec;
        if (typeSpec.type === "Intrinsic") {
            this.compileIntrinsicVariableGet(typeSpec, body);
        }
        return typeSpec;
    }
    /**
     * Compiles a built-in function invocation
     * @param func Expression to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileBuiltinFunctionInvocation(func, body) {
        // --- Prepare function argument types
        const argTypes = func.arguments.map((arg) => {
            const type = this.compileExpressionWithNoEmit(arg, body);
            return type;
        });
        const hasF64Arg = argTypes.filter((arg) => arg.type === "Intrinsic" && arg.underlying === "f64").length > 0;
        const argType = argTypes.length == 0 ? undefined : argTypes[0];
        const argIns = argType ? argType.underlying : "";
        let waType = argIns === "" ? wa_nodes_1.WaType.i32 : WatSharpCompiler_1.waTypeMappings[argIns];
        let resultType = argType;
        // --- Compile function argument types
        func.arguments.forEach((arg, index) => {
            this.compileExpression(arg, body);
            if (func.name === "min" || func.name === "max") {
                this.castIntrinsicToIntrinsic(hasF64Arg ? f64Desc : f32Desc, argTypes[index], body);
            }
        });
        // --- Inject the appropriate operation
        switch (func.name) {
            case "abs": {
                if (argIns.startsWith("i")) {
                    // --- Integer absolute value
                    const local = this.createTempLocal(waType);
                    this.inject([
                        FunctionBuilder_1.localTee(local),
                        FunctionBuilder_1.constVal(waType, 0),
                        FunctionBuilder_1.lt(waType, true),
                        FunctionBuilder_1.ifBlock([FunctionBuilder_1.localGet(local), FunctionBuilder_1.constVal(waType, -1), FunctionBuilder_1.mul(waType)], [FunctionBuilder_1.localGet(local)], waType),
                    ], body);
                }
                else if (argIns.startsWith("f")) {
                    this.inject(FunctionBuilder_1.abs(waType), body);
                }
                break;
            }
            case "ceil":
                if (waType === wa_nodes_1.WaType.i32 || waType === wa_nodes_1.WaType.i64) {
                    this.reportError("W150", func, "ceil");
                    return null;
                }
                this.inject(FunctionBuilder_1.ceil(waType), body);
                break;
            case "clz":
                if (waType === wa_nodes_1.WaType.f32 || waType === wa_nodes_1.WaType.f64) {
                    this.reportError("W151", func, "clz");
                    return null;
                }
                this.inject(FunctionBuilder_1.clz(waType), body);
                break;
            case "copysign":
                if (waType === wa_nodes_1.WaType.i32 || waType === wa_nodes_1.WaType.i64) {
                    this.reportError("W150", func, "copysign");
                    return null;
                }
                this.inject(FunctionBuilder_1.copysign(waType), body);
                break;
            case "ctz":
                if (waType === wa_nodes_1.WaType.f32 || waType === wa_nodes_1.WaType.f64) {
                    this.reportError("W151", func, "ctz");
                    return null;
                }
                this.inject(FunctionBuilder_1.ctz(waType), body);
                break;
            case "floor":
                if (waType === wa_nodes_1.WaType.i32 || waType === wa_nodes_1.WaType.i64) {
                    this.reportError("W150", func, "floor");
                    return null;
                }
                this.inject(FunctionBuilder_1.floor(waType), body);
                break;
            case "nearest":
                if (waType === wa_nodes_1.WaType.i32 || waType === wa_nodes_1.WaType.i64) {
                    this.reportError("W150", func, "nearest");
                    return null;
                }
                this.inject(FunctionBuilder_1.nearest(waType), body);
                break;
            case "neg":
                if (waType === wa_nodes_1.WaType.i32 || waType === wa_nodes_1.WaType.i64) {
                    this.reportError("W150", func, "neg");
                    return null;
                }
                this.inject(FunctionBuilder_1.neg(waType), body);
                break;
            case "popcnt":
                if (waType === wa_nodes_1.WaType.f32 || waType === wa_nodes_1.WaType.f64) {
                    this.reportError("W151", func, "popcnt");
                    return null;
                }
                this.inject(FunctionBuilder_1.popcnt(waType), body);
                break;
            case "sqrt":
                if (waType === wa_nodes_1.WaType.i32 || waType === wa_nodes_1.WaType.i64) {
                    this.reportError("W150", func, "sqrt");
                    return null;
                }
                this.inject(FunctionBuilder_1.sqrt(waType), body);
                break;
            case "min":
            case "max":
                resultType = hasF64Arg ? f64Desc : f32Desc;
                waType = hasF64Arg ? wa_nodes_1.WaType.f64 : wa_nodes_1.WaType.f32;
                for (let i = 1; i < func.arguments.length; i++) {
                    this.inject(func.name === "min" ? FunctionBuilder_1.min(waType) : FunctionBuilder_1.max(waType), body);
                }
                break;
        }
        return resultType;
    }
    /**
     * Compiles a function invocation
     * @param invoc Expression to compile
     * @param body Body to put the statements in
     * @returns Type specification of the result
     */
    compileFunctionInvocation(invoc, body) {
        // --- Check if the invoked function exists
        const calledDecl = this.wsCompiler.declarations.get(invoc.name);
        if (!calledDecl) {
            this.reportError("W153", invoc, invoc.name);
            return null;
        }
        // --- Only function and table invocations are allowed
        if (calledDecl.type !== "FunctionDeclaration" &&
            calledDecl.type !== "TableDeclaration" &&
            calledDecl.type !== "ImportedFunctionDeclaration") {
            this.reportError("W153", invoc, invoc.name);
            return null;
        }
        // --- Can we inline it?
        let inlineIt = false;
        let funcId = -1;
        if (calledDecl.type === "FunctionDeclaration") {
            inlineIt = calledDecl.canBeInlined;
            funcId = calledDecl.funcId;
        }
        // --- Prepare the function parameters
        const funcParams = calledDecl.type === "ImportedFunctionDeclaration"
            ? calledDecl.parSpecs.map((sp, idx) => ({
                name: `$param${idx}`,
                spec: sp,
            }))
            : calledDecl.params;
        const funcResult = calledDecl.resultType;
        // --- Check the dispatch expression
        if (calledDecl.type === "TableDeclaration") {
            if (!invoc.dispatcher) {
                this.reportError("W159", invoc);
                return null;
            }
        }
        else if (invoc.dispatcher) {
            this.reportError("W160", invoc);
            return null;
        }
        // --- Test if number of arguments equals the number of parameters
        if (funcParams.length !== invoc.arguments.length) {
            this.reportError("W154", invoc, funcParams.length, invoc.arguments.length);
            return null;
        }
        // --- We use this during inline optimization
        let lastInlineParamName;
        let startInstructionIndex = 0;
        const builderBody = this._builder.body;
        // --- Match and convert parameter types one-by-one
        for (let i = 0; i < funcParams.length; i++) {
            startInstructionIndex = builderBody.length;
            const par = funcParams[i];
            const waParType = par.spec.type === "Pointer"
                ? wa_nodes_1.WaType.i32
                : WatSharpCompiler_1.waTypeMappings[par.spec.underlying];
            const argType = this.compileExpression(invoc.arguments[i], body);
            if (!argType) {
                this.reportError("W142", par);
                return null;
            }
            const waArgType = argType.type === "Pointer"
                ? wa_nodes_1.WaType.i32
                : WatSharpCompiler_1.waTypeMappings[argType.underlying];
            // --- Do not allow implicit type mappings from floating-point to integer
            if ((waParType === wa_nodes_1.WaType.i32 || waParType === wa_nodes_1.WaType.i64) &&
                (waArgType === wa_nodes_1.WaType.f32 || waArgType === wa_nodes_1.WaType.f64)) {
                this.reportError("W155", invoc);
                return null;
            }
            // --- Argument is OK, cast it to the parameter's type
            this.castForStorage(par.spec, argType, body);
            // --- Handle parameters of an inline function
            if (inlineIt) {
                const inlineParamName = `$${funcId}${WatSharpCompiler_1.createParameterName(par.name)}`;
                this.createCompileTimeLocal(inlineParamName, par.spec);
                this.inject(FunctionBuilder_1.localSet(inlineParamName), body);
                lastInlineParamName = inlineParamName;
            }
        }
        // --- Call the function
        if (calledDecl.type === "TableDeclaration") {
            // --- Non-inlined function call
            this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, calledDecl.entryIndex), body);
            const exprType = this.compileExpression(invoc.dispatcher, body);
            if (exprType) {
                this.castForStorage(i32Desc, exprType, body);
                this.inject(FunctionBuilder_1.add(wa_nodes_1.WaType.i32), body);
                this.inject(FunctionBuilder_1.callIndirect(WatSharpCompiler_1.createTableName(invoc.name)), body);
            }
        }
        else {
            // --- We call the function, count the invocations
            if (calledDecl.type !== "ImportedFunctionDeclaration") {
                calledDecl.invocationCount++;
            }
            if (inlineIt) {
                // --- Just copy the locals and the body of the inline function
                this.wsCompiler.getFunctionLocals(calledDecl.name).forEach((loc) => {
                    if (!this._builder.locals.some((l) => l.id === loc.id)) {
                        this._builder.locals.push(loc);
                    }
                });
                this.wsCompiler
                    .getFunctionBodyInstructions(calledDecl.name)
                    .forEach((ins) => {
                    this.inject(ins, body);
                });
                // --- Do optimization
                wat_optimizer_1.optimizeConstants(builderBody);
                wat_optimizer_1.optimizeLastInlineParam(builderBody, lastInlineParamName, startInstructionIndex);
                wat_optimizer_1.optimizeConstants(builderBody);
            }
            else {
                this.inject(FunctionBuilder_1.call(WatSharpCompiler_1.createGlobalName(invoc.name)), body);
            }
            if (inlineIt) {
                // --- Obtain the return value from the inline function
                if (calledDecl.type !== "ImportedFunctionDeclaration" &&
                    calledDecl.hasReturn) {
                    const resultName = `$${funcId}$res`;
                    this.createCompileTimeLocal(resultName, calledDecl.resultType);
                    this.inject(FunctionBuilder_1.localGet(resultName), body);
                }
            }
        }
        // --- Done
        return funcResult !== null && funcResult !== void 0 ? funcResult : voidDesc;
    }
    /**
     * Casts a storage type to another storage type
     * @param left
     * @param right
     * @param body Body to put the statements in
     */
    castForStorage(left, right, body, value) {
        switch (left.type) {
            case "Intrinsic":
                if (right.type !== "Intrinsic") {
                    this.reportError("W141", right);
                    return;
                }
                this.castIntrinsicToIntrinsic(left, right, body, value);
                break;
            case "Pointer":
                if (right.type === "Pointer") {
                    return;
                }
                if (right.type !== "Intrinsic" ||
                    right.underlying === "f64" ||
                    right.underlying === "f32") {
                    this.reportError("W141", right);
                    return;
                }
                if (right.underlying === "i64" || right.underlying === "u64") {
                    this.inject(FunctionBuilder_1.wrap64(), body);
                }
                break;
        }
    }
    /**
     * Casts an intinsice type to another intrinsic type
     * @param left Left value
     * @param right Right expression
     * @param body Array of instrcutions
     * @param value optional value to test
     */
    castIntrinsicToIntrinsic(left, right, body, value) {
        if (left.underlying === right.underlying) {
            return;
        }
        if ((left.underlying === "i32" && right.underlying === "u32") ||
            (left.underlying === "u32" && right.underlying === "i32")) {
            return;
        }
        if ((left.underlying === "i64" && right.underlying === "u64") ||
            (left.underlying === "u64" && right.underlying === "i64")) {
            return;
        }
        const compiler = this;
        switch (right.underlying) {
            case "i64":
            case "u64":
                switch (left.underlying) {
                    case "f32":
                        this.inject(FunctionBuilder_1.convert32(wa_nodes_1.WaType.i64, right.underlying === "i64"), body);
                        return;
                    case "f64":
                        this.inject(FunctionBuilder_1.convert64(wa_nodes_1.WaType.i64, right.underlying === "i64"), body);
                        return;
                    case "i32":
                    case "u32":
                        this.inject(FunctionBuilder_1.wrap64(), body);
                        return;
                    case "i16":
                    case "u16":
                        this.inject(FunctionBuilder_1.wrap64(), body);
                        tighten(0xffff, 16, left.underlying, body, value);
                        return;
                    case "i8":
                    case "u8":
                        this.inject(FunctionBuilder_1.wrap64(), body);
                        tighten(0xff, 24, left.underlying, body, value);
                        return;
                    case "bool":
                        this.inject(FunctionBuilder_1.eqz(wa_nodes_1.WaType.i64), body);
                        this.inject(FunctionBuilder_1.eqz(wa_nodes_1.WaType.i32), body);
                        return;
                }
                break;
            case "i32":
            case "u32":
            case "i16":
            case "u16":
            case "i8":
            case "u8":
                switch (left.underlying) {
                    case "f32":
                        this.inject(FunctionBuilder_1.convert32(wa_nodes_1.WaType.i32, right.underlying.startsWith("i")), body);
                        return;
                    case "f64":
                        this.inject(FunctionBuilder_1.convert64(wa_nodes_1.WaType.i32, right.underlying.startsWith("i")), body);
                        return;
                    case "i64":
                        this.inject(FunctionBuilder_1.extend32(true), body);
                        return;
                    case "u64":
                        this.inject(FunctionBuilder_1.extend32(false), body);
                        return;
                    case "i16":
                    case "u16":
                        tighten(0xffff, 16, left.underlying, body, value);
                        return;
                    case "i8":
                    case "u8":
                        tighten(0xff, 24, left.underlying, body, value);
                        return;
                    case "bool":
                        this.inject(FunctionBuilder_1.eqz(wa_nodes_1.WaType.i32), body);
                        this.inject(FunctionBuilder_1.eqz(wa_nodes_1.WaType.i32), body);
                        return;
                }
                break;
            case "f64":
                switch (left.underlying) {
                    case "f32":
                        this.inject(FunctionBuilder_1.demote64(), body);
                        return;
                    case "i64":
                        this.inject(FunctionBuilder_1.trunc64(wa_nodes_1.WaType.f64, true), body);
                        return;
                    case "u64":
                        this.inject(FunctionBuilder_1.trunc64(wa_nodes_1.WaType.f64, false), body);
                        return;
                    case "i32":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f64, true), body);
                        return;
                    case "u32":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f64, false), body);
                        return;
                    case "i16":
                    case "u16":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f64, false), body);
                        tighten(0xffff, 16, left.underlying, body, value);
                        return;
                    case "i8":
                    case "u8":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f64, false), body);
                        tighten(0xff, 24, left.underlying, body, value);
                        return;
                    case "bool":
                        this.inject(FunctionBuilder_1.trunc64(wa_nodes_1.WaType.f64, true), body);
                        this.inject(FunctionBuilder_1.eqz(wa_nodes_1.WaType.i64), body);
                        this.inject(FunctionBuilder_1.eqz(wa_nodes_1.WaType.i32), body);
                        return;
                }
                break;
            case "f32":
                switch (left.underlying) {
                    case "f64":
                        this.inject(FunctionBuilder_1.promote32(), body);
                        return;
                    case "i64":
                        this.inject(FunctionBuilder_1.trunc64(wa_nodes_1.WaType.f32, true), body);
                        return;
                    case "u64":
                        this.inject(FunctionBuilder_1.trunc64(wa_nodes_1.WaType.f32, false), body);
                        return;
                    case "i32":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f32, true), body);
                        return;
                    case "u32":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f32, false), body);
                        return;
                    case "i16":
                    case "u16":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f32, false), body);
                        tighten(0xffff, 16, left.underlying, body, value);
                        return;
                    case "i8":
                    case "u8":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f32, false), body);
                        tighten(0xff, 24, left.underlying, body, value);
                        return;
                    case "bool":
                        this.inject(FunctionBuilder_1.trunc32(wa_nodes_1.WaType.f32, true), body);
                        this.inject(FunctionBuilder_1.eqz(wa_nodes_1.WaType.i32), body);
                        this.inject(FunctionBuilder_1.eqz(wa_nodes_1.WaType.i32), body);
                        return;
                }
                break;
        }
        /**
         * Demotes a 32-bit value to a smaller one
         * @param mask Bit mask
         * @param bits Bit count
         * @param typename Type name
         * @param body Body to put the statements in
         * @param value: Optional value to check if tightening is needed at all
         */
        function tighten(mask, bits, typename, body, value) {
            if (value && typeof value === "number") {
                const rightBits = 32 - bits;
                const lower = typename.startsWith("i") ? -(Math.pow(2, (rightBits - 1))) : 0;
                const upper = lower + Math.pow(2, rightBits);
                if (value >= lower && value <= upper) {
                    return;
                }
            }
            compiler.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, mask), body);
            compiler.inject(FunctionBuilder_1.and(wa_nodes_1.WaType.i32), body);
            if (typename.startsWith("i")) {
                compiler.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, bits), body);
                compiler.inject(FunctionBuilder_1.shl(wa_nodes_1.WaType.i32), body);
                compiler.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, bits), body);
                compiler.inject(FunctionBuilder_1.shr(wa_nodes_1.WaType.i32, true), body);
            }
        }
    }
    /**
     * Resolves an identifier to a declaration
     * @param id
     */
    resolveIdentifier(id) {
        const local = this._locals.get(id.name);
        if (local) {
            return {
                local,
            };
        }
        const decl = this.wsCompiler.declarations.get(id.name);
        if (!decl) {
            this.reportError("W142", id, id.name);
            return null;
        }
        if (decl.type === "GlobalDeclaration") {
            return {
                global: decl,
            };
        }
        if (decl.type === "VariableDeclaration") {
            return {
                var: decl,
            };
        }
        if (decl.type === "DataDeclaration") {
            return {
                data: decl,
            };
        }
        this.reportError("W142", id, id.name);
        return null;
    }
    /**
     * Calculates the address of the specified expression
     * @param expr Address expression
     * @param body Body to put the statements in
     */
    calculateAddressOf(expr, body) {
        switch (expr.type) {
            case "Identifier": {
                // --- Only variables have an address
                const resolvedId = this.resolveIdentifier(expr);
                if (resolvedId === null) {
                    this.reportError("W146", expr);
                    return null;
                }
                // --- Is it a local pointer?
                if (resolvedId.local && resolvedId.local.type.type === "Pointer") {
                    const local = resolvedId.local;
                    this.inject(FunctionBuilder_1.localGet(local.name), body);
                    return {
                        address: 0,
                        spec: local.type,
                        getterCompiled: true,
                    };
                }
                if (!resolvedId.var && !resolvedId.data) {
                    this.reportError("W146", expr);
                    return null;
                }
                const address = resolvedId.var
                    ? resolvedId.var.address
                    : resolvedId.data.address;
                const spec = resolvedId.var
                    ? resolvedId.var.spec
                    : resolvedId.data.spec;
                // --- Inject variable address if requested
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, address), body);
                // --- Retrieve address/type information
                return {
                    address,
                    spec,
                };
            }
            case "DereferenceExpression": {
                // --- Start with the calculation of the operand address
                const operandAddr = this.calculateAddressOf(expr.operand, body);
                if (operandAddr === null) {
                    return null;
                }
                // --- Member access needs a struct object
                if (operandAddr.spec.type !== "Pointer") {
                    this.reportError("W152", expr);
                    return null;
                }
                // --- Load the pointer from the address
                if (!operandAddr.getterCompiled) {
                    this.inject(FunctionBuilder_1.load(wa_nodes_1.WaType.i32), body);
                }
                // --- Retrieve address/type information
                return {
                    address: operandAddr.address,
                    spec: operandAddr.spec.spec,
                };
            }
            case "MemberAccess": {
                // --- Start with the calculation of the object address
                const leftAddress = this.calculateAddressOf(expr.object, body);
                if (leftAddress === null) {
                    return null;
                }
                // --- Member access needs a struct object
                if (leftAddress.spec.type !== "Struct") {
                    this.reportError("W147", expr);
                    return null;
                }
                // --- Obtain struct field information
                const field = leftAddress.spec.fields.filter((fi) => fi.id === expr.member);
                if (!field || field.length === 0) {
                    this.reportError("W147", expr);
                    return null;
                }
                // --- Field exists, add its offset to the address
                let address = leftAddress.address;
                const offset = field[0].offset;
                if (offset !== 0) {
                    this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, offset), body);
                    this.inject(FunctionBuilder_1.add(wa_nodes_1.WaType.i32), body);
                }
                return {
                    address,
                    spec: field[0].spec,
                };
            }
            case "ItemAccess": {
                // --- Start with the calculation of the object address
                const arrayAddress = this.calculateAddressOf(expr.array, body);
                if (arrayAddress === null) {
                    return null;
                }
                // --- Calculate the item size
                let address = arrayAddress.address;
                if (arrayAddress.spec.type !== "Array") {
                    this.reportError("W149", expr);
                    return null;
                }
                const itemSize = this.wsCompiler.getSizeof(arrayAddress.spec.spec);
                const indexType = this.compileExpression(expr.index, body);
                if (indexType === null) {
                    return null;
                }
                this.castForStorage(i32Desc, indexType, body, expr.index.value);
                this.inject(FunctionBuilder_1.constVal(wa_nodes_1.WaType.i32, itemSize), body);
                this.inject(FunctionBuilder_1.mul(wa_nodes_1.WaType.i32), body);
                this.inject(FunctionBuilder_1.add(wa_nodes_1.WaType.i32), body);
                return {
                    address,
                    spec: arrayAddress.spec.spec,
                };
            }
            default:
                this.reportError("W170", expr);
                break;
        }
        return null;
    }
    // ==========================================================================
    // Inline function compilation
    /**
     * Prepares the current function for inlining after it was compiled
     * The compiler calls this method only on functions explicitly marked
     * for inlining
     */
    prepareForInlining() {
        this.func.canBeInlined = false;
        // --- No inlining for functions with more than two parameters
        if (this.func.params.length > 2) {
            return;
        }
        // --- No inlining over 32 instructions
        if (wat_helpers_1.countInstructions(this._builder.body) > 80) {
            return;
        }
        // --- No inlining for functions that contain calls to other functions
        if (wat_helpers_1.findInstruction(this._builder.body, (ins) => ins.type === "Call" || ins.type === "CallIndirect")) {
            return;
        }
        // --- No inlining with multiple returns
        let returnCount = 0;
        wat_helpers_1.visitInstructions(this._builder.body, (ins) => {
            if (ins.type === "Return") {
                returnCount++;
            }
        });
        if (returnCount > 1) {
            return;
        }
        // --- Return should be in the main branch of the function
        let mainReturnIndex = -1;
        this._builder.body.forEach((ins, index) => {
            if (ins.type === "Return") {
                mainReturnIndex = index;
            }
        });
        if (returnCount > 0 && mainReturnIndex < 0) {
            return;
        }
        // --- Ok, this function can be inlined, so let's carry out a few trnasforms
        this.func.canBeInlined = true;
        // --- Change the return statement to set_local to store function result
        if (mainReturnIndex >= 0) {
            this._builder.body[mainReturnIndex] = FunctionBuilder_1.localSet(`$${this.func.funcId}$res`);
            this.func.hasReturn = true;
        }
        // --- Transform all local names to include the function identifier
        this._builder.locals.forEach((l) => {
            l.id = `$${this.func.funcId}${l.id}`;
        });
        wat_helpers_1.visitInstructions(this._builder.body, (ins) => {
            if (ins.type === "LocalGet" ||
                ins.type === "LocalSet" ||
                ins.type === "LocalTee") {
                ins.id = `$${this.func.funcId}${ins.id}`;
            }
        });
    }
    // ==========================================================================
    // Helpers for locals
    /**
     * Creates a temporary local with the specified type
     * @param type
     */
    createTempLocal(type, prefix = "") {
        const underlying = wa_nodes_1.WaType[type].toString();
        const tmpName = `$tloc$${prefix}${underlying}`;
        if (!this._tempLocals.has(type)) {
            const local = this._builder.addLocal(tmpName, type);
            this.locals.set(tmpName, {
                fromParameter: false,
                name: tmpName,
                waType: type,
                type: {
                    type: "Intrinsic",
                    underlying,
                },
            });
            this._tempLocals.add(type);
            this.addTrace(() => [
                "local",
                0,
                this.wsCompiler.waTree.renderLocal(local),
            ]);
        }
        return tmpName;
    }
    /**
     * Creates a compile-time local with the specified name and type
     * @param name Local name
     * @param spec Local type specification
     */
    createCompileTimeLocal(name, spec) {
        // --- Avoid declaring the same local
        if (this.locals.has(name)) {
            return;
        }
        // --- Add the new local
        const localWaType = spec.type === "Pointer"
            ? wa_nodes_1.WaType.i32
            : WatSharpCompiler_1.waTypeMappings[spec.underlying];
        this._builder.addLocal(name, localWaType);
        this.locals.set(name, {
            fromParameter: false,
            name,
            waType: localWaType,
            type: {
                type: "Intrinsic",
                underlying: wa_nodes_1.WaType[localWaType].toString(),
            },
        });
    }
    // ==========================================================================
    // Helpers
    /**
     * Injects the specifiec WebAssembly instructions into the function
     * @param emit Should emit code?
     * @param instr Instructions to inject
     * @param body Body to inject the instructions
     */
    inject(instr, body) {
        if (this._suspendCount > 0) {
            return;
        }
        if (Array.isArray(instr)) {
            body.push(...instr);
        }
        else {
            body.push(instr);
        }
    }
    /**
     * Reports the specified error
     * @param errorCode Error code
     * @param token Token that represents the error's position
     * @param options Error message options
     */
    reportError(errorCode, node, ...options) {
        this.wsCompiler.reportError(errorCode, node, ...options);
    }
}
exports.FunctionCompiler = FunctionCompiler;
/**
 * Tests if the expression in a commutative binary operation
 * @param expr Expression to test
 * @returns Binary operation
 */
function isCommutativeOp(expr) {
    if (expr.type !== "BinaryExpression") {
        return null;
    }
    switch (expr.operator) {
        case "!=":
        case "==":
        case "&":
        case "*":
        case "+":
        case "^":
        case "|":
            return expr;
        default:
            return null;
    }
}
/**
 * Creates a literal value
 * @param value Value to wrap into a literal
 * @param asU64 Wrap a 64-bit value as u64?
 */
function createLiteral(value, asU64 = false) {
    return {
        type: "Literal",
        value: typeof value === "bigint"
            ? asU64
                ? BigInt.asUintN(64, value)
                : BigInt.asIntN(64, value)
            : value,
        literalSource: typeof value === "number"
            ? Number.isInteger(value)
                ? source_tree_1.LiteralSource.Int
                : source_tree_1.LiteralSource.Real
            : source_tree_1.LiteralSource.BigInt,
    };
}
// --- Intrinsic type instances
const voidDesc = {
    type: "Void",
};
const i32Desc = {
    type: "Intrinsic",
    underlying: "i32",
};
const i64Desc = {
    type: "Intrinsic",
    underlying: "i64",
};
const f32Desc = {
    type: "Intrinsic",
    underlying: "f32",
};
const f64Desc = {
    type: "Intrinsic",
    underlying: "f64",
};


/***/ }),

/***/ "./src/compiler/WatSharpCompiler.ts":
/*!******************************************!*\
  !*** ./src/compiler/WatSharpCompiler.ts ***!
  \******************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.createBreakLabel = exports.createLoopLabel = exports.createTableName = exports.createParameterName = exports.createLocalName = exports.createGlobalName = exports.mapFunctionParameterType = exports.bitwiseNotMasks = exports.waTypeMappings = exports.defaultCompilerOptions = exports.WatSharpCompiler = void 0;
const errors_1 = __webpack_require__(/*! ../core/errors */ "./src/core/errors.ts");
const source_tree_1 = __webpack_require__(/*! ./source-tree */ "./src/compiler/source-tree.ts");
const WatSharpParser_1 = __webpack_require__(/*! ./WatSharpParser */ "./src/compiler/WatSharpParser.ts");
const expression_resolver_1 = __webpack_require__(/*! ./expression-resolver */ "./src/compiler/expression-resolver.ts");
const FunctionCompiler_1 = __webpack_require__(/*! ./FunctionCompiler */ "./src/compiler/FunctionCompiler.ts");
const WaTree_1 = __webpack_require__(/*! ../wa-ast/WaTree */ "./src/wa-ast/WaTree.ts");
const wa_nodes_1 = __webpack_require__(/*! ../wa-ast/wa-nodes */ "./src/wa-ast/wa-nodes.ts");
/**
 * This class implements the WAT# compiler
 */
class WatSharpCompiler {
    constructor(source, includeHandler, preprocessorSymbols, options) {
        this.source = source;
        this.includeHandler = includeHandler;
        this.preprocessorSymbols = preprocessorSymbols;
        this.options = options;
        // --- Keep track of error messages
        this._errors = [];
        // --- Stores the WebAssemble tree
        this._waTree = null;
        // --- Trace helper members
        this._shouldTrace = false;
        this._traceMessages = [];
        // --- The number of table entries to emit
        this._tableEntries = 0;
        // --- Bodies of the compiled functions
        this._functionBodies = new Map();
        this._parser = new WatSharpParser_1.WatSharpParser(source, includeHandler, preprocessorSymbols);
        if (!options) {
            this.options = exports.defaultCompilerOptions;
        }
    }
    /**
     * Turn on comiler tracing
     */
    trace() {
        this._shouldTrace = true;
    }
    /**
     * Execute the compilation of the source
     */
    compile() {
        // --- Step #1: parse the program and sign any syntax issues.
        this._parser.parseProgram();
        this._errors = this._parser.errors.slice(0);
        if (this._errors.length > 0) {
            return null;
        }
        // --- Step #2: resolve dependencies among declarations
        this.resolveDependencies();
        if (this._errors.length > 0) {
            return null;
        }
        this.flattenTypes();
        // --- Step #3: emit declaration nodes
        this._waTree = new WaTree_1.WaTree();
        this.emitDeclarationNodes();
        // --- Step #4: process function bodies
        this.processFunctionBodies();
        if (this._errors.length > 0) {
            return null;
        }
        // --- Step #5: complete code emission
        this.completeCodeEmission();
        // --- Step #6: render the tree
        return this._waTree.render();
    }
    /**
     * The errors raised during the parse phase
     */
    get errors() {
        return this._errors;
    }
    /**
     * Indicates if there were any errors during the parse phase
     */
    get hasErrors() {
        return this._errors.length > 0;
    }
    /**
     * Gets the declarations of the WAT# program
     */
    get declarations() {
        return this._parser.declarations;
    }
    /**
     * Gets the WebAssembly tree with the emitted code
     */
    get waTree() {
        return this._waTree;
    }
    /**
     * Gets compiler trace messages
     */
    get traceMessages() {
        return this._traceMessages;
    }
    /**
     * Gets the instructions for the specified function
     * @param name Function name
     */
    getFunctionBodyInstructions(name) {
        return this._functionBodies.get(name).builder.body;
    }
    /**
     * Gets the locals of the specified function
     * @param name Function name
     */
    getFunctionLocals(name) {
        return this._functionBodies.get(name).builder.locals;
    }
    /**
     * Adds a trace message
     * @param traceFactory Factory function to generate trace message
     */
    addTrace(traceFactory) {
        if (this._shouldTrace) {
            const trace = traceFactory();
            this._traceMessages.push({
                source: trace[0],
                point: trace[1],
                message: trace[2],
            });
        }
    }
    // ==========================================================================
    // Declaration resolution
    /**
     * Resolve unresolved declaration dependencies
     */
    resolveDependencies(spec) {
        const compiler = this;
        // --- Declarations to resolve
        const declarations = this._parser.declarations;
        // --- Queue to hold dependent declarations to resolve
        const resolutionQueue = [];
        if (spec) {
            resolutionQueue.push({ typeSpec: spec });
        }
        else {
            const items = [...declarations.values()].map((item) => ({ decl: item }));
            resolutionQueue.push(...items);
        }
        // --- Stack of name to check circular references
        const namedStack = [];
        // --- Next address in memory
        let nextMemoryAddress = 0;
        // --- Iterate through the queue of items to resolve
        while (resolutionQueue.length > 0) {
            const item = resolutionQueue.shift();
            let itemToReport;
            // --- Resolve a declaration, type, or expression
            try {
                if (item.decl) {
                    itemToReport = item.decl;
                    resolveDeclarationDependency(item.decl);
                }
                else if (item.typeSpec) {
                    itemToReport = item.typeSpec;
                    resolveTypeSpecification(item.typeSpec);
                }
                else if (item.expr) {
                    itemToReport = item.expr;
                    resolveExpression(item.expr);
                }
            }
            catch (err) {
                this.reportError("W107", itemToReport, err.toString());
            }
        }
        /**
         * Resolve a single declaration
         * @param decl Declaration instance
         */
        function resolveDeclarationDependency(decl) {
            // --- Declaration has already been resolved
            if (decl.resolved) {
                return;
            }
            // --- Push the name to the stack
            namedStack.push(decl.name);
            // --- Resolve a particular type of declaration
            switch (decl.type) {
                case "ConstDeclaration":
                    resolveExpression(decl.expr);
                    if (decl.expr.value !== undefined) {
                        decl.value = expression_resolver_1.applyTypeCast(decl.underlyingType, decl.expr.value);
                    }
                    break;
                case "DataDeclaration":
                    decl.exprs.forEach((expr) => resolveExpression(expr));
                    decl.spec = {
                        type: "Array",
                        spec: {
                            type: "Intrinsic",
                            underlying: decl.underlyingType,
                            sizeof: source_tree_1.instrisicSizes[decl.underlyingType],
                        },
                        size: {
                            type: "Literal",
                            value: decl.exprs.length,
                            source: source_tree_1.LiteralSource.Int,
                        },
                    };
                    decl.address = nextMemoryAddress;
                    nextMemoryAddress =
                        decl.address +
                            decl.exprs.length * source_tree_1.instrisicSizes[decl.underlyingType];
                    break;
                case "GlobalDeclaration":
                    if (decl.initExpr) {
                        resolveExpression(decl.initExpr);
                    }
                    break;
                case "TableDeclaration":
                    resolveTableDeclaration(decl);
                    break;
                case "TypeDeclaration":
                    resolveTypeSpecification(decl.spec);
                    break;
                case "VariableDeclaration":
                    resolveTypeSpecification(decl.spec);
                    if (decl.addressAlias) {
                        const aliasDecl = declarations.get(decl.addressAlias.name);
                        if (!aliasDecl) {
                            compiler.reportError("W101", decl.addressAlias);
                            break;
                        }
                        if (aliasDecl.type !== "VariableDeclaration") {
                            compiler.reportError("W162", decl.addressAlias);
                            break;
                        }
                        if (aliasDecl.address === undefined) {
                            compiler.reportError("W163", decl.addressAlias);
                            break;
                        }
                        decl.address = aliasDecl.address;
                    }
                    if (decl.address === undefined) {
                        decl.address = nextMemoryAddress;
                    }
                    nextMemoryAddress = decl.address + compiler.getSizeof(decl.spec);
                    break;
            }
            // --- Done. Remove the name fromt he stack and complete the resolution
            namedStack.pop();
            decl.resolved = true;
            return;
        }
        /**
         * Resolve the given type specification
         * @param typeSpec Type specification to resolve
         */
        function resolveTypeSpecification(spec) {
            var _a;
            // --- Type has already been resolved
            if (spec.resolved) {
                return;
            }
            // --- Let's make the type resolved
            spec.resolved = true;
            switch (spec.type) {
                case "Intrinsic":
                    spec.sizeof = source_tree_1.instrisicSizes[spec.underlying];
                    break;
                case "Pointer":
                    spec.sizeof = 4;
                    resolutionQueue.push({
                        typeSpec: spec.spec,
                    });
                    break;
                case "Struct":
                    spec.sizeof = 0;
                    for (let i = 0; i < spec.fields.length; i++) {
                        spec.fields[i].offset = spec.sizeof;
                        resolveTypeSpecification(spec.fields[i].spec);
                        spec.sizeof += compiler.getSizeof(spec.fields[i].spec);
                    }
                    break;
                case "Array":
                    resolveTypeSpecification(spec.spec);
                    resolveExpression(spec.size);
                    const arraySize = Number((_a = spec.size.value) !== null && _a !== void 0 ? _a : 0);
                    spec.sizeof = arraySize * compiler.getSizeof(spec.spec);
                    break;
                case "NamedType":
                    // --- Check circular references
                    if (namedStack.includes(spec.name)) {
                        compiler.reportError("W103", spec, spec.name);
                        break;
                    }
                    const decl = declarations.get(spec.name);
                    if (decl) {
                        if (decl.type === "TypeDeclaration") {
                            // --- Use a stack to detect circular references
                            namedStack.push(spec.name);
                            resolveDeclarationDependency(decl);
                            namedStack.pop();
                            break;
                        }
                        else {
                            // --- Name is not a type declaration
                            compiler.reportError("W102", spec, spec.name);
                            return null;
                        }
                    }
                    else {
                        // --- Unknown declaration
                        compiler.reportError("W101", spec, spec.name);
                        return null;
                    }
                default:
                    spec.resolved = false;
                    return null;
            }
        }
        /**
         * Resolves the specified expression
         * @param expr Expression to resolve
         */
        function resolveExpression(expr) {
            // --- "Outsource" to resolveConstantExpression
            expression_resolver_1.resolveConstantExpression(expr, 
            // --- This function resolves identifiers in constant expressions
            (id) => {
                // --- Check for circular reference
                if (namedStack.includes(id.name)) {
                    compiler.reportError("W103", id, id.name);
                    return;
                }
                // --- Resolve the const declaration
                const decl = declarations.get(id.name);
                if (decl) {
                    if (decl.type === "ConstDeclaration") {
                        // --- Use a stack to detect circular references
                        namedStack.push(id.name);
                        resolveDeclarationDependency(decl);
                        id.value = decl.value;
                        namedStack.pop();
                    }
                    else {
                        // --- This is not a const declaration
                        compiler.reportError("W108", id, id.name);
                    }
                }
                else {
                    // --- Unknown declaration
                    compiler.reportError("W101", id, id.name);
                }
            }, 
            // --- This function resolves the "sizeof" operator
            (typeSpec) => {
                resolveTypeSpecification(typeSpec);
                return typeSpec.resolved ? compiler.getSizeof(typeSpec) : 0;
            }, 
            // --- Use this function to report errors
            compiler.reportError.bind(compiler));
        }
        /**
         * Resolves the specified table declaration
         * @param table Table declaration
         */
        function resolveTableDeclaration(table) {
            // --- Resolve the result type
            if (table.resultType) {
                resolveTypeSpecification(table.resultType);
            }
            // --- Resolve parameter types
            table.params.forEach((p) => resolveTypeSpecification(p.spec));
            // --- Iterate through all table idetifiers
            table.entryIndex = compiler._tableEntries;
            table.ids.forEach((id) => {
                const decl = declarations.get(id);
                if (!decl) {
                    // --- Unknown identifier
                    compiler.reportError("W101", table, id);
                    return;
                }
                if (decl.type !== "FunctionDeclaration") {
                    // --- This ID is not a function
                    compiler.reportError("W109", table, id);
                    return;
                }
                if (!matchSignature(decl, table)) {
                    return;
                }
                // --- Sign the usage of this function
                decl.invocationCount++;
                compiler._tableEntries++;
            });
            table.resolved = true;
        }
        /**
         * Matches the specified function signature with the provided parameters and result type
         * @param func Function signature to check
         * @param resultType Result type specification
         * @param params Function parameters
         */
        function matchSignature(func, table) {
            // --- Void versus non-void result type
            if ((!func.resultType && table.resultType) ||
                (!table.resultType && func.resultType)) {
                compiler.reportError("W158", table, func.name);
                return false;
            }
            // --- Compare non-void result types
            if (func.resultType && table.resultType) {
                if (!matchFunctionParamaterTypes(func.resultType, table.resultType)) {
                    compiler.reportError("W158", table, func.name);
                    return false;
                }
            }
            // --- Compare parameters
            if (func.params.length !== table.params.length) {
                compiler.reportError("W158", table, func.name);
                return false;
            }
            func.params.forEach((fp, index) => {
                if (!matchFunctionParamaterTypes(fp.spec, table.params[index].spec)) {
                    compiler.reportError("W158", table, func.name);
                    return false;
                }
            });
            return true;
        }
        /**
         * Tests if two function parameter types match or not
         * @param type1
         * @param type2
         */
        function matchFunctionParamaterTypes(type1, type2) {
            if (type1.type === "Intrinsic") {
                // --- Match intrinsic types
                if (type2.type !== "Intrinsic" ||
                    type1.underlying !== type2.underlying) {
                    return false;
                }
            }
            else if (type1.type === "Pointer" && type2.type !== "Pointer") {
                return false;
            }
            return true;
        }
    }
    /**
     * Flattens type declarations by replacing named types specifications with direct
     * ones
     */
    flattenTypes() {
        const compiler = this;
        for (const decl of this._parser.declarations.values()) {
            switch (decl.type) {
                case "TypeDeclaration":
                case "VariableDeclaration":
                    decl.spec = flattenType(decl.spec);
                    break;
            }
        }
        function flattenType(spec) {
            if (spec.flattened) {
                return spec;
            }
            spec.flattened = true;
            switch (spec.type) {
                case "Array":
                case "Pointer":
                    spec.spec = flattenType(spec.spec);
                    break;
                case "Struct":
                    spec.fields.forEach((fi) => {
                        fi.spec = flattenType(fi.spec);
                    });
                    break;
                case "NamedType":
                    const typeDecl = compiler._parser.declarations.get(spec.name);
                    if (typeDecl.type !== "TypeDeclaration") {
                        break;
                    }
                    return flattenType(typeDecl.spec);
            }
            return spec;
        }
    }
    /**
     * Processes table declarations
     */
    emitTables() {
        let header = false;
        let itemCount = 0;
        for (const table of this._parser.declarations.values()) {
            if (table.type !== "TableDeclaration") {
                continue;
            }
            this.waTree.separatorLine();
            if (!header) {
                this.waTree.comment("Table definitions");
                header = true;
            }
            itemCount += table.ids.length;
            const resultType = mapFunctionParameterType(table.resultType);
            const params = table.params.map((p) => ({
                id: createParameterName(p.name),
                type: mapFunctionParameterType(p.spec),
            }));
            this.waTree.typeDef(createTableName(table.name), params, resultType);
            this.waTree.element(table.entryIndex, table.ids.map((id) => createGlobalName(id)));
        }
        this.waTree.setTable(itemCount, "$tables$");
    }
    /**
     * Get the size of a type specification
     * @param typeSpec
     */
    getSizeof(typeSpec) {
        if (typeSpec.type !== "NamedType") {
            return typeSpec.sizeof;
        }
        const decl = this._parser.declarations.get(typeSpec.name);
        if (decl.type !== "TypeDeclaration" || !decl.spec.sizeof) {
            return 0;
        }
        return decl.spec.sizeof;
    }
    // ==========================================================================
    // Emit declaration nodes
    /**
     * Emit the code for module declarations
     */
    emitDeclarationNodes() {
        this.emitImports();
        this.emitExports();
        this.emitGlobals();
        this.emitVariableMap();
        this.emitData();
        this.emitTables();
    }
    /**
     * Emit module exports code
     */
    emitExports() {
        let header = false;
        for (const decl of this.declarations.values()) {
            if (decl.type === "FunctionDeclaration" && decl.isExport) {
                if (!header) {
                    this._waTree.separatorLine();
                    this._waTree.comment("Exported functions");
                    header = true;
                }
                this._waTree.functionExport(createGlobalName(decl.name), decl.name);
            }
        }
    }
    /**
     * Emit module exports code
     */
    emitImports() {
        for (const decl of this.declarations.values()) {
            if (decl.type === "ImportedFunctionDeclaration") {
                this._waTree.functionImport(createGlobalName(decl.name), decl.name1, decl.name2, decl.parSpecs.map((ps) => ({
                    type: exports.waTypeMappings[ps.underlying],
                })));
            }
        }
    }
    /**
     * Emit globals code
     */
    emitGlobals() {
        var _a;
        let header = false;
        for (const decl of this.declarations.values()) {
            if (decl.type === "GlobalDeclaration") {
                if (!header) {
                    this._waTree.separatorLine();
                    this._waTree.comment("Global declarations");
                    header = true;
                }
                this._waTree.global(createGlobalName(decl.name), exports.waTypeMappings[decl.underlyingType], (_a = decl === null || decl === void 0 ? void 0 : decl.initExpr) === null || _a === void 0 ? void 0 : _a.value);
            }
        }
    }
    /**
     * Emit globals code
     */
    emitVariableMap() {
        this._waTree.separatorLine();
        this._waTree.comment("Variable mappings");
        for (const decl of this.declarations.values()) {
            if (decl.type === "VariableDeclaration") {
                this._waTree.comment(`0x${decl.address
                    .toString(16)
                    .padStart(8, "0")} (${decl.address
                    .toString(10)
                    .padStart(10, " ")}) [${this.getSizeof(decl.spec)
                    .toString(10)
                    .padStart(10, " ")}]: ${decl.name}`);
            }
        }
    }
    emitData() {
        this._waTree.separatorLine();
        this._waTree.comment("Data declarations");
        for (const decl of this.declarations.values()) {
            if (decl.type === "DataDeclaration") {
                const bytes = [];
                const dataType = decl.spec.spec
                    .underlying;
                const size = source_tree_1.instrisicSizes[dataType];
                for (const expr of decl.exprs) {
                    let value = BigInt(expr.value);
                    for (let i = 0; i < size; i++) {
                        bytes.push(Number(value & BigInt(0xff)));
                        value >>= BigInt(8);
                    }
                }
                this._waTree.data(decl.address, bytes);
            }
        }
    }
    // ==========================================================================
    // Process function bodies
    /**
     * Processes the bodies of functions
     */
    processFunctionBodies() {
        // --- Filter for function declarations
        const functionDeclarations = [];
        for (const decl of this.declarations.values()) {
            if (decl.type === "FunctionDeclaration") {
                functionDeclarations.push(decl);
            }
        }
        // --- Compile inline functions first
        functionDeclarations
            .filter((fd) => fd.isInline)
            .forEach((f) => this.processFunctionBody(f));
        // --- Carry on with no-inline functions
        functionDeclarations
            .filter((fd) => !fd.isInline)
            .forEach((f) => this.processFunctionBody(f));
    }
    /**
     * Processes the specified function
     * @param func Function to process
     */
    processFunctionBody(func) {
        const fCompiler = new FunctionCompiler_1.FunctionCompiler(this, func);
        fCompiler.process();
        if (func.isInline) {
            fCompiler.prepareForInlining();
        }
        this._functionBodies.set(func.name, {
            builder: fCompiler.builder,
            func,
        });
    }
    // ==========================================================================
    // Complete code emission
    completeCodeEmission() {
        // --- Add used function declarations
        for (const funcBody of this._functionBodies.values()) {
            if (funcBody.func.isExport ||
                (!funcBody.func.canBeInlined && funcBody.func.invocationCount > 0)) {
                this.waTree.separatorLine();
                this.waTree.addFunc(funcBody.builder);
            }
        }
        // --- Calculate memory size
        let maxAddr = 0;
        for (const decl of this.declarations.values()) {
            if (decl.type !== "VariableDeclaration") {
                continue;
            }
            const thisEndAddr = decl.address + this.getSizeof(decl.spec);
            if (isNaN(thisEndAddr)) {
                throw new Error(`Unexpected memory size of variable ${decl.name}`);
            }
            if (thisEndAddr > maxAddr) {
                maxAddr = thisEndAddr;
            }
        }
        const memPages = (maxAddr >> 16) + 1;
        this._waTree.setMemorySpecification(memPages);
    }
    // ==========================================================================
    // Error reporting
    /**
     * Reports the specified error
     * @param errorCode Error code
     * @param token Token that represents the error's position
     * @param options Error message options
     */
    reportError(errorCode, node, ...options) {
        var _a;
        let errorText = (_a = errors_1.errorMessages[errorCode]) !== null && _a !== void 0 ? _a : "Unkonwn error";
        if (options) {
            options.forEach((_o, idx) => (errorText = replace(errorText, `{${idx}}`, options[idx].toString())));
        }
        this._errors.push({
            code: errorCode,
            text: errorText,
            line: node.startLine,
            column: node.startColumn,
            position: node.startPosition,
        });
        function replace(input, placeholder, replacement) {
            do {
                input = input.replace(placeholder, replacement);
            } while (input.includes(placeholder));
            return input;
        }
    }
}
exports.WatSharpCompiler = WatSharpCompiler;
/**
 * Default compiler options
 */
exports.defaultCompilerOptions = {
    generateComments: true,
};
/**
 * Mappings from intrinsic types to WA types
 */
exports.waTypeMappings = {
    bool: wa_nodes_1.WaType.i32,
    i8: wa_nodes_1.WaType.i32,
    u8: wa_nodes_1.WaType.i32,
    i16: wa_nodes_1.WaType.i32,
    u16: wa_nodes_1.WaType.i32,
    i32: wa_nodes_1.WaType.i32,
    u32: wa_nodes_1.WaType.i32,
    i64: wa_nodes_1.WaType.i64,
    u64: wa_nodes_1.WaType.i64,
    f32: wa_nodes_1.WaType.f32,
    f64: wa_nodes_1.WaType.f64,
};
/**
 * Mask constants for the bitwise NOT operation
 */
exports.bitwiseNotMasks = {
    bool: 0xff,
    i8: 0xff,
    u8: 0xff,
    i16: 0xffff,
    u16: 0xffff,
    i32: 4294967295,
    u32: 4294967295,
    i64: BigInt("0xffffffffffffffff"),
    u64: BigInt("0xffffffffffffffff"),
    f32: 0,
    f64: 0,
};
/**
 * Maps a function parameter type to an intrinsic type
 * @param spec Type specification
 */
function mapFunctionParameterType(spec) {
    if (spec) {
        if (spec.type === "Pointer") {
            return wa_nodes_1.WaType.i32;
        }
        if (spec.type === "Intrinsic") {
            return exports.waTypeMappings[spec.underlying];
        }
    }
    return null;
}
exports.mapFunctionParameterType = mapFunctionParameterType;
/**
 * Creates a WebAssembly function name
 * @param name Name to use as input
 */
function createGlobalName(name) {
    return `$${name}`;
}
exports.createGlobalName = createGlobalName;
/**
 * Creates a WebAssembly local name
 * @param name Name to use as input
 */
function createLocalName(name) {
    return `$loc$${name}`;
}
exports.createLocalName = createLocalName;
/**
 * Creates a WebAssembly parameter name
 * @param name Name to use as input
 */
function createParameterName(name) {
    return `$par$${name}`;
}
exports.createParameterName = createParameterName;
/**
 * Creates a WebAssembly function name
 * @param name Name to use as input
 */
function createTableName(name) {
    return `$tbl$${name}`;
}
exports.createTableName = createTableName;
/**
 * Creates a WebAssembly loop label
 * @param depth Loop depth
 */
function createLoopLabel(depth) {
    return `$loop$${depth}`;
}
exports.createLoopLabel = createLoopLabel;
/**
 * Creates a WebAssembly break label
 * @param depth Loop depth
 */
function createBreakLabel(depth) {
    return `$break$${depth}`;
}
exports.createBreakLabel = createBreakLabel;


/***/ }),

/***/ "./src/compiler/WatSharpLexer.ts":
/*!***************************************!*\
  !*** ./src/compiler/WatSharpLexer.ts ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.WatSharpLexer = void 0;
const tokens_1 = __webpack_require__(/*! ../core/tokens */ "./src/core/tokens.ts");
/**
 * This class implements the lexer of WAT#
 */
class WatSharpLexer {
    /**
     * Initializes the tokenizer with the input stream
     * @param input Input source code stream
     */
    constructor(input) {
        this.input = input;
        // --- Already fetched tokens
        this._ahead = [];
        // --- Prefetched character (from the next token)
        this._prefetched = null;
        // --- Prefetched character position (from the next token)
        this._prefetchedPos = null;
        // --- Prefetched character column (from the next token)
        this._prefetchedColumn = null;
        // --- The last end-of-line comment
        this._lastComment = null;
    }
    /**
     * Resets the last comment
     */
    resetComment() {
        this._lastComment = null;
    }
    /**
     * Gets the last end-of-line comment
     */
    get lastComment() {
        return this._lastComment;
    }
    /**
     * Fethches the next token without advancing to its position
     * @param ws If true, retrieve whitespaces too
     */
    peek(ws = false) {
        return this.ahead(0, ws);
    }
    /**
     *
     * @param n Number of token positions to read ahead
     * @param ws If true, retrieve whitespaces too
     */
    ahead(n = 1, ws = false) {
        if (n > 16) {
            throw new Error("Cannot look ahead more than 16 tokens");
        }
        // --- Prefetch missing tokens
        while (this._ahead.length <= n) {
            const token = this.fetch();
            if (token.type === tokens_1.TokenType.EolComment) {
                this._lastComment = token.text;
            }
            if (tokens_1.isEof(token)) {
                return token;
            }
            if (ws || (!ws && !tokens_1.isWs(token))) {
                this._ahead.push(token);
            }
        }
        return this._ahead[n];
    }
    /**
     * Fethces the nex token and advances the stream position
     * @param ws If true, retrieve whitespaces too
     */
    get(ws = false) {
        if (this._ahead.length > 0) {
            const token = this._ahead.shift();
            if (!token) {
                throw new Error("Token expected");
            }
            return token;
        }
        while (true) {
            const token = this.fetch();
            if (token.type === tokens_1.TokenType.EolComment) {
                this._lastComment = token.text;
            }
            if (tokens_1.isEof(token) || ws || (!ws && !tokens_1.isWs(token))) {
                return token;
            }
        }
    }
    /**
     * Fetches the next token from the input stream
     */
    fetch() {
        const lexer = this;
        const input = this.input;
        const startPos = this._prefetchedPos || input.position;
        const line = input.line;
        const startColumn = this._prefetchedColumn || input.column;
        let text = "";
        let tokenType = tokens_1.TokenType.Eof;
        let lastEndPos = input.position;
        let lastEndColumn = input.column;
        let ch = null;
        let useResolver = false;
        let phase = LexerPhase.Start;
        while (true) {
            // --- Get the next character
            ch = fetchNextChar();
            // --- In case of EOF, return the current token data
            if (ch === null) {
                return makeToken();
            }
            // --- Set the intial token type to unknown for the other characters
            if (tokenType === tokens_1.TokenType.Eof) {
                tokenType = tokens_1.TokenType.Unknown;
            }
            // --- Follow the lexer state machine
            switch (phase) {
                // ====================================================================
                // Process the first character
                case LexerPhase.Start:
                    switch (ch) {
                        // --- Go on with whitespaces
                        case " ":
                        case "\t":
                        case "\n":
                        case "\r":
                            phase = LexerPhase.InWhiteSpace;
                            tokenType = tokens_1.TokenType.Ws;
                            break;
                        // --- Divide DivideAsgn, BlockComment, or EolComment
                        case "/":
                            phase = LexerPhase.Slash;
                            tokenType = tokens_1.TokenType.Divide;
                            break;
                        // --- Asterisk, MultiplyAsgn
                        case "*":
                            phase = LexerPhase.Asterisk;
                            tokenType = tokens_1.TokenType.Asterisk;
                            break;
                        case "%":
                            phase = LexerPhase.Remainder;
                            tokenType = tokens_1.TokenType.Remainder;
                            break;
                        // --- Plus, AddAsgn
                        case "+":
                            phase = LexerPhase.Plus;
                            tokenType = tokens_1.TokenType.Plus;
                            break;
                        // --- Minus, SubtractAsgn
                        case "-":
                            phase = LexerPhase.Minus;
                            tokenType = tokens_1.TokenType.Minus;
                            break;
                        // --- Xor, XorAsgn
                        case "^":
                            phase = LexerPhase.Xor;
                            tokenType = tokens_1.TokenType.Xor;
                            break;
                        // --- Or, OrAsgn
                        case "|":
                            phase = LexerPhase.Or;
                            tokenType = tokens_1.TokenType.Or;
                            break;
                        // --- Ampersand, AndAsgn
                        case "&":
                            phase = LexerPhase.Ampersand;
                            tokenType = tokens_1.TokenType.Ampersand;
                            break;
                        case ";":
                            return completeToken(tokens_1.TokenType.Semicolon);
                        case ",":
                            return completeToken(tokens_1.TokenType.Comma);
                        case "(":
                            return completeToken(tokens_1.TokenType.LParent);
                        case ")":
                            return completeToken(tokens_1.TokenType.RParent);
                        case ":":
                            phase = LexerPhase.Colon;
                            tokenType = tokens_1.TokenType.Colon;
                            break;
                        case "[":
                            return completeToken(tokens_1.TokenType.LSquare);
                        case "]":
                            return completeToken(tokens_1.TokenType.RSquare);
                        case "?":
                            return completeToken(tokens_1.TokenType.QuestionMark);
                        case "~":
                            return completeToken(tokens_1.TokenType.BinaryNot);
                        case "{":
                            return completeToken(tokens_1.TokenType.LBrace);
                        case "}":
                            return completeToken(tokens_1.TokenType.RBrace);
                        // --- Asgn, Equal
                        case "=":
                            phase = LexerPhase.Equal;
                            tokenType = tokens_1.TokenType.Asgn;
                            break;
                        // --- Not, NotEqual
                        case "!":
                            phase = LexerPhase.Exclamation;
                            tokenType = tokens_1.TokenType.Not;
                            break;
                        // --- LessThan, LessThanOrEqual, ShiftLeft
                        case "<":
                            phase = LexerPhase.AngleLeft;
                            tokenType = tokens_1.TokenType.LessThan;
                            break;
                        // --- GreaterThan, GreaterThanOrEqual, ShiftRight, SignedShiftRight
                        case ">":
                            phase = LexerPhase.AngleRight;
                            tokenType = tokens_1.TokenType.GreaterThan;
                            break;
                        case "0":
                            phase = LexerPhase.Zero;
                            tokenType = tokens_1.TokenType.DecimalLiteral;
                            break;
                        case ".":
                            phase = LexerPhase.Dot;
                            tokenType = tokens_1.TokenType.Dot;
                            break;
                        case '"':
                            phase = LexerPhase.String;
                            break;
                        default:
                            if (tokens_1.isIdStart(ch)) {
                                useResolver = true;
                                phase = LexerPhase.IdTail;
                                tokenType = tokens_1.TokenType.Identifier;
                            }
                            else if (tokens_1.isDecimalDigit(ch)) {
                                phase = LexerPhase.DecimalOrReal;
                                tokenType = tokens_1.TokenType.DecimalLiteral;
                            }
                            else {
                                completeToken(tokens_1.TokenType.Unknown);
                            }
                            break;
                    }
                    break;
                // ====================================================================
                // Process comments
                // --- Looking for the end of whitespace
                case LexerPhase.InWhiteSpace:
                    if (ch !== " " && ch !== "\t" && ch !== "\r" && ch !== "\n") {
                        return makeToken();
                    }
                    break;
                // --- Wait for an "*" that may complete a block comment
                case LexerPhase.BlockCommentTrail1:
                    if (ch === "*") {
                        phase = LexerPhase.BlockCommentTrail2;
                    }
                    break;
                // --- Wait for a "/" that may complete a block comment
                case LexerPhase.BlockCommentTrail2:
                    if (ch === "/") {
                        return completeToken(tokens_1.TokenType.BlockComment);
                    }
                    break;
                case LexerPhase.InlineCommentTrail:
                    if (ch === "\n") {
                        return completeToken();
                    }
                    break;
                // ====================================================================
                // Process identifiers
                case LexerPhase.IdTail:
                    if (!tokens_1.isIdContinuation(ch)) {
                        return makeToken();
                    }
                    break;
                // ====================================================================
                // Process miscellenous tokens
                case LexerPhase.Slash:
                    if (ch === "=") {
                        return completeToken(tokens_1.TokenType.DivideAsgn);
                    }
                    if (ch === "*") {
                        phase = LexerPhase.BlockCommentTrail1;
                    }
                    else if (ch === "/") {
                        phase = LexerPhase.InlineCommentTrail;
                        tokenType = tokens_1.TokenType.EolComment;
                    }
                    else {
                        return makeToken();
                    }
                    break;
                case LexerPhase.Asterisk:
                    return ch === "="
                        ? completeToken(tokens_1.TokenType.MultiplyAsgn)
                        : makeToken();
                case LexerPhase.Remainder:
                    return ch === "="
                        ? completeToken(tokens_1.TokenType.RemainderAsgn)
                        : makeToken();
                case LexerPhase.Plus:
                    return ch === "=" ? completeToken(tokens_1.TokenType.AddAsgn) : makeToken();
                case LexerPhase.Minus:
                    return ch === "="
                        ? completeToken(tokens_1.TokenType.SubtractAsgn)
                        : makeToken();
                case LexerPhase.Xor:
                    return ch === "=" ? completeToken(tokens_1.TokenType.XorAsgn) : makeToken();
                case LexerPhase.Or:
                    return ch === "=" ? completeToken(tokens_1.TokenType.OrAsgn) : makeToken();
                case LexerPhase.Ampersand:
                    return ch === "=" ? completeToken(tokens_1.TokenType.AndAsgn) : makeToken();
                case LexerPhase.Colon:
                    return ch === "=" ? completeToken(tokens_1.TokenType.CopyAsgn) : makeToken();
                case LexerPhase.Equal:
                    return ch === "=" ? completeToken(tokens_1.TokenType.Equal) : makeToken();
                case LexerPhase.Exclamation:
                    return ch === "=" ? completeToken(tokens_1.TokenType.NotEqual) : makeToken();
                case LexerPhase.AngleLeft:
                    if (ch === "=") {
                        return completeToken(tokens_1.TokenType.LessThanOrEqual);
                    }
                    if (ch === "<") {
                        phase = LexerPhase.ShiftLeft;
                        tokenType = tokens_1.TokenType.ShiftLeft;
                        break;
                    }
                    return makeToken();
                case LexerPhase.AngleRight:
                    if (ch === "=") {
                        return completeToken(tokens_1.TokenType.GreaterThanOrEqual);
                    }
                    if (ch !== ">") {
                        return makeToken();
                    }
                    phase = LexerPhase.SignedShiftRight;
                    tokenType = tokens_1.TokenType.SignedShiftRight;
                    break;
                case LexerPhase.ShiftLeft:
                    return ch === "="
                        ? completeToken(tokens_1.TokenType.ShiftLeftAsgn)
                        : makeToken();
                case LexerPhase.SignedShiftRight:
                    if (ch === ">") {
                        phase = LexerPhase.ShiftRight;
                        tokenType = tokens_1.TokenType.ShiftRight;
                        break;
                    }
                    return ch === "="
                        ? completeToken(tokens_1.TokenType.SignedShiftRightAsgn)
                        : makeToken();
                case LexerPhase.ShiftRight:
                    return ch === "="
                        ? completeToken(tokens_1.TokenType.ShiftRightAsgn)
                        : makeToken();
                // ====================================================================
                // --- Literals
                case LexerPhase.Zero:
                    if (ch === "x") {
                        phase = LexerPhase.HexaFirst;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    else if (ch === "b") {
                        phase = LexerPhase.BinaryFirst;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    else if (tokens_1.isDecimalDigit(ch) || ch === "_" || ch === "'") {
                        phase = LexerPhase.DecimalOrReal;
                    }
                    else if (ch === ".") {
                        phase = LexerPhase.RealFractionalFirst;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    else if (ch === "e" || ch === "E") {
                        phase = LexerPhase.RealExponent;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    else {
                        return makeToken();
                    }
                    break;
                case LexerPhase.Dot:
                    if (!tokens_1.isDecimalDigit(ch)) {
                        return makeToken();
                    }
                    phase = LexerPhase.RealFractionalTail;
                    tokenType = tokens_1.TokenType.RealLiteral;
                    break;
                case LexerPhase.HexaFirst:
                    if (ch === "_" || ch === "'") {
                        break;
                    }
                    if (!tokens_1.isHexadecimalDigit(ch)) {
                        return makeToken();
                    }
                    phase = LexerPhase.HexaTail;
                    tokenType = tokens_1.TokenType.HexadecimalLiteral;
                    break;
                case LexerPhase.HexaTail:
                    if (!tokens_1.isHexadecimalDigit(ch) && ch !== "_" && ch !== "'") {
                        return makeToken();
                    }
                    break;
                case LexerPhase.BinaryFirst:
                    if (ch === "_" || ch === "'") {
                        break;
                    }
                    if (!tokens_1.isBinaryDigit(ch)) {
                        return makeToken();
                    }
                    phase = LexerPhase.BinaryTail;
                    tokenType = tokens_1.TokenType.BinaryLiteral;
                    break;
                case LexerPhase.BinaryTail:
                    if (!tokens_1.isBinaryDigit(ch) && ch !== "_" && ch !== "'") {
                        return makeToken();
                    }
                    tokenType = tokens_1.TokenType.BinaryLiteral;
                    break;
                case LexerPhase.DecimalOrReal:
                    if (tokens_1.isDecimalDigit(ch) || ch === "_" || ch === "'") {
                        break;
                    }
                    else if (ch === ".") {
                        phase = LexerPhase.RealFractionalFirst;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    else if (ch === "e" || ch === "E") {
                        phase = LexerPhase.RealExponent;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    else {
                        return makeToken();
                    }
                    break;
                case LexerPhase.RealFractionalFirst:
                    if (tokens_1.isDecimalDigit(ch)) {
                        phase = LexerPhase.RealFractionalTail;
                        tokenType = tokens_1.TokenType.RealLiteral;
                    }
                    else if (ch === "e" || ch === "E") {
                        phase = LexerPhase.RealExponent;
                    }
                    else {
                        return makeToken();
                    }
                    break;
                case LexerPhase.RealFractionalTail:
                    if (ch === "e" || ch === "E") {
                        phase = LexerPhase.RealExponent;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    else if (!tokens_1.isDecimalDigit(ch) && ch !== "_" && ch !== "'") {
                        return makeToken();
                    }
                    break;
                case LexerPhase.RealExponent:
                    if (ch === "+" || ch === "-") {
                        phase = LexerPhase.RealExponentSign;
                    }
                    else if (tokens_1.isDecimalDigit(ch)) {
                        phase = LexerPhase.RealExponentTail;
                        tokenType = tokens_1.TokenType.RealLiteral;
                    }
                    else {
                        return makeToken();
                    }
                    break;
                case LexerPhase.RealExponentSign:
                    if (tokens_1.isDecimalDigit(ch)) {
                        phase = LexerPhase.RealExponentTail;
                        tokenType = tokens_1.TokenType.RealLiteral;
                    }
                    else {
                        return makeToken();
                    }
                    break;
                case LexerPhase.RealExponentTail:
                    if (!tokens_1.isDecimalDigit(ch)) {
                        return makeToken();
                    }
                    break;
                case LexerPhase.String:
                    if (ch === '"') {
                        return completeToken(tokens_1.TokenType.StringLiteral);
                    }
                    else if (tokens_1.isRestrictedInString(ch)) {
                        return completeToken(tokens_1.TokenType.Unknown);
                    }
                    else if (ch === "\\") {
                        phase = LexerPhase.StringBackSlash;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    break;
                // Start of string character escape
                case LexerPhase.StringBackSlash:
                    switch (ch) {
                        case "b":
                        case "f":
                        case "n":
                        case "r":
                        case "t":
                        case "v":
                        case "0":
                        case "'":
                        case '"':
                        case "\\":
                            phase = LexerPhase.String;
                            break;
                        default:
                            if (ch === "x") {
                                phase = LexerPhase.StringHexa1;
                            }
                            else {
                                phase = LexerPhase.String;
                            }
                    }
                    break;
                // First hexadecimal digit of string character escape
                case LexerPhase.StringHexa1:
                    if (tokens_1.isHexadecimalDigit(ch)) {
                        phase = LexerPhase.StringHexa2;
                    }
                    else {
                        return completeToken(tokens_1.TokenType.Unknown);
                    }
                    break;
                // Second hexadecimal digit of character escape
                case LexerPhase.StringHexa2:
                    if (tokens_1.isHexadecimalDigit(ch)) {
                        phase = LexerPhase.String;
                    }
                    else {
                        return completeToken(tokens_1.TokenType.Unknown);
                    }
                    break;
                // ====================================================================
                // --- We cannot continue
                default:
                    return makeToken();
            }
            // --- Append the char to the current text
            appendTokenChar();
            // --- Go on with parsing the next character
        }
        /**
         * Appends the last character to the token, and manages positions
         */
        function appendTokenChar() {
            text += ch;
            lexer._prefetched = null;
            lexer._prefetchedPos = null;
            lexer._prefetchedColumn = null;
            lastEndPos = input.position;
            lastEndColumn = input.position;
        }
        /**
         * Fetches the next character from the input stream
         */
        function fetchNextChar() {
            let ch;
            if (!lexer._prefetched) {
                lexer._prefetchedPos = input.position;
                lexer._prefetchedColumn = input.column;
                lexer._prefetched = input.get();
            }
            return lexer._prefetched;
        }
        /**
         * Packs the specified type of token to send back
         * @param type
         */
        function makeToken() {
            var _a;
            if (useResolver) {
                tokenType =
                    (_a = resolverHash[text]) !== null && _a !== void 0 ? _a : (tokens_1.isIdStart(text[0]) && text[text.length - 1] !== "'"
                        ? tokens_1.TokenType.Identifier
                        : tokens_1.TokenType.Unknown);
            }
            return {
                text,
                type: tokenType,
                location: {
                    fileIndex: input.fileIndex,
                    startPosition: startPos,
                    endPosition: lastEndPos,
                    startLine: line,
                    endLine: line,
                    startColumn,
                    endColumn: lastEndColumn,
                },
            };
        }
        /**
         * Add the last character to the token and return it
         */
        function completeToken(suggestedType) {
            appendTokenChar();
            // --- Send back the token
            if (suggestedType !== undefined) {
                tokenType = suggestedType;
            }
            return makeToken();
        }
    }
}
exports.WatSharpLexer = WatSharpLexer;
/**
 * This enum indicates the current lexer phase
 */
var LexerPhase;
(function (LexerPhase) {
    // Start getting a token
    LexerPhase[LexerPhase["Start"] = 0] = "Start";
    // Collecting whitespace
    LexerPhase[LexerPhase["InWhiteSpace"] = 1] = "InWhiteSpace";
    LexerPhase[LexerPhase["InlineCommentTrail"] = 2] = "InlineCommentTrail";
    LexerPhase[LexerPhase["BlockCommentTrail1"] = 3] = "BlockCommentTrail1";
    LexerPhase[LexerPhase["BlockCommentTrail2"] = 4] = "BlockCommentTrail2";
    LexerPhase[LexerPhase["Slash"] = 5] = "Slash";
    LexerPhase[LexerPhase["Asterisk"] = 6] = "Asterisk";
    LexerPhase[LexerPhase["Remainder"] = 7] = "Remainder";
    LexerPhase[LexerPhase["Plus"] = 8] = "Plus";
    LexerPhase[LexerPhase["Minus"] = 9] = "Minus";
    LexerPhase[LexerPhase["Xor"] = 10] = "Xor";
    LexerPhase[LexerPhase["Or"] = 11] = "Or";
    LexerPhase[LexerPhase["Ampersand"] = 12] = "Ampersand";
    LexerPhase[LexerPhase["Colon"] = 13] = "Colon";
    LexerPhase[LexerPhase["Equal"] = 14] = "Equal";
    LexerPhase[LexerPhase["Exclamation"] = 15] = "Exclamation";
    LexerPhase[LexerPhase["AngleLeft"] = 16] = "AngleLeft";
    LexerPhase[LexerPhase["AngleRight"] = 17] = "AngleRight";
    LexerPhase[LexerPhase["ShiftLeft"] = 18] = "ShiftLeft";
    LexerPhase[LexerPhase["ShiftRight"] = 19] = "ShiftRight";
    LexerPhase[LexerPhase["SignedShiftRight"] = 20] = "SignedShiftRight";
    LexerPhase[LexerPhase["IdTail"] = 21] = "IdTail";
    LexerPhase[LexerPhase["Dot"] = 22] = "Dot";
    LexerPhase[LexerPhase["Zero"] = 23] = "Zero";
    LexerPhase[LexerPhase["HexaFirst"] = 24] = "HexaFirst";
    LexerPhase[LexerPhase["HexaTail"] = 25] = "HexaTail";
    LexerPhase[LexerPhase["BinaryFirst"] = 26] = "BinaryFirst";
    LexerPhase[LexerPhase["BinaryTail"] = 27] = "BinaryTail";
    LexerPhase[LexerPhase["DecimalOrReal"] = 28] = "DecimalOrReal";
    LexerPhase[LexerPhase["RealFractionalFirst"] = 29] = "RealFractionalFirst";
    LexerPhase[LexerPhase["RealFractionalTail"] = 30] = "RealFractionalTail";
    LexerPhase[LexerPhase["RealExponent"] = 31] = "RealExponent";
    LexerPhase[LexerPhase["RealExponentSign"] = 32] = "RealExponentSign";
    LexerPhase[LexerPhase["RealExponentTail"] = 33] = "RealExponentTail";
    LexerPhase[LexerPhase["String"] = 34] = "String";
    LexerPhase[LexerPhase["StringBackSlash"] = 35] = "StringBackSlash";
    LexerPhase[LexerPhase["StringHexa1"] = 36] = "StringHexa1";
    LexerPhase[LexerPhase["StringHexa2"] = 37] = "StringHexa2";
})(LexerPhase || (LexerPhase = {}));
const resolverHash = {
    export: tokens_1.TokenType.Export,
    bool: tokens_1.TokenType.Bool,
    i8: tokens_1.TokenType.I8,
    sbyte: tokens_1.TokenType.I8,
    u8: tokens_1.TokenType.U8,
    byte: tokens_1.TokenType.U8,
    i16: tokens_1.TokenType.I16,
    short: tokens_1.TokenType.I16,
    u16: tokens_1.TokenType.U16,
    ushort: tokens_1.TokenType.U16,
    i32: tokens_1.TokenType.I32,
    int: tokens_1.TokenType.I32,
    u32: tokens_1.TokenType.U32,
    uint: tokens_1.TokenType.U32,
    i64: tokens_1.TokenType.I64,
    long: tokens_1.TokenType.I64,
    u64: tokens_1.TokenType.U64,
    ulong: tokens_1.TokenType.U64,
    f32: tokens_1.TokenType.F32,
    float: tokens_1.TokenType.F32,
    f64: tokens_1.TokenType.F64,
    double: tokens_1.TokenType.F64,
    void: tokens_1.TokenType.Void,
    Infinity: tokens_1.TokenType.Infinity,
    NaN: tokens_1.TokenType.NaN,
    true: tokens_1.TokenType.True,
    false: tokens_1.TokenType.False,
    clz: tokens_1.TokenType.Clz,
    ctz: tokens_1.TokenType.Ctz,
    popcnt: tokens_1.TokenType.PopCnt,
    abs: tokens_1.TokenType.Abs,
    ceil: tokens_1.TokenType.Ceil,
    floor: tokens_1.TokenType.Floor,
    trunc: tokens_1.TokenType.Trunc,
    nearest: tokens_1.TokenType.Nearest,
    sqrt: tokens_1.TokenType.Sqrt,
    min: tokens_1.TokenType.Min,
    max: tokens_1.TokenType.Max,
    neg: tokens_1.TokenType.Neg,
    copysign: tokens_1.TokenType.CopySign,
    inline: tokens_1.TokenType.Inline,
    struct: tokens_1.TokenType.Struct,
    sizeof: tokens_1.TokenType.Sizeof,
    const: tokens_1.TokenType.Const,
    global: tokens_1.TokenType.Global,
    type: tokens_1.TokenType.Type,
    table: tokens_1.TokenType.Table,
    data: tokens_1.TokenType.Data,
    import: tokens_1.TokenType.Import,
    local: tokens_1.TokenType.Local,
    if: tokens_1.TokenType.If,
    else: tokens_1.TokenType.Else,
    do: tokens_1.TokenType.Do,
    while: tokens_1.TokenType.While,
    break: tokens_1.TokenType.Break,
    continue: tokens_1.TokenType.Continue,
    return: tokens_1.TokenType.Return,
};


/***/ }),

/***/ "./src/compiler/WatSharpParser.ts":
/*!****************************************!*\
  !*** ./src/compiler/WatSharpParser.ts ***!
  \****************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.WatSharpParser = void 0;
const WatSharpLexer_1 = __webpack_require__(/*! ./WatSharpLexer */ "./src/compiler/WatSharpLexer.ts");
const PreprocessorParser_1 = __webpack_require__(/*! ../preprocessor/PreprocessorParser */ "./src/preprocessor/PreprocessorParser.ts");
const errors_1 = __webpack_require__(/*! ../core/errors */ "./src/core/errors.ts");
const tokens_1 = __webpack_require__(/*! ../core/tokens */ "./src/core/tokens.ts");
const token_traits_1 = __webpack_require__(/*! ./token-traits */ "./src/compiler/token-traits.ts");
const source_tree_1 = __webpack_require__(/*! ./source-tree */ "./src/compiler/source-tree.ts");
const MultiChunkInputStream_1 = __webpack_require__(/*! ../core/MultiChunkInputStream */ "./src/core/MultiChunkInputStream.ts");
class WatSharpParser {
    constructor(source, includeHandler, preprocessorSymbols, preprocess = true) {
        this.source = source;
        this.includeHandler = includeHandler;
        // --- Keep track of error messages
        this._parseErrors = [];
        // --- Source chunks to parse
        this._sourceChunks = null;
        // --- Use this lexer
        this._lexer = null;
        // --- Declarations parsed
        this._declarations = new Map();
        // --- Sequential function declaration ID
        this._nextFuncId = 0;
        // --- Prepare the preprocessor
        this._preprocessor = new PreprocessorParser_1.PreprocessorParser(source, 0, includeHandler !== null && includeHandler !== void 0 ? includeHandler : this.handleIncludeFiles);
        if (preprocessorSymbols) {
            preprocessorSymbols.forEach((s) => (this._preprocessor.preprocessorSymbols[s] = true));
        }
        // --- Carry out preprocessing, if declared so
        if (preprocess) {
            this.preprocessFiles();
        }
    }
    /**
     * The errors raised during the parse phase
     */
    get errors() {
        return this._parseErrors;
    }
    /**
     * Indicates if there were any errors during the parse phase
     */
    get hasErrors() {
        return this._parseErrors.length > 0;
    }
    /**
     * Gets the declarations of the WAT# program
     */
    get declarations() {
        return this._declarations;
    }
    /**
     * Preprocesses the source files
     */
    preprocessFiles() {
        try {
            this._sourceChunks = this._preprocessor.preprocessSource();
        }
        catch (err) {
            this._parseErrors = this._preprocessor.errors.slice(0);
            throw err;
        }
        // --- Prepare the lexer
        this._lexer = new WatSharpLexer_1.WatSharpLexer(new MultiChunkInputStream_1.MultiChunkInputStream(this._sourceChunks));
    }
    /**
     * Parse the entire WAT# program
     */
    parseProgram() {
        while (this._lexer.peek().type !== tokens_1.TokenType.Eof) {
            try {
                this.parseDeclaration();
            }
            catch (_a) {
                const lastErr = this._parseErrors[this._parseErrors.length - 1];
                // --- Skip the remaining part of the declaration
                if (lastErr &&
                    lastErr.code !== "W006" &&
                    !lastErr.code.startsWith("W1")) {
                    let token;
                    do {
                        token = this._lexer.get();
                    } while (token.type !== tokens_1.TokenType.Eof &&
                        token.type !== tokens_1.TokenType.Semicolon &&
                        token.type !== tokens_1.TokenType.RBrace);
                }
            }
        }
    }
    // ==========================================================================
    // Declarations
    /**
     * declaration
     *   : constDeclaration
     *   | globalDeclaration
     *   | typeDeclaration
     *   | tableDeclaration
     *   | dataDeclaration
     *   | importedFunctionDeclaration
     *   | functionDeclaration
     *   ;
     */
    parseDeclaration() {
        const { start, traits } = this.getParsePoint();
        if (traits.typeStart) {
            const spec = this.parseTypeSpecification();
            if (!spec) {
                this.reportError("W008");
                return;
            }
            const idToken = this._lexer.peek();
            if (idToken.type !== tokens_1.TokenType.Identifier) {
                this.reportError("W004");
                return;
            }
            this._lexer.get();
            const next = this._lexer.peek();
            if (next.type === tokens_1.TokenType.LBrace || next.type === tokens_1.TokenType.Semicolon) {
                this.parseVariableDeclarationTail(start, idToken.text, spec);
                return;
            }
            else if (next.type === tokens_1.TokenType.LParent) {
                this.parseFuntionDeclarationTail(start, idToken.text, spec);
                return;
            }
            this.reportError("W018");
            return;
        }
        switch (start.type) {
            case tokens_1.TokenType.Eof:
                return;
            case tokens_1.TokenType.Semicolon:
                this._lexer.get();
                break;
            case tokens_1.TokenType.Const:
                this.parseConstDeclaration();
                return;
            case tokens_1.TokenType.Global:
                this.parseGlobalDeclaration();
                return;
            case tokens_1.TokenType.Type:
                this.parseTypeDeclaration();
                return;
            case tokens_1.TokenType.Table:
                this.parseTableDeclaration();
                return;
            case tokens_1.TokenType.Data:
                this.parseDataDeclaration();
                return;
            case tokens_1.TokenType.Import:
                this.parseImportedFunctionDeclaration();
                return;
            case tokens_1.TokenType.Export:
            case tokens_1.TokenType.Inline:
            case tokens_1.TokenType.Void:
                this.parseFunctionDeclaration();
                return;
            default:
                this.reportError("W003", start, start.text);
                return;
        }
    }
    /**
     * constDeclaration
     *   : "const" intrinsicType identifier "=" expr ";"
     *   ;
     */
    parseConstDeclaration() {
        const keyword = this._lexer.get();
        const { start, traits } = this.getParsePoint();
        if (!traits.intrinsicType) {
            this.reportError("W005", start, start.text);
            return;
        }
        this._lexer.get();
        const id = this.expectToken(tokens_1.TokenType.Identifier, "W004");
        this.expectToken(tokens_1.TokenType.Asgn, "W007");
        const expr = this.parseExpr();
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.addDeclaration("ConstDeclaration", {
            name: id.text,
            underlyingType: tokens_1.TokenType[start.type].toLowerCase(),
            expr,
        }, keyword, semicolon);
    }
    /**
     * globalDeclaration
     *   : "global" intrinsicType identifier ("=" expr)? ";"
     *   ;
     */
    parseGlobalDeclaration() {
        const keyword = this._lexer.get();
        const { start, traits } = this.getParsePoint();
        if (!traits.intrinsicType) {
            this.reportError("W005", start, start.text);
            return;
        }
        this._lexer.get();
        const id = this.expectToken(tokens_1.TokenType.Identifier, "W004");
        let initExpr;
        if (this._lexer.peek().type === tokens_1.TokenType.Asgn) {
            this._lexer.get();
            initExpr = this.parseExpr();
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.addDeclaration("GlobalDeclaration", {
            name: id.text,
            underlyingType: tokens_1.TokenType[start.type].toLowerCase(),
            initExpr,
        }, keyword, semicolon);
    }
    /**
     * typeDeclaration
     *  : "type" identifier "=" typeSpecification ";"
     *  ;
     */
    parseTypeDeclaration() {
        const keyword = this._lexer.get();
        const id = this.expectToken(tokens_1.TokenType.Identifier, "W004");
        this.expectToken(tokens_1.TokenType.Asgn, "W007");
        const spec = this.parseTypeSpecification();
        if (spec === null) {
            this.reportError("W008");
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.addDeclaration("TypeDeclaration", {
            name: id.text,
            spec,
        }, keyword, semicolon);
    }
    /**
     * tableDeclaration
     *  : "table" (intrinsicType | "void") identifier "(" functionParam? ("," functionParam)* ")"
     *    "{" identifier ("," identifier)* "}" ";"
     *  ;
     */
    parseTableDeclaration() {
        const keyword = this._lexer.get();
        let resultType;
        // --- Table reult type
        const { start, traits } = this.getParsePoint();
        if (traits.typeStart) {
            resultType = this.parseTypeSpecification();
        }
        else if (start.type === tokens_1.TokenType.Void) {
            this._lexer.get();
        }
        else {
            this.reportError("W008");
            return;
        }
        // --- Check the result type
        if (resultType &&
            resultType.type !== "Intrinsic" &&
            resultType.type !== "Pointer") {
            this.reportError("W020");
            return;
        }
        const id = this.expectToken(tokens_1.TokenType.Identifier, "W004");
        this.expectToken(tokens_1.TokenType.LParent, "W016");
        const params = this.parseFunctionParameters();
        this.expectToken(tokens_1.TokenType.LBrace, "W009");
        const ids = [];
        do {
            const idToken = this._lexer.peek();
            if (idToken.type !== tokens_1.TokenType.Identifier) {
                this.reportError("W004");
                return;
            }
            ids.push(idToken.text);
            this._lexer.get();
            if (this._lexer.peek().type !== tokens_1.TokenType.Comma) {
                break;
            }
            this._lexer.get();
        } while (true);
        this.expectToken(tokens_1.TokenType.RBrace, "W010");
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.addDeclaration("TableDeclaration", {
            name: id.text,
            ids,
            params,
            resultType,
        }, keyword, semicolon);
    }
    /**
     * dataDeclaration
     *   : "data" (integralType)? identifier "[" expr? ("," expr)* "]"
     *   ;
     */
    parseDataDeclaration() {
        const keyword = this._lexer.get();
        const { start, traits } = this.getParsePoint();
        let underlyingType = "u8";
        if (traits.intrinsicType) {
            // --- Intrinsic type definition found
            if (start.type === tokens_1.TokenType.F32 || start.type === tokens_1.TokenType.F64) {
                this.reportError("W013");
                return;
            }
            underlyingType = tokens_1.TokenType[start.type].toLowerCase();
            this._lexer.get();
        }
        const id = this.expectToken(tokens_1.TokenType.Identifier, "W004");
        this.expectToken(tokens_1.TokenType.LSquare, "W011");
        const exprs = [];
        do {
            const expr = this.parseExpr();
            if (expr) {
                exprs.push(expr);
            }
            else {
                this.reportError("W002");
                return;
            }
            if (this._lexer.peek().type !== tokens_1.TokenType.Comma) {
                break;
            }
            this._lexer.get();
        } while (true);
        this.expectToken(tokens_1.TokenType.RSquare, "W012");
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.addDeclaration("DataDeclaration", {
            name: id.text,
            underlyingType,
            exprs,
        }, keyword, semicolon);
    }
    /**
     * variableDeclaration
     *   : typeSpecification identifier ( "{" identifier "}" )? ";"
     *   ;
     */
    parseVariableDeclarationTail(start, name, spec) {
        let addressAlias;
        let next = this._lexer.peek();
        if (next.type === tokens_1.TokenType.LBrace) {
            this._lexer.get();
            next = this._lexer.peek();
            if (next.type === tokens_1.TokenType.Identifier) {
                addressAlias = this.createExpressionNode("Identifier", {
                    name: next.text,
                }, next, next);
            }
            else {
                this.reportError("W004");
                return;
            }
            this._lexer.get();
            this.expectToken(tokens_1.TokenType.RBrace);
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.addDeclaration("VariableDeclaration", {
            name,
            spec,
            addressAlias,
        }, start, semicolon);
    }
    /**
     * importedFunctionDeclaration
     *   : "import" ("void" | intrinsicType) identifier stringLiteral stringLiteral
     *     "(" intrinsicType? ("," intrinsicType)* ")"
     *   ;
     */
    parseImportedFunctionDeclaration() {
        const keyword = this._lexer.get();
        let resultType;
        if (this._lexer.peek().type === tokens_1.TokenType.Void) {
            this._lexer.get();
        }
        else {
            const spec = this.parseTypeSpecification();
            if (!spec) {
                this.reportError("W014");
                return;
            }
            if (spec.type !== "Intrinsic") {
                this.reportError("W015");
                return;
            }
            resultType = spec;
        }
        const id = this.expectToken(tokens_1.TokenType.Identifier, "W004");
        const name1Literal = this.expectToken(tokens_1.TokenType.StringLiteral).text;
        const name1 = name1Literal.substr(1, name1Literal.length - 2);
        const name2Literal = this.expectToken(tokens_1.TokenType.StringLiteral).text;
        const name2 = name2Literal.substr(1, name2Literal.length - 2);
        this.expectToken(tokens_1.TokenType.LParent, "W016");
        let parSpecs = [];
        do {
            if (this._lexer.peek().type === tokens_1.TokenType.RParent) {
                break;
            }
            const parSpec = this.parseTypeSpecification();
            if (!parSpec) {
                this.reportError("W014");
                return;
            }
            if (parSpec.type !== "Intrinsic") {
                this.reportError("W015");
                return;
            }
            parSpecs.push(parSpec);
            if (this._lexer.peek().type !== tokens_1.TokenType.Comma) {
                break;
            }
            this._lexer.get();
        } while (true);
        this.expectToken(tokens_1.TokenType.RParent, "W017");
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.addDeclaration("ImportedFunctionDeclaration", {
            name: id.text,
            name1,
            name2,
            resultType,
            parSpecs,
        }, keyword, semicolon);
    }
    /**
     * functionDeclaration
     *   : functionModifier? ("void" | intrinsicType) identifier
     *     "(" functionParam? ("," functionParam)* ")"
     *     "{" bodyStatement* "}"
     *   ;
     *
     * functionModifier
     *   : "export"
     *   | "inline"
     *   ;
     *
     * functionParam
     *   : typeSpecification identifier
     *   ;
     */
    parseFunctionDeclaration() {
        let isExport;
        let isInline;
        let resultType;
        const keyword = this._lexer.peek();
        if (keyword.type === tokens_1.TokenType.Void) {
            this._lexer.get();
        }
        else {
            if (keyword.type === tokens_1.TokenType.Export) {
                isExport = true;
            }
            else if (keyword.type === tokens_1.TokenType.Inline) {
                isInline = true;
            }
            this._lexer.get();
            const { start, traits } = this.getParsePoint();
            if (traits.typeStart) {
                resultType = this.parseTypeSpecification();
            }
            else if (start.type === tokens_1.TokenType.Void) {
                this._lexer.get();
            }
            else {
                this.reportError("W008");
                return;
            }
        }
        const id = this.expectToken(tokens_1.TokenType.Identifier);
        this.parseFuntionDeclarationTail(keyword, id.text, resultType, isExport, isInline);
    }
    /**
     * Parses the tail of a function declaration
     * @param id Function identifier
     * @param resultType Function result type
     */
    parseFuntionDeclarationTail(start, name, resultType, isExport = false, isInline = false) {
        // --- We are before the opening parenthesis
        this.expectToken(tokens_1.TokenType.LParent, "W016");
        // --- Check the result type
        if (resultType &&
            resultType.type !== "Intrinsic" &&
            resultType.type !== "Pointer") {
            this.reportError("W020");
            return;
        }
        const params = this.parseFunctionParameters();
        if (!params) {
            return;
        }
        const body = [];
        this.parseBlockStatement(body, 0);
        this.addDeclaration("FunctionDeclaration", {
            funcId: this._nextFuncId++,
            name,
            resultType,
            params,
            isExport,
            isInline,
            body,
            invocationCount: 0
        }, start, start);
    }
    /**
     * Parses function parameters
     */
    parseFunctionParameters() {
        let params = [];
        do {
            const { start, traits } = this.getParsePoint();
            if (!traits.typeStart) {
                break;
            }
            const paramType = this.parseTypeSpecification();
            if (!paramType) {
                this.reportError("W008");
                return null;
            }
            if (paramType.type !== "Intrinsic" && paramType.type !== "Pointer") {
                this.reportError("W019");
                return null;
            }
            const id = this.expectToken(tokens_1.TokenType.Identifier, "W021");
            params.push({
                type: "FunctionParameter",
                name: id.text,
                spec: paramType,
                startPosition: start.location.startPosition,
                endPosition: id.location.endPosition,
                startLine: start.location.startLine,
                endLine: id.location.endLine,
                startColumn: start.location.startColumn,
                endColumn: id.location.endColumn,
            });
            if (this._lexer.peek().type !== tokens_1.TokenType.Comma) {
                break;
            }
            this._lexer.get();
        } while (true);
        this.expectToken(tokens_1.TokenType.RParent, "W017");
        if (this._lexer.peek().type !== tokens_1.TokenType.LBrace) {
            this.reportError("W009");
            return null;
        }
        return params;
    }
    /**
     * Parses a block statement
     */
    parseBlockStatement(body, loopDepth) {
        this._lexer.get();
        let next;
        while ((next = this._lexer.peek()).type !== tokens_1.TokenType.RBrace &&
            next.type !== tokens_1.TokenType.Eof) {
            try {
                this.parseStatement(body, loopDepth);
            }
            catch (err) {
                const lastError = this._parseErrors[this._parseErrors.length - 1];
                if (lastError &&
                    lastError.code !== "W006" &&
                    !lastError.code.startsWith("W1")) {
                    break;
                }
            }
        }
        this.expectToken(tokens_1.TokenType.RBrace, "W010");
    }
    /**
     * bodyStatement :=
     *   | blockStatement
     *   | localVariableDeclaration
     *   | assignment
     *   | localFunctionInvocation
     *   | controlFlowStatement
     *   ;
     */
    parseStatement(body, loopDepth) {
        const next = this._lexer.peek();
        // --- Check for local function invocation
        if (next.type === tokens_1.TokenType.Identifier &&
            this._lexer.ahead(1).type === tokens_1.TokenType.LParent) {
            this.parseLocalFunctionInvocation(body);
            return;
        }
        // --- Check for assignement
        if (next.type === tokens_1.TokenType.Identifier ||
            next.type === tokens_1.TokenType.Asterisk ||
            next.type === tokens_1.TokenType.LParent) {
            this.parseAssignment(body, loopDepth);
            return;
        }
        // --- Check other statements
        switch (next.type) {
            case tokens_1.TokenType.LBrace:
                this.parseBlockStatement(body, loopDepth);
                break;
            case tokens_1.TokenType.Semicolon:
                this._lexer.get();
                return;
            case tokens_1.TokenType.Local:
                this.parseLocalVariable(body);
                return;
            case tokens_1.TokenType.If:
                this.parseIfStatement(body, loopDepth);
                return;
            case tokens_1.TokenType.Do:
                this.parseDoStatement(body, loopDepth + 1);
                return;
            case tokens_1.TokenType.While:
                this.parseWhileStatement(body, loopDepth + 1);
                return;
            case tokens_1.TokenType.Break:
                this.parseBreakStatement(body, loopDepth);
                return;
            case tokens_1.TokenType.Continue:
                this.parseContinueStatement(body, loopDepth);
                return;
            case tokens_1.TokenType.Return:
                this.parseReturnStatement(body);
                return;
            default:
                this.reportError("W023");
                return;
        }
    }
    /**
     * localFunctionInvocation
     *   : identifier "(" expr? ("," expr)* ")" ";"
     *   ;
     */
    parseLocalFunctionInvocation(body) {
        const id = this._lexer.get();
        const args = this.parseFunctionArgs();
        if (args === null) {
            return;
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        const funcExpr = this.createExpressionNode("FunctionInvocation", {
            name: id.text,
            arguments: args.args,
            dispatcher: args.dispatcher,
        }, id, id);
        this.createStatementNode(body, "LocalFunctionInvocation", {
            name: id.text,
            invoked: funcExpr,
        }, id, semicolon);
    }
    /**
     * localVariableStatement
     *   : "local" (intrinsicType | pointerType) identifier ("=" expr) ";"
     *   ;
     */
    parseLocalVariable(body) {
        const start = this._lexer.get();
        const spec = this.parseTypeSpecification();
        if (!spec) {
            this.reportError("W008");
            return;
        }
        const id = this._lexer.peek();
        if (id.type !== tokens_1.TokenType.Identifier) {
            this.reportError("W004");
            return;
        }
        this._lexer.get();
        let initExpr;
        if (this._lexer.peek().type === tokens_1.TokenType.Asgn) {
            this._lexer.get();
            initExpr = this.parseExpr();
            if (!initExpr) {
                this.reportError("W002");
            }
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.createStatementNode(body, "LocalVariable", {
            name: id.text,
            spec,
            initExpr,
        }, id, semicolon);
    }
    /**
     * assignment
     *   : leftValue ( "=" | "+=" | "-=" | "*=" | "/=" | "%=" | "^=" | "&=" | "|="
     *     | "<<=" | ">>=" | ">>>=" )
     *     expr ";"
     *   ;
     */
    parseAssignment(body, loopDepth) {
        const lval = this.parseLeftValue();
        if (!lval) {
            return;
        }
        const { start, traits } = this.getParsePoint();
        if (!traits.assignmentOp) {
            this.reportError("W024");
            return;
        }
        this._lexer.get();
        const expr = this.parseExpr();
        if (!expr) {
            return;
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.createStatementNode(body, "Assignment", {
            lval,
            asgn: start.text,
            expr,
        }, start, semicolon);
    }
    /**
     *
     * leftValue
     *   : "*" leftValue
     *   | addressable
     *   ;
     */
    parseLeftValue() {
        const start = this._lexer.peek();
        if (start.type === tokens_1.TokenType.Asterisk) {
            this._lexer.get();
            const operand = this.parseLeftValue();
            if (!operand) {
                return;
            }
            return this.createExpressionNode("DereferenceExpression", {
                operand,
            }, start, start);
        }
        else if (start.type === tokens_1.TokenType.Identifier ||
            start.type === tokens_1.TokenType.LParent) {
            return this.parseAddressable();
        }
        this.reportError("W025");
        return null;
    }
    /**
     * addressable
     *   : (identifier | ( "(" leftValue ")" ) (("[" expr "]") | ("." addressable))*
     *   ;
     */
    parseAddressable() {
        const start = this._lexer.peek();
        let lval;
        if (start.type === tokens_1.TokenType.Identifier) {
            this._lexer.get();
            lval = this.createExpressionNode("Identifier", {
                name: start.text,
            }, start, start);
        }
        else if (start.type === tokens_1.TokenType.LParent) {
            this._lexer.get();
            lval = this.parseLeftValue();
            if (!lval) {
                return null;
            }
            this.expectToken(tokens_1.TokenType.RParent, "W017");
        }
        else {
            this.reportError("W004");
            return null;
        }
        do {
            const next = this._lexer.peek();
            if (next.type === tokens_1.TokenType.Dot) {
                // --- Process member Lval
                this._lexer.get();
                const idToken = this._lexer.peek();
                if (idToken.type !== tokens_1.TokenType.Identifier) {
                    this.reportError("W004");
                    return null;
                }
                this._lexer.get();
                lval = this.createExpressionNode("MemberAccess", {
                    member: idToken.text,
                    object: lval,
                }, next, idToken);
            }
            else if (next.type === tokens_1.TokenType.LSquare) {
                // --- Process index access
                this._lexer.get();
                const index = this.parseExpr();
                this.expectToken(tokens_1.TokenType.RSquare);
                lval = this.createExpressionNode("ItemAccess", {
                    index,
                    array: lval,
                }, next, this._lexer.peek());
            }
            else {
                break;
            }
        } while (true);
        return lval;
    }
    /**
     * ifStatement
     *   : "if" "(" condition ")" statement ("else" statement)?
     *   ;
     */
    parseIfStatement(body, loopDepth) {
        const start = this._lexer.get();
        this.expectToken(tokens_1.TokenType.LParent, "W016");
        const test = this.parseExpr();
        if (test === null) {
            this.reportError("W002");
            return;
        }
        this.expectToken(tokens_1.TokenType.RParent, "W017");
        const consequent = [];
        this.parseStatement(consequent, loopDepth);
        let alternate;
        const next = this._lexer.peek();
        if (next.type === tokens_1.TokenType.Else) {
            this._lexer.get();
            alternate = [];
            this.parseStatement(alternate, loopDepth);
        }
        this.createStatementNode(body, "If", {
            test,
            consequent,
            alternate,
        }, start, start);
    }
    /**
     * doWhileStatement
     *   : "do" statement "while" "(" expr ")" ";"
     *   ;
     */
    parseDoStatement(body, loopDepth) {
        const start = this._lexer.get();
        const loopBody = [];
        this.parseStatement(loopBody, loopDepth);
        this.expectToken(tokens_1.TokenType.While, "W022");
        this.expectToken(tokens_1.TokenType.LParent, "W016");
        const test = this.parseExpr();
        if (test === null) {
            this.reportError("W002");
            return;
        }
        this.expectToken(tokens_1.TokenType.RParent, "W017");
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.createStatementNode(body, "Do", {
            loopBody,
            test,
        }, start, semicolon);
    }
    /**
     * whileStatement
     *   : "while" "(" expr ")" statement
     *   ;
     */
    parseWhileStatement(body, loopDepth) {
        const start = this._lexer.get();
        this.expectToken(tokens_1.TokenType.LParent, "W016");
        const test = this.parseExpr();
        if (test === null) {
            this.reportError("W002");
            return;
        }
        this.expectToken(tokens_1.TokenType.RParent, "W017");
        const loopBody = [];
        this.parseStatement(loopBody, loopDepth);
        this.createStatementNode(body, "While", {
            loopBody,
            test,
        }, start, start);
    }
    /**
     * breakStatement
     *   : "break" ";"
     *   ;
     */
    parseBreakStatement(body, loopDepth) {
        const start = this._lexer.get();
        if (loopDepth < 1) {
            this.reportError("W110");
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.createStatementNode(body, "Break", {}, start, semicolon);
    }
    /**
     * continueStatement
     *   : "continue" ";"
     *   ;
     */
    parseContinueStatement(body, loopDepth) {
        const start = this._lexer.get();
        if (loopDepth < 1) {
            this.reportError("W111");
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.createStatementNode(body, "Continue", {}, start, semicolon);
    }
    /**
     * returnStatement
     *   : "return" expr? ";"
     *   ;
     */
    parseReturnStatement(body) {
        const start = this._lexer.get();
        const next = this._lexer.peek();
        let expr;
        if (next.type !== tokens_1.TokenType.Semicolon) {
            // --- Try to obtain the return expression
            expr = this.parseExpr();
        }
        const semicolon = this.expectToken(tokens_1.TokenType.Semicolon, "W006");
        this.createStatementNode(body, "Return", {
            expr,
        }, start, semicolon);
    }
    /**
     * Adds a new declaration to the existing ones
     * @param type Declaration type
     * @param stump Declaration to add
     * @param startToken: Start token
     * @param endToken: End token
     */
    addDeclaration(type, stump, startToken, endToken) {
        if (!stump.name) {
            throw new Error("A declaration must have a name.");
        }
        if (this._declarations.has(stump.name)) {
            this.reportError("W100", undefined, stump.name);
        }
        const updatedDecl = Object.assign({}, stump, {
            type,
            startPosition: startToken.location.startPosition,
            endPosition: endToken.location.endPosition,
            startLine: startToken.location.startLine,
            startColumn: startToken.location.startColumn,
            endLine: endToken.location.endLine,
            endColumn: endToken.location.startColumn,
            order: this._declarations.size,
        });
        this._declarations.set(stump.name, updatedDecl);
        return updatedDecl;
    }
    // ==========================================================================
    // Expression parsing
    /**
     * expr
     *   : parExpr
     *   | brackExpr
     *   | conditionalExpr
     *   ;
     */
    parseExpr() {
        const traits = this.getParsePoint().traits;
        if (traits.expressionStart) {
            return this.parseCondExpr();
        }
        return null;
    }
    /**
     * conditionalExpr
     *   : orExpr ( "?" expr ":" expr )?
     *   ;
     */
    parseCondExpr() {
        const startToken = this._lexer.peek();
        const condExpr = this.parseOrExpr();
        if (!condExpr) {
            return null;
        }
        if (!this.skipToken(tokens_1.TokenType.QuestionMark)) {
            return condExpr;
        }
        const trueExpr = this.getExpression();
        this.expectToken(tokens_1.TokenType.Colon);
        const falseExpr = this.getExpression();
        const endToken = this._lexer.peek();
        return this.createExpressionNode("ConditionalExpression", {
            condition: condExpr,
            consequent: trueExpr,
            alternate: falseExpr,
        }, startToken, endToken);
    }
    /**
     * orExpr
     *   : xorExpr ( "|" xorExpr )?
     *   ;
     */
    parseOrExpr() {
        let leftExpr = this.parseXorExpr();
        if (!leftExpr) {
            return null;
        }
        while (this.skipToken(tokens_1.TokenType.Or)) {
            const startToken = this._lexer.peek();
            const rightExpr = this.parseXorExpr();
            if (!rightExpr) {
                this.reportError("W002");
                return null;
            }
            const endToken = this._lexer.peek();
            leftExpr = this.createExpressionNode("BinaryExpression", {
                operator: "|",
                left: leftExpr,
                right: rightExpr,
            }, startToken, endToken);
        }
        return leftExpr;
    }
    /**
     * xorExpr
     *   : andExpr ( "^" andExpr )?
     *   ;
     */
    parseXorExpr() {
        let leftExpr = this.parseAndExpr();
        if (!leftExpr) {
            return null;
        }
        while (this.skipToken(tokens_1.TokenType.Xor)) {
            const startToken = this._lexer.peek();
            const rightExpr = this.parseAndExpr();
            if (!rightExpr) {
                this.reportError("W002");
                return null;
            }
            const endToken = this._lexer.peek();
            leftExpr = this.createExpressionNode("BinaryExpression", {
                operator: "^",
                left: leftExpr,
                right: rightExpr,
            }, startToken, endToken);
        }
        return leftExpr;
    }
    /**
     * andExpr
     *   : equExpr ( "&" equExpr )?
     *   ;
     */
    parseAndExpr() {
        let leftExpr = this.parseEquExpr();
        if (!leftExpr) {
            return null;
        }
        while (this.skipToken(tokens_1.TokenType.Ampersand)) {
            const startToken = this._lexer.peek();
            const rightExpr = this.parseEquExpr();
            if (!rightExpr) {
                this.reportError("W002");
                return null;
            }
            const endToken = this._lexer.peek();
            leftExpr = this.createExpressionNode("BinaryExpression", {
                operator: "&",
                left: leftExpr,
                right: rightExpr,
            }, startToken, endToken);
        }
        return leftExpr;
    }
    /**
     * equExpr
     *   : relExpr ( ( "==" | "!=" ) relExpr )?
     *   ;
     */
    parseEquExpr() {
        let leftExpr = this.parseRelExpr();
        if (!leftExpr) {
            return null;
        }
        let opType;
        while ((opType = this.skipTokens(tokens_1.TokenType.Equal, tokens_1.TokenType.NotEqual))) {
            const startToken = this._lexer.peek();
            const rightExpr = this.parseRelExpr();
            if (!rightExpr) {
                this.reportError("W002");
                return null;
            }
            const endToken = this._lexer.peek();
            leftExpr = this.createExpressionNode("BinaryExpression", {
                type: "BinaryExpression",
                operator: opType.text,
                left: leftExpr,
                right: rightExpr,
            }, startToken, endToken);
        }
        return leftExpr;
    }
    /**
     * relExpr
     *   : shiftExpr ( ( "<" | "<=" | ">" | ">=" ) shiftExpr )?
     *   ;
     */
    parseRelExpr() {
        let leftExpr = this.parseShiftExpr();
        if (!leftExpr) {
            return null;
        }
        let opType;
        while ((opType = this.skipTokens(tokens_1.TokenType.LessThan, tokens_1.TokenType.LessThanOrEqual, tokens_1.TokenType.GreaterThan, tokens_1.TokenType.GreaterThanOrEqual))) {
            const startToken = this._lexer.peek();
            const rightExpr = this.parseShiftExpr();
            if (!rightExpr) {
                this.reportError("W002");
                return null;
            }
            const endToken = this._lexer.peek();
            leftExpr = this.createExpressionNode("BinaryExpression", {
                operator: opType.text,
                left: leftExpr,
                right: rightExpr,
            }, startToken, endToken);
        }
        return leftExpr;
    }
    /**
     * shiftExpr
     *   : addExpr ( ( "<<" | ">>" ) addExpr )?
     *   ;
     */
    parseShiftExpr() {
        let leftExpr = this.parseAddExpr();
        if (!leftExpr) {
            return null;
        }
        let opType;
        while ((opType = this.skipTokens(tokens_1.TokenType.ShiftLeft, tokens_1.TokenType.ShiftRight, tokens_1.TokenType.SignedShiftRight))) {
            const startToken = this._lexer.peek();
            const rightExpr = this.parseAddExpr();
            if (!rightExpr) {
                this.reportError("W002");
                return null;
            }
            const endToken = this._lexer.peek();
            leftExpr = this.createExpressionNode("BinaryExpression", {
                operator: opType.text,
                left: leftExpr,
                right: rightExpr,
            }, startToken, endToken);
        }
        return leftExpr;
    }
    /**
     * addExpr
     *   : multExpr ( ( "+" | "-" ) multExpr )?
     *   ;
     */
    parseAddExpr() {
        let leftExpr = this.parseMultExpr();
        if (!leftExpr) {
            return null;
        }
        let opType;
        while ((opType = this.skipTokens(tokens_1.TokenType.Plus, tokens_1.TokenType.Minus))) {
            const startToken = this._lexer.peek();
            const rightExpr = this.parseMultExpr();
            if (!rightExpr) {
                this.reportError("W002");
                return null;
            }
            const endToken = this._lexer.peek();
            leftExpr = this.createExpressionNode("BinaryExpression", {
                operator: opType.text,
                left: leftExpr,
                right: rightExpr,
            }, startToken, endToken);
        }
        return leftExpr;
    }
    /**
     * multExpr
     *   : memberOrIxdexExpr ( ( "*" | "/" | "%") memberOrIndexExpr )?
     *   ;
     */
    parseMultExpr() {
        let leftExpr = this.parseMemberOrIndexExpr();
        if (!leftExpr) {
            return null;
        }
        let opType;
        while ((opType = this.skipTokens(tokens_1.TokenType.Asterisk, tokens_1.TokenType.Divide, tokens_1.TokenType.Remainder))) {
            const startToken = this._lexer.peek();
            const rightExpr = this.parseMemberOrIndexExpr();
            if (!rightExpr) {
                this.reportError("W002");
                return null;
            }
            const endToken = this._lexer.peek();
            leftExpr = this.createExpressionNode("BinaryExpression", {
                operator: opType.text,
                left: leftExpr,
                right: rightExpr,
            }, startToken, endToken);
        }
        return leftExpr;
    }
    /**
     * memberOrIndexExpression
     *   : memberExpression
     *   | indexExpression
     *   ;
     *
     * memberExpression
     *   : primaryExpression "." expr
     *   ;
     *
     * indexExpression
     *   : primaryExpression "[" expr "]"
     *   ;
     */
    parseMemberOrIndexExpr() {
        let operand = this.parseUnaryExpr();
        if (!operand) {
            return null;
        }
        do {
            const next = this._lexer.peek();
            if (next.type === tokens_1.TokenType.Dot) {
                // --- Process member access
                this._lexer.get();
                const idToken = this._lexer.peek();
                if (idToken.type !== tokens_1.TokenType.Identifier) {
                    this.reportError("W004");
                    return null;
                }
                this._lexer.get();
                operand = this.createExpressionNode("MemberAccess", {
                    object: operand,
                    member: idToken.text,
                }, next, idToken);
            }
            else if (next.type === tokens_1.TokenType.LSquare) {
                // --- Process index access
                this._lexer.get();
                const indexExpr = this.parseExpr();
                this.expectToken(tokens_1.TokenType.RSquare);
                operand = this.createExpressionNode("ItemAccess", {
                    array: operand,
                    index: indexExpr,
                }, next, this._lexer.peek());
            }
            else {
                break;
            }
        } while (true);
        return operand;
    }
    /**
     * unaryExpr
     *   : ( | "+" | "-" | "~" | "!" | "&" ) primaryExpr
     *   ;
     */
    parseUnaryExpr() {
        const { start, traits } = this.getParsePoint();
        if (!traits.unaryOp) {
            return this.parsePrimaryExpr();
        }
        this._lexer.get();
        const unaryOperand = this.parsePrimaryExpr();
        if (!unaryOperand) {
            return null;
        }
        return this.createExpressionNode("UnaryExpression", {
            operator: start.text,
            operand: unaryOperand,
        }, start, start);
    }
    /**
     * dereferenceExpr
     *   : "*" primaryExpr
     *   ;
     */
    parseDereferenceExpr() {
        const start = this._lexer.get();
        const operand = this.parsePrimaryExpr();
        if (!operand) {
            this.reportError("W002");
            return null;
        }
        return this.createExpressionNode("DereferenceExpression", {
            operand,
        }, start, start);
    }
    /**
     * primaryExpr
     *   : "sizeof" "(" typeSpec ")"
     *   | funcInvocation
     *   | literal
     *   | identifier
     *   | unaryExpr
     *   | "(" expr ")"
     *   ;
     */
    parsePrimaryExpr() {
        const { start, traits } = this.getParsePoint();
        if (traits.intrinsicType) {
            return this.parseTypeCastExpression();
        }
        if (traits.builtInFunc) {
            this._lexer.get();
            const args = this.parseFunctionArgs();
            if (!args) {
                return null;
            }
            return this.createExpressionNode("BuiltInFunctionInvocation", {
                name: start.text,
                arguments: args.args,
                dispatcher: args.dispatcher,
            }, start, start);
        }
        switch (start.type) {
            case tokens_1.TokenType.Sizeof:
                this._lexer.get();
                this.expectToken(tokens_1.TokenType.LParent, "W016");
                const typeSpec = this.parseTypeSpecification();
                this.expectToken(tokens_1.TokenType.RParent);
                return this.createExpressionNode("SizeOfExpression", {
                    spec: typeSpec,
                }, start, start);
            case tokens_1.TokenType.LParent:
                this._lexer.get();
                const parenthesizedExpr = this.parseExpr();
                if (!parenthesizedExpr) {
                    return null;
                }
                this.expectToken(tokens_1.TokenType.RParent, "W017");
                return parenthesizedExpr;
            case tokens_1.TokenType.Identifier:
                const idToken = this._lexer.get();
                if (this._lexer.peek().type === tokens_1.TokenType.LParent) {
                    const args = this.parseFunctionArgs();
                    if (!args) {
                        return null;
                    }
                    return this.createExpressionNode("FunctionInvocation", {
                        name: idToken.text,
                        arguments: args.args,
                        dispatcher: args.dispatcher,
                    }, idToken, idToken);
                }
                return this.createExpressionNode("Identifier", {
                    name: idToken.text,
                }, idToken, idToken);
            case tokens_1.TokenType.False:
            case tokens_1.TokenType.True:
                this._lexer.get();
                return this.createExpressionNode("Literal", {
                    value: start.type === tokens_1.TokenType.True ? 1 : 0,
                    source: source_tree_1.LiteralSource.Int
                }, start, start);
            case tokens_1.TokenType.BinaryLiteral:
                this._lexer.get();
                return this.parseBinaryLiteral(start);
            case tokens_1.TokenType.DecimalLiteral:
                this._lexer.get();
                return this.parseDecimalLiteral(start);
            case tokens_1.TokenType.HexadecimalLiteral:
                this._lexer.get();
                return this.parseHexadecimalLiteral(start);
            case tokens_1.TokenType.RealLiteral:
                this._lexer.get();
                return this.parseRealLiteral(start);
            case tokens_1.TokenType.Infinity:
                this._lexer.get();
                return this.createExpressionNode("Literal", {
                    value: Infinity,
                }, start, start);
            case tokens_1.TokenType.NaN:
                this._lexer.get();
                return this.createExpressionNode("Literal", {
                    value: NaN,
                }, start, start);
            case tokens_1.TokenType.Plus:
            case tokens_1.TokenType.Minus:
            case tokens_1.TokenType.BinaryNot:
            case tokens_1.TokenType.Not:
            case tokens_1.TokenType.Ampersand:
                return this.parseUnaryExpr();
            case tokens_1.TokenType.Asterisk:
                return this.parseDereferenceExpr();
        }
        return null;
    }
    /**
     * typeCastExpr
     *   : intrinsicType "(" expr ")"
     */
    parseTypeCastExpression() {
        const start = this._lexer.get();
        this.expectToken(tokens_1.TokenType.LParent, "W016");
        const expr = this.parseExpr();
        if (!expr) {
            this.reportError("W002");
            return null;
        }
        this.expectToken(tokens_1.TokenType.RParent);
        return this.createExpressionNode("TypeCast", {
            name: tokens_1.TokenType[start.type].toLowerCase(),
            operand: expr,
        }, start, start);
    }
    /**
     * functionArgsExpr
     *   : "(" expr? ("," expr)* ")" ("[" expr "]")?
     *   ;
     */
    parseFunctionArgs() {
        const args = [];
        let dispatcher;
        this.expectToken(tokens_1.TokenType.LParent, "W016");
        do {
            const { traits } = this.getParsePoint();
            if (!traits.expressionStart) {
                break;
            }
            const expr = this.parseExpr();
            if (!expr) {
                this.reportError("W002");
                return null;
            }
            args.push(expr);
            if (!this.skipToken(tokens_1.TokenType.Comma)) {
                break;
            }
        } while (true);
        this.expectToken(tokens_1.TokenType.RParent);
        const next = this._lexer.peek();
        if (next.type === tokens_1.TokenType.LSquare) {
            // --- Optional dispatcher
            this._lexer.get();
            dispatcher = this.parseExpr();
            if (!dispatcher) {
                return null;
            }
            this.expectToken(tokens_1.TokenType.RSquare, "W012");
        }
        return {
            args,
            dispatcher,
        };
    }
    /**
     * Parses a binary literal
     * @param token Literal token
     */
    parseBinaryLiteral(token) {
        let value;
        const bigValue = BigInt(token.text.replace(/[_']/g, ""));
        if (bigValue < Number.MIN_SAFE_INTEGER ||
            bigValue > Number.MAX_SAFE_INTEGER) {
            value = bigValue;
        }
        else {
            value = parseInt(token.text.substr(2).replace(/[_']/g, ""), 2);
        }
        return this.createExpressionNode("Literal", {
            value,
            source: typeof value === "number" ? source_tree_1.LiteralSource.Int : source_tree_1.LiteralSource.BigInt,
        }, token, token);
    }
    /**
     * Parses a decimal literal
     * @param token Literal token
     */
    parseDecimalLiteral(token) {
        let value;
        const bigValue = BigInt(token.text.replace(/[_']/g, ""));
        if (bigValue < Number.MIN_SAFE_INTEGER ||
            bigValue > Number.MAX_SAFE_INTEGER) {
            value = bigValue;
        }
        else {
            value = parseInt(token.text.replace(/[_']/g, ""), 10);
        }
        return this.createExpressionNode("Literal", {
            value,
            source: typeof value === "bigint"
                ? source_tree_1.LiteralSource.BigInt
                : Number.isInteger(value)
                    ? source_tree_1.LiteralSource.Int
                    : source_tree_1.LiteralSource.Real,
        }, token, token);
    }
    /**
     * Parses a hexadecimal literal
     * @param token Literal token
     */
    parseHexadecimalLiteral(token) {
        let value;
        const bigValue = BigInt(token.text.replace(/[_']/g, ""));
        if (bigValue < Number.MIN_SAFE_INTEGER ||
            bigValue > Number.MAX_SAFE_INTEGER) {
            value = bigValue;
        }
        else {
            value = parseInt(token.text.substr(2).replace(/[_']/g, ""), 16);
        }
        return this.createExpressionNode("Literal", {
            value,
            source: typeof value === "number" ? source_tree_1.LiteralSource.Int : source_tree_1.LiteralSource.BigInt,
        }, token, token);
    }
    /**
     * Parses a real literal
     * @param token Literal token
     */
    parseRealLiteral(token) {
        let value = parseFloat(token.text.replace(/[_']/g, ""));
        return this.createExpressionNode("Literal", {
            value,
            source: source_tree_1.LiteralSource.Real,
        }, token, token);
    }
    /**
     * Gets an expression
     * @param optional Is the expression optional?
     * @param leadingComma Test for leading comma?
     */
    getExpression(optional = false, leadingComma = false) {
        if (leadingComma) {
            if (!this.skipToken(tokens_1.TokenType.Comma)) {
                if (!optional) {
                    this.reportError("W001");
                }
                return null;
            }
            else {
                // --- We have a comma, so the expression in not optional
                optional = false;
            }
        }
        const expr = this.parseExpr();
        if (expr) {
            return expr;
        }
        if (!optional) {
            this.reportError("W002");
        }
        return null;
    }
    // ==========================================================================
    // Type specification parsing
    /**
     * typeSpecification
     *   : primaryTypeSpecification ("[" expr "]")*
     *   ;
     */
    parseTypeSpecification() {
        let typeSpec = this.parsePrimaryTypeSpec();
        if (!typeSpec) {
            return null;
        }
        do {
            const next = this._lexer.peek();
            if (next.type !== tokens_1.TokenType.LSquare) {
                break;
            }
            this._lexer.get();
            const size = this.parseExpr();
            if (size === null) {
                this.reportError("W002");
            }
            this.expectToken(tokens_1.TokenType.RSquare);
            typeSpec = this.createTypeSpecNode("Array", {
                spec: typeSpec,
                size,
            }, next, next);
        } while (true);
        return typeSpec;
    }
    /**
     * primaryTypeSpecification
     *   : instrinsicType
     *   | identifier
     *   | "*" primaryTypeSpecification
     *   | "(" typeSpecification ")"
     *   | structType
     *   ;
     */
    parsePrimaryTypeSpec() {
        const { start, traits } = this.getParsePoint();
        if (traits.intrinsicType) {
            this._lexer.get();
            return this.createTypeSpecNode("Intrinsic", {
                underlying: tokens_1.TokenType[start.type].toLowerCase(),
            }, start, start);
        }
        switch (start.type) {
            case tokens_1.TokenType.Identifier:
                this._lexer.get();
                return this.createTypeSpecNode("NamedType", {
                    name: start.text,
                }, start, start);
            case tokens_1.TokenType.Asterisk:
                this._lexer.get();
                return this.createTypeSpecNode("Pointer", {
                    spec: this.parsePrimaryTypeSpec(),
                }, start, start);
            case tokens_1.TokenType.LParent:
                this._lexer.get();
                const typeSpec = this.parseTypeSpecification();
                this.expectToken(tokens_1.TokenType.RParent);
                return typeSpec;
            case tokens_1.TokenType.Struct:
                return this.parseStructType();
        }
        return null;
    }
    /**
     * structType
     *   : "struct" "{" fieldDef ("," fieldDef)* "}"
     *   ;
     
     * fieldDef
     *   : typeSpecification identifier
     *   ;
     */
    parseStructType() {
        const start = this._lexer.get();
        this.expectToken(tokens_1.TokenType.LBrace, "W009");
        const struct = this.createTypeSpecNode("Struct", {
            fields: [],
        }, start, start);
        while (true) {
            // --- Get field type
            const fieldSpec = this.parseTypeSpecification();
            const id = this._lexer.peek();
            if (id.type !== tokens_1.TokenType.Identifier) {
                this.reportError("W004");
                return null;
            }
            this._lexer.get();
            struct.fields.push(this.createTypeSpecNode("StructField", {
                id: id.text,
                spec: fieldSpec,
            }, id, id));
            if (!this.skipToken(tokens_1.TokenType.Comma)) {
                break;
            }
        }
        this.expectToken(tokens_1.TokenType.RBrace, "W010");
        return struct;
    }
    // ==========================================================================
    // Default include file handling
    /**
     * This method handles the include files with the preprocessor
     * @param filename
     */
    handleIncludeFiles(baseFileIndex, filename) {
        return {
            fileIndex: 0,
            source: "",
        };
    }
    // ==========================================================================
    // Helpers
    /**
     * Gets the current parse point
     */
    getParsePoint() {
        const start = this._lexer.peek();
        const traits = token_traits_1.getTokenTraits(start.type);
        return { start, traits };
    }
    /**
     * Tests the type of the next token
     * @param type Expected token type
     */
    expectToken(type, errorCode, allowEof) {
        const next = this._lexer.peek();
        if (next.type === type || (allowEof && next.type === tokens_1.TokenType.Eof)) {
            // --- Skip the expected token
            return this._lexer.get();
        }
        this.reportError(errorCode !== null && errorCode !== void 0 ? errorCode : "W003", next, next.text);
        return null;
    }
    /**
     * Skips the next token if the type is the specified one
     * @param type Token type to check
     */
    skipToken(type) {
        const next = this._lexer.peek();
        if (next.type === type) {
            this._lexer.get();
            return next;
        }
        return null;
    }
    /**
     * Skips the next token if the type is the specified one
     * @param type Token type to check
     */
    skipTokens(...types) {
        const next = this._lexer.peek();
        for (const type of types) {
            if (next.type === type) {
                this._lexer.get();
                return next;
            }
        }
        return null;
    }
    /**
     * Creates an expression node
     * @param type Expression type
     * @param stump Stump properties
     * @param startToken The token that starts the expression
     * @param endToken The token that ends the expression
     */
    createExpressionNode(type, stump, startToken, endToken) {
        const startPosition = startToken.location.startPosition;
        const endPosition = endToken.location.startPosition;
        return Object.assign({}, stump, {
            type,
            startPosition,
            endPosition,
            startLine: startToken.location.startLine,
            startColumn: startToken.location.startColumn,
            endLine: endToken.location.endLine,
            endColumn: endToken.location.startColumn,
        });
    }
    /**
     * Creates an expression node
     * @param type Expression type
     * @param stump Stump properties
     * @param startToken The token that starts the expression
     * @param endToken The token that ends the expression
     */
    createTypeSpecNode(type, stump, startToken, endToken) {
        const startPosition = startToken.location.startPosition;
        const endPosition = endToken.location.startPosition;
        return Object.assign({}, stump, {
            type,
            startPosition,
            endPosition,
            startLine: startToken.location.startLine,
            startColumn: startToken.location.startColumn,
            endLine: endToken.location.endLine,
            endColumn: endToken.location.startColumn,
        });
    }
    /**
     * Creates an expression node
     * @param body List of statements to add the new statement
     * @param type Expression type
     * @param stump Stump properties
     * @param startToken The token that starts the expression
     * @param endToken The token that ends the expression
     */
    createStatementNode(body, type, stump, startToken, endToken) {
        const startPosition = startToken.location.startPosition;
        const endPosition = endToken.location.startPosition;
        const newStatement = Object.assign({}, stump, {
            type,
            startPosition,
            endPosition,
            startLine: startToken.location.startLine,
            startColumn: startToken.location.startColumn,
            endLine: endToken.location.endLine,
            endColumn: endToken.location.startColumn,
        });
        body.push(newStatement);
        return newStatement;
    }
    /**
     * Reports the specified error
     * @param errorCode Error code
     * @param token Token that represents the error's position
     * @param options Error message options
     */
    reportError(errorCode, token, ...options) {
        var _a;
        let errorText = (_a = errors_1.errorMessages[errorCode]) !== null && _a !== void 0 ? _a : "Unkonwn error";
        if (options) {
            options.forEach((o, idx) => (errorText = replace(errorText, `{${idx}}`, options[idx].toString())));
        }
        if (!token) {
            token = this._lexer.peek();
        }
        this._parseErrors.push({
            code: errorCode,
            text: errorText,
            line: token.location.startLine,
            column: token.location.startColumn,
            position: token.location.startPosition,
        });
        throw new errors_1.ParserError(errorText, errorCode);
        function replace(input, placeholder, replacement) {
            do {
                input = input.replace(placeholder, replacement);
            } while (input.includes(placeholder));
            return input;
        }
    }
}
exports.WatSharpParser = WatSharpParser;


/***/ }),

/***/ "./src/compiler/expression-resolver.ts":
/*!*********************************************!*\
  !*** ./src/compiler/expression-resolver.ts ***!
  \*********************************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.renderExpression = exports.applyBuiltInFunction = exports.applyBinaryOperation = exports.applyUnaryOperation = exports.applyTypeCast = exports.resolveConstantExpression = void 0;
/**
 * Resolves the specified expression to a constant value
 * @param expr Expression to resolve
 * @param constResolver Function to resolve nested constant values
 * @param sizeOfResolver Function to resolve "sizeof"
 */
function resolveConstantExpression(expr, constResolver, sizeOfResolver, reportError) {
    switch (expr.type) {
        case "FunctionInvocation":
        case "ItemAccess":
        case "MemberAccess":
            // --- You cannot use this constructs in a constant expression
            reportError("W104", expr);
            return;
        case "Literal":
            // --- A literal is resolved during the parse phase
            return;
        case "BuiltInFunctionInvocation":
            let allResolved = true;
            expr.arguments.forEach((arg) => {
                resolveConstantExpression(arg, constResolver, sizeOfResolver, reportError);
                allResolved = allResolved && arg.value !== undefined;
            });
            if (allResolved) {
                expr.value = applyBuiltInFunction(expr.name, expr.arguments.map((arg) => arg.value));
            }
            return;
        case "ConditionalExpression":
            // --- Resolve all the three operands
            resolveConstantExpression(expr.condition, constResolver, sizeOfResolver, reportError);
            resolveConstantExpression(expr.consequent, constResolver, sizeOfResolver, reportError);
            resolveConstantExpression(expr.alternate, constResolver, sizeOfResolver, reportError);
            if (expr.condition.value !== undefined &&
                expr.consequent.value !== undefined &&
                expr.alternate.value !== undefined) {
                // --- Evaluate if all operands resolved successfully
                expr.value = expr.condition.value
                    ? expr.consequent.value
                    : expr.alternate.value;
            }
            return;
        case "Identifier":
            constResolver(expr);
            return;
        case "SizeOfExpression":
            const sizeOf = sizeOfResolver(expr.spec);
            if (sizeOf) {
                expr.value = sizeOf;
            }
            return;
        case "TypeCast":
            resolveConstantExpression(expr.operand, constResolver, sizeOfResolver, reportError);
            if (expr.operand.value !== undefined) {
                expr.value = applyTypeCast(expr.name, expr.operand.value);
            }
            return;
        case "UnaryExpression":
            resolveConstantExpression(expr.operand, constResolver, sizeOfResolver, reportError);
            if (expr.operand.value !== undefined) {
                if (expr.operator === "&") {
                    reportError("W105", expr);
                    return;
                }
                if (expr.operator === "~") {
                    const arg = toBits(expr.operand.value);
                    if (typeof arg === "number" && isNaN(arg)) {
                        reportError("W106", expr);
                        return;
                    }
                }
                expr.value = applyUnaryOperation(expr.operator, expr.operand.value);
            }
            return;
        case "BinaryExpression":
            resolveConstantExpression(expr.left, constResolver, sizeOfResolver, reportError);
            resolveConstantExpression(expr.right, constResolver, sizeOfResolver, reportError);
            if (expr.left.value !== undefined && expr.right.value !== undefined) {
                expr.value = applyBinaryOperation(expr.operator, expr.left.value, expr.right.value);
            }
            return;
    }
}
exports.resolveConstantExpression = resolveConstantExpression;
/**
 * Applies a type cast on the specified value
 * @param typeName Name of the type to cast the value
 * @param value Value to cast
 */
function applyTypeCast(typeName, value) {
    switch (typeName) {
        case "bool":
            return value === 0 ? 0 : 1;
        case "i8":
            return Number(BigInt.asIntN(8, BigInt(typeof value === "number" ? Math.trunc(value) : value)));
        case "u8":
            return Number(BigInt.asUintN(8, BigInt(typeof value === "number" ? Math.trunc(value) : value)));
        case "i16":
            return Number(BigInt.asIntN(16, BigInt(typeof value === "number" ? Math.trunc(value) : value)));
        case "u16":
            return Number(BigInt.asUintN(16, BigInt(typeof value === "number" ? Math.trunc(value) : value)));
        case "i32":
            return Number(BigInt.asIntN(32, BigInt(typeof value === "number" ? Math.trunc(value) : value)));
        case "u32":
            return Number(BigInt.asUintN(32, BigInt(typeof value === "number" ? Math.trunc(value) : value)));
        case "i64":
            return BigInt.asIntN(64, BigInt(value));
        case "u64":
            return BigInt.asUintN(64, BigInt(value));
        case "f32":
            return Math.fround(Number(value));
        case "f64":
            return value;
        default:
            throw new Error("Invalid type to cast a constant value");
    }
}
exports.applyTypeCast = applyTypeCast;
/**
 * Applies an unary operation to an operand
 * @param op Unary operation type
 * @param operand Operand value
 */
function applyUnaryOperation(op, operand) {
    switch (op) {
        case "!":
            return operand ? 0 : 1;
        case "-":
            return -operand;
        case "~":
            return ~toBits(operand);
    }
    return operand;
}
exports.applyUnaryOperation = applyUnaryOperation;
/**
 * Applies an unary operation to an operand
 * @param op Unary operation type
 * @param left Left operand value
 * @param right Right operand value
 */
function applyBinaryOperation(op, left, right) {
    switch (op) {
        case "+":
            return typeof left === "number" && typeof right === "number"
                ? left + right
                : BigInt(left) + BigInt(right);
        case "-":
            return typeof left === "number" && typeof right === "number"
                ? left - right
                : BigInt(left) - BigInt(right);
        case "*":
            return typeof left === "number" && typeof right === "number"
                ? left * right
                : BigInt(left) * BigInt(right);
        case "/":
            return typeof left === "number" && typeof right === "number"
                ? left / right
                : BigInt(left) / BigInt(right);
        case "%":
            return typeof left === "number" && typeof right === "number"
                ? left % right
                : BigInt(left) % BigInt(right);
        case "<<":
            return typeof left === "number" && typeof right === "number"
                ? left << right
                : BigInt(left) << BigInt(right);
        case ">>":
            return typeof left === "number" && typeof right === "number"
                ? left >> right
                : BigInt(left) >> BigInt(right);
        case ">>>":
            return typeof left === "number" && typeof right === "number"
                ? left >>> right
                : Number(left) >>> Number(right);
        case "&":
            if ((typeof left === "number" && !Number.isInteger(left)) ||
                (typeof right === "number" && !Number.isInteger(right))) {
                throw new Error("Bitwise bit operators must use integer operands");
            }
            return typeof left === "number" && typeof right === "number"
                ? left & right
                : BigInt(left) & BigInt(right);
        case "^":
            if ((typeof left === "number" && !Number.isInteger(left)) ||
                (typeof right === "number" && !Number.isInteger(right))) {
                throw new Error("Bitwise bit operators must use integer operands");
            }
            return typeof left === "number" && typeof right === "number"
                ? left ^ right
                : BigInt(left) ^ BigInt(right);
        case "|":
            if ((typeof left === "number" && !Number.isInteger(left)) ||
                (typeof right === "number" && !Number.isInteger(right))) {
                throw new Error("Bitwise bit operators must use integer operands");
            }
            return typeof left === "number" && typeof right === "number"
                ? left | right
                : BigInt(left) | BigInt(right);
        case "!=":
            return left !== right ? 1 : 0;
        case "<":
            return left < right ? 1 : 0;
        case "<=":
            return left <= right ? 1 : 0;
        case "==":
            return left === right ? 1 : 0;
        case ">":
            return left > right ? 1 : 0;
        case ">=":
            return left >= right ? 1 : 0;
    }
}
exports.applyBinaryOperation = applyBinaryOperation;
/**
 * Applies a built-in function
 * @param name Function name
 * @param args Function arguments
 */
function applyBuiltInFunction(name, args) {
    let arg;
    switch (name) {
        case "abs":
            arg = singleArg();
            return arg < 0 ? -arg : arg;
        case "ceil":
            arg = singleArg();
            return typeof arg === "number" ? Math.ceil(arg) : arg;
        case "floor":
            arg = singleArg();
            return typeof arg === "number" ? Math.floor(arg) : arg;
        case "max":
            atLeastOneArg();
            return bigIntMax(...args);
        case "min":
            atLeastOneArg();
            return bigIntMin(...args);
        case "nearest":
            arg = singleArg();
            if (typeof arg === "bigint")
                return arg;
            const ceil = Math.ceil(arg);
            const floor = Math.floor(arg);
            return Math.abs(ceil - arg) < Math.abs(floor - arg) ? ceil : floor;
        case "sqrt":
            arg = singleArg();
            if (arg < 0) {
                throw new Error("Argument of sqrt must be a non-zero number");
            }
            return typeof arg === "number" ? Math.sqrt(arg) : sqrtBigInt(arg);
        case "trunc":
            arg = singleArg();
            return typeof arg === "number" ? Math.trunc(arg) : arg;
    }
    throw new Error(`The '${name}' function cannot be used in constant expressions`);
    function singleArg() {
        if (args.length != 1) {
            throw new Error(`'${name}' must have a single argument`);
        }
        return args[0];
    }
    function atLeastOneArg() {
        if (args.length == 0) {
            throw new Error(`'${name}' must have at least one argument`);
        }
    }
}
exports.applyBuiltInFunction = applyBuiltInFunction;
/**
 * Converts the argument to 32-bit or 64-bit signed integer
 * @param arg Argument to convert
 * @returns NaN, if arg is a float; otherwise the 32-bit or 64-bit
 * signed representation of arg
 */
function toBits(arg) {
    if (typeof arg === "bigint") {
        return BigInt.asIntN(64, arg);
    }
    if (!Number.isInteger(arg)) {
        return NaN;
    }
    return Number(BigInt.asIntN(32, BigInt(arg)));
}
/**
 * Calculates the square root of a BigInt
 * @param arg Argument
 */
function sqrtBigInt(arg) {
    if (arg < 0) {
        throw new Error("Argument of sqrt must be a non-zero number");
    }
    if (arg < 2) {
        return arg;
    }
    function newtonIteration(n, x0) {
        const x1 = (n / x0 + x0) >> BigInt(1);
        if (x0 === x1 || x0 === x1 - BigInt(1)) {
            return x0;
        }
        return newtonIteration(n, x1);
    }
    return newtonIteration(arg, BigInt(1));
}
/**
 * Math.max alternative that works with number and BigInt
 * @param args
 */
function bigIntMax(...args) {
    return args.reduce((m, e) => (e > m ? e : m));
}
/**
 * Math.min alternative that works with number and BigInt
 * @param args
 */
function bigIntMin(...args) {
    return args.reduce((m, e) => (e < m ? e : m));
}
/**
 * Renders the specified expression
 * @param expr Expression to render
 */
function renderExpression(expr) {
    switch (expr.type) {
        case "Literal":
            return expr.value.toString();
        case "Identifier":
            return expr.name;
        case "BinaryExpression":
            return `(${renderExpression(expr.left)}${expr.operator}${renderExpression(expr.right)})`;
        case "FunctionInvocation":
        case "BuiltInFunctionInvocation":
            return `${expr.name}(${expr.arguments
                .map((arg) => renderExpression(arg))
                .join(",")})`;
        case "ConditionalExpression":
            return `(${renderExpression(expr.condition)}?${renderExpression(expr.consequent)}:${renderExpression(expr.alternate)})`;
        case "ItemAccess":
            return `(${renderExpression(expr.array)}[${renderExpression(expr.index)}])`;
        case "MemberAccess":
            return `(${renderExpression(expr.object)}.${expr.member})`;
        case "SizeOfExpression":
            return "sizeof()";
        case "TypeCast":
            return `${expr.name}(${renderExpression(expr.operand)})`;
        case "UnaryExpression":
            return `${expr.operator}(${renderExpression(expr.operand)})`;
    }
}
exports.renderExpression = renderExpression;


/***/ }),

/***/ "./src/compiler/source-tree.ts":
/*!*************************************!*\
  !*** ./src/compiler/source-tree.ts ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.instrisicSizes = exports.LiteralSource = void 0;
/**
 * Type of source literal
 */
var LiteralSource;
(function (LiteralSource) {
    LiteralSource[LiteralSource["Int"] = 0] = "Int";
    LiteralSource[LiteralSource["BigInt"] = 1] = "BigInt";
    LiteralSource[LiteralSource["Real"] = 2] = "Real";
})(LiteralSource = exports.LiteralSource || (exports.LiteralSource = {}));
/**
 * Hash object for the size of intrinsic types
 */
exports.instrisicSizes = {
    bool: 1,
    i8: 1,
    u8: 1,
    i16: 2,
    u16: 2,
    i32: 4,
    u32: 4,
    i64: 8,
    u64: 8,
    f32: 4,
    f64: 8,
};


/***/ }),

/***/ "./src/compiler/token-traits.ts":
/*!**************************************!*\
  !*** ./src/compiler/token-traits.ts ***!
  \**************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.getTokenTraits = void 0;
const tokens_1 = __webpack_require__(/*! ../core/tokens */ "./src/core/tokens.ts");
/**
 * Gets the traits of the specified token type
 * @param type Token type
 */
function getTokenTraits(type) {
    var _a;
    return (_a = tokenTraits.get(type)) !== null && _a !== void 0 ? _a : {};
}
exports.getTokenTraits = getTokenTraits;
/**
 * This map contains the traits of token types
 */
const tokenTraits = new Map();
// ----------------------------------------------------------------------------
// A
tokenTraits.set(tokens_1.TokenType.Abs, { expressionStart: true, builtInFunc: true });
tokenTraits.set(tokens_1.TokenType.AddAsgn, { assignmentOp: true });
tokenTraits.set(tokens_1.TokenType.Ampersand, { expressionStart: true, unaryOp: true });
tokenTraits.set(tokens_1.TokenType.AndAsgn, { assignmentOp: true });
tokenTraits.set(tokens_1.TokenType.Asgn, { assignmentOp: true });
tokenTraits.set(tokens_1.TokenType.Asterisk, {
    expressionStart: true,
    typeStart: true,
});
// ----------------------------------------------------------------------------
// B
tokenTraits.set(tokens_1.TokenType.BinaryLiteral, { expressionStart: true });
tokenTraits.set(tokens_1.TokenType.BinaryNot, { expressionStart: true, unaryOp: true });
tokenTraits.set(tokens_1.TokenType.Bool, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
// ----------------------------------------------------------------------------
// C
tokenTraits.set(tokens_1.TokenType.Ceil, { expressionStart: true, builtInFunc: true });
tokenTraits.set(tokens_1.TokenType.Clz, { expressionStart: true, builtInFunc: true });
tokenTraits.set(tokens_1.TokenType.CopyAsgn, { assignmentOp: true });
tokenTraits.set(tokens_1.TokenType.CopySign, {
    expressionStart: true,
    builtInFunc: true,
});
tokenTraits.set(tokens_1.TokenType.Ctz, { expressionStart: true, builtInFunc: true });
// ----------------------------------------------------------------------------
// D
tokenTraits.set(tokens_1.TokenType.DecimalLiteral, { expressionStart: true });
tokenTraits.set(tokens_1.TokenType.DivideAsgn, { assignmentOp: true });
// ----------------------------------------------------------------------------
// F
tokenTraits.set(tokens_1.TokenType.F32, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.F64, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.False, { expressionStart: true });
tokenTraits.set(tokens_1.TokenType.Floor, {
    expressionStart: true,
    builtInFunc: true,
    typeStart: true,
});
// ----------------------------------------------------------------------------
// H
tokenTraits.set(tokens_1.TokenType.HexadecimalLiteral, { expressionStart: true });
// ----------------------------------------------------------------------------
// I
tokenTraits.set(tokens_1.TokenType.I8, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.I16, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.I32, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.I64, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.Identifier, {
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.Infinity, { expressionStart: true });
// ----------------------------------------------------------------------------
// L
tokenTraits.set(tokens_1.TokenType.LParent, { expressionStart: true, typeStart: true });
// ----------------------------------------------------------------------------
// M
tokenTraits.set(tokens_1.TokenType.Max, { expressionStart: true, builtInFunc: true });
tokenTraits.set(tokens_1.TokenType.Min, { expressionStart: true, builtInFunc: true });
tokenTraits.set(tokens_1.TokenType.Minus, { expressionStart: true, unaryOp: true });
tokenTraits.set(tokens_1.TokenType.MultiplyAsgn, { assignmentOp: true });
// ----------------------------------------------------------------------------
// N
tokenTraits.set(tokens_1.TokenType.NaN, { expressionStart: true });
tokenTraits.set(tokens_1.TokenType.Neg, { expressionStart: true, builtInFunc: true });
tokenTraits.set(tokens_1.TokenType.Nearest, {
    expressionStart: true,
    builtInFunc: true,
});
tokenTraits.set(tokens_1.TokenType.Not, { expressionStart: true, unaryOp: true });
// ----------------------------------------------------------------------------
// O
tokenTraits.set(tokens_1.TokenType.OrAsgn, { assignmentOp: true });
// ----------------------------------------------------------------------------
// P
tokenTraits.set(tokens_1.TokenType.Plus, { expressionStart: true, unaryOp: true });
tokenTraits.set(tokens_1.TokenType.PopCnt, { expressionStart: true, builtInFunc: true });
// ----------------------------------------------------------------------------
// R
tokenTraits.set(tokens_1.TokenType.RealLiteral, { expressionStart: true });
tokenTraits.set(tokens_1.TokenType.RemainderAsgn, { assignmentOp: true });
// ----------------------------------------------------------------------------
// S
tokenTraits.set(tokens_1.TokenType.ShiftLeftAsgn, { assignmentOp: true });
tokenTraits.set(tokens_1.TokenType.ShiftRightAsgn, { assignmentOp: true });
tokenTraits.set(tokens_1.TokenType.SignedShiftRightAsgn, { assignmentOp: true });
tokenTraits.set(tokens_1.TokenType.Sizeof, { expressionStart: true });
tokenTraits.set(tokens_1.TokenType.Sqrt, { expressionStart: true, builtInFunc: true });
tokenTraits.set(tokens_1.TokenType.Struct, { typeStart: true });
tokenTraits.set(tokens_1.TokenType.SubtractAsgn, { assignmentOp: true });
// ----------------------------------------------------------------------------
// T
tokenTraits.set(tokens_1.TokenType.True, { expressionStart: true });
tokenTraits.set(tokens_1.TokenType.Trunc, { expressionStart: true, builtInFunc: true });
// ----------------------------------------------------------------------------
// U
tokenTraits.set(tokens_1.TokenType.U8, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.U16, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.U32, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
tokenTraits.set(tokens_1.TokenType.U64, {
    intrinsicType: true,
    expressionStart: true,
    typeStart: true,
});
// ----------------------------------------------------------------------------
// X
tokenTraits.set(tokens_1.TokenType.XorAsgn, { assignmentOp: true });


/***/ }),

/***/ "./src/compiler/wat-helpers.ts":
/*!*************************************!*\
  !*** ./src/compiler/wat-helpers.ts ***!
  \*************************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.countInstructions = exports.findInstruction = exports.instructionsActionLoop = exports.visitInstructions = void 0;
/**
 * Visits all instruction recursively to carry out the specified action
 * @param instrs Array of instructions to visit
 * @param action Action to execute on a visited instruction
 */
function visitInstructions(instrs, action, startIndex = 0) {
    for (let i = startIndex; i < instrs.length; i++) {
        const instr = instrs[i];
        action(instr, instrs, i);
        switch (instr.type) {
            case "If":
                visitInstructions(instr.consequtive, action);
                if (instr.alternate) {
                    visitInstructions(instr.alternate, action);
                }
                break;
            case "Block":
            case "Loop":
                visitInstructions(instr.body, action);
        }
    }
}
exports.visitInstructions = visitInstructions;
/**
 * Visits the instruction tree recursively to execute an action until there
 * are any changes
 * @param instrs Instructions
 * @param action Action to carry out
 */
function instructionsActionLoop(instrs, action) {
    let changeCount = 0;
    let changed;
    do {
        changed = false;
        for (let i = 0; i < instrs.length; i++) {
            changed = action(instrs, i);
            if (changed) {
                changeCount++;
                break;
            }
            const instr = instrs[i];
            switch (instr.type) {
                case "If":
                    changeCount += instructionsActionLoop(instr.consequtive, action);
                    if (instr.alternate) {
                        changeCount += instructionsActionLoop(instr.alternate, action);
                    }
                    break;
                case "Block":
                case "Loop":
                    changeCount += instructionsActionLoop(instr.body, action);
                    break;
            }
        }
    } while (changed);
    return changeCount;
}
exports.instructionsActionLoop = instructionsActionLoop;
/**
 * Tests if there is any instruction in the tree with a specific predicate
 * @param instrs Instrcutions to test recursively
 * @param predicate Predicate to test
 */
function findInstruction(instrs, predicate) {
    for (let i = 0; i < instrs.length; i++) {
        const instr = instrs[i];
        switch (instr.type) {
            case "If":
                let found = findInstruction(instr.consequtive, predicate);
                if (found) {
                    return true;
                }
                if (instr.alternate) {
                    found = findInstruction(instr.alternate, predicate);
                }
                if (found) {
                    return true;
                }
                break;
            case "Block":
            case "Loop": {
                let found = findInstruction(instr.body, predicate);
                if (found) {
                    return true;
                }
            }
            default:
                if (predicate(instr)) {
                    return true;
                }
                break;
        }
    }
    return false;
}
exports.findInstruction = findInstruction;
/**
 * Gets the number of WebAssembly instrcutions
 * @param instrs Instructions to count recursively
 */
function countInstructions(instrs) {
    let count = 0;
    for (let i = 0; i < instrs.length; i++) {
        count++;
        const instr = instrs[i];
        switch (instr.type) {
            case "If":
                count += countInstructions(instr.consequtive);
                if (instr.alternate) {
                    count += countInstructions(instr.alternate);
                }
                break;
            case "Block":
            case "Loop":
                count += countInstructions(instr.body);
        }
    }
    return count;
}
exports.countInstructions = countInstructions;


/***/ }),

/***/ "./src/compiler/wat-optimizer.ts":
/*!***************************************!*\
  !*** ./src/compiler/wat-optimizer.ts ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.optimizeLastInlineParam = exports.optimizeLocalUsage = exports.optimizeConstants = exports.optimizeWat = void 0;
const FunctionBuilder_1 = __webpack_require__(/*! ../wa-ast/FunctionBuilder */ "./src/wa-ast/FunctionBuilder.ts");
const wa_nodes_1 = __webpack_require__(/*! ../wa-ast/wa-nodes */ "./src/wa-ast/wa-nodes.ts");
const wat_helpers_1 = __webpack_require__(/*! ./wat-helpers */ "./src/compiler/wat-helpers.ts");
/**
 * Optimizes the specified set of instructions
 * @param instrs WA instructions to optimize
 */
function optimizeWat(instrs) {
    let changeCount;
    do {
        changeCount = 0;
        changeCount += removeDeadCode(instrs);
        changeCount += convertToBranchIf(instrs);
        changeCount += removeRedundantBranch(instrs);
        changeCount += reduceBranchIf(instrs);
        changeCount += optimizeConstantOperations(instrs);
        changeCount += reduceIntegerCasts(instrs);
        changeCount += optimizeLocalAccessors(instrs);
        changeCount += optimizeLocalTees(instrs);
        changeCount += optimizeStoreOps(instrs);
        changeCount += optimizeConstantDuplication(instrs);
        changeCount += optimizeEmptyLoop(instrs);
        changeCount += optimizeEmptyBlock(instrs);
        changeCount += peelLoop(instrs);
        changeCount += peelBlock(instrs);
    } while (changeCount);
}
exports.optimizeWat = optimizeWat;
/**
 * Optimizes constants
 * @param instrs WA instructions to optimize
 */
function optimizeConstants(instrs) {
    let changeCount;
    do {
        changeCount = optimizeConstantOperations(instrs);
    } while (changeCount);
}
exports.optimizeConstants = optimizeConstants;
/**
 * Optimizes local usages
 * @param funcBuilder Builder to optimize
 */
function optimizeLocalUsage(funcBuilder) {
    const localUsages = new Set();
    wat_helpers_1.visitInstructions(funcBuilder.body, (ins) => {
        if (ins.type.startsWith("Local")) {
            localUsages.add(ins.id);
        }
    });
    funcBuilder.locals = funcBuilder.locals.filter((l) => localUsages.has(l.id));
}
exports.optimizeLocalUsage = optimizeLocalUsage;
/**
 * Optimizes the usage of the last inline parameter
 * @param builderBody Body to optimize
 * @param paramName Name of the last inline parameter
 * @param startIndex Start index of the inline invocation
 */
function optimizeLastInlineParam(builderBody, paramName, startIndex) {
    // --- No inline parameter
    if (!paramName) {
        return;
    }
    // --- Let's count the getters for the last param
    const maxLength = builderBody.length;
    let getters = 0;
    let getterParent;
    let getterIndex = -1;
    wat_helpers_1.visitInstructions(builderBody, (ins, parent, index) => {
        if (ins.type === "LocalGet" && ins.id === paramName) {
            getters++;
            getterParent = parent;
            getterIndex = index;
        }
    }, startIndex);
    // --- Multiple getters, no optimization
    if (getters > 1 || getterIndex < 0 || !getterParent) {
        return;
    }
    // --- Check for replaceable pattern (constVal, localGet, globalGet)
    if ((builderBody[startIndex].type === "ConstVal" ||
        builderBody[startIndex].type === "LocalGet" ||
        builderBody[startIndex].type === "GlobalGet") &&
        startIndex + 1 < maxLength &&
        builderBody[startIndex + 1].type === "LocalSet" &&
        builderBody[startIndex + 1].id === paramName &&
        getterParent[getterIndex].id === paramName) {
        // --- Replace the getter with the constant
        getterParent[getterIndex] = builderBody[startIndex];
        // --- Remove the constant and the setter
        builderBody.splice(startIndex, 2);
    }
}
exports.optimizeLastInlineParam = optimizeLastInlineParam;
/**
 * Removes constant operations and replaces them with their equivalent
 * constant value
 * @param instrs WA instructions to optimize
 */
function optimizeConstantOperations(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        let changed = false;
        if (isConstant(ins, index)) {
            // --- We found a constant
            if (isUnary(ins, index + 1)) {
                // --- We can reduce an unary operation
                changed = reduceUnary(ins, index);
            }
            else if (isConstant(ins, index + 1) && isBinary(ins, index + 2)) {
                // --- We can reduce a binary operation
                changed = reduceBinary(ins, index);
            }
            else if (isBinary(ins, index + 1)) {
                if (isConstant(ins, index + 2) && isBinary(ins, index + 3)) {
                    changed = reduceCascadedBinary(ins, index);
                }
                else {
                    changed = reduceSecondConstOfBinary(ins, index);
                }
            }
            else if (isEqz(ins, index + 1) && isEqz(ins, index + 2)) {
                const value = ins[index];
                ins[index] = FunctionBuilder_1.constVal(value.valueType, value.value ? 1 : 0);
                ins.splice(index + 1, 2);
                changed = true;
            }
        }
        return changed;
    });
}
/**
 * Removed dead code from the specified block
 * @param instrs WA instructions to optimize
 * @param depth Current depth;
 */
function removeDeadCode(instrs, depth = 0) {
    let changeCount = 0;
    let retIndex = -1;
    for (let i = instrs.length - 1; i >= 0; i--) {
        const instr = instrs[i];
        switch (instr.type) {
            case "Return":
                retIndex = depth === 0 ? i : i + 1;
                break;
            case "Branch":
                retIndex = i + 1;
                break;
            case "If":
                changeCount += removeDeadCode(instr.consequtive, depth + 1);
                if (instr.alternate) {
                    changeCount += removeDeadCode(instr.alternate, depth + 1);
                }
                break;
            case "Block":
            case "Loop":
                changeCount += removeDeadCode(instr.body, depth + 1);
                break;
        }
    }
    if (retIndex >= 0) {
        const deleteCount = instrs.length - retIndex;
        instrs.splice(retIndex, deleteCount);
        if (deleteCount > 0) {
            changeCount++;
        }
    }
    return changeCount;
}
/**
 * Converts "if" instructions with a single "br" to "br_if"
 * @param instrs Instructions to convert
 */
function convertToBranchIf(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isIf(ins, index)) {
            const ifInstr = ins[index];
            if (ifInstr.consequtive.length === 1 &&
                !ifInstr.alternate &&
                ifInstr.consequtive[0].type === "Branch") {
                ins[index] = FunctionBuilder_1.branchIf(ifInstr.consequtive[0].label);
                return true;
            }
        }
        return false;
    });
}
/**
 * Reduces "const" and "br_if"
 * @param instrs Instructions to convert
 */
function reduceBranchIf(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isConstant(ins, index) && isBranchIf(ins, index + 1)) {
            const constVal = ins[index];
            const branchIf = ins[index + 1];
            if (constVal.value) {
                ins[index] = FunctionBuilder_1.branch(branchIf.label);
                ins.splice(index + 1, 1);
                return true;
            }
            else {
                ins.splice(index, 2);
            }
        }
        return false;
    });
}
/**
 * Removes redundant branch statements
 * @param instrs Instructions to convert
 */
function removeRedundantBranch(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if ((isBranch(ins, index) || isBranchIf(ins, index)) &&
            isBranch(ins, index + 1)) {
            const br1Instr = ins[index];
            const br2Instr = ins[index + 1];
            if (br1Instr.label === br2Instr.label) {
                ins[index] = FunctionBuilder_1.branch(br1Instr.label);
                ins.splice(index + 1, 1);
                return true;
            }
        }
        return false;
    });
}
/**
 * Changes "local_set" and "local_get" to "local_tee"
 * @param instrs Instructions to convert
 */
function optimizeLocalAccessors(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isLocalSet(ins, index) && isLocalGet(ins, index + 1)) {
            const localSet = ins[index];
            const localGet = ins[index + 1];
            if (localSet.id === localGet.id) {
                ins[index] = FunctionBuilder_1.localTee(localSet.id);
                ins.splice(index + 1, 1);
                return true;
            }
        }
        return false;
    });
}
/**
 * Removes single local_tee instructions
 * @param instrs Instructions to convert
 */
function optimizeLocalTees(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isLocalTee(ins, index)) {
            const localTee = ins[index];
            let teeCount = 0;
            wat_helpers_1.visitInstructions(ins, (it) => {
                if (it.type.startsWith("Local") &&
                    it.id === localTee.id) {
                    teeCount++;
                }
            });
            if (teeCount === 1) {
                ins.splice(index, 1);
                return true;
            }
        }
        return false;
    });
}
/**
 * Optimizes "load" and "store" operations by using an offset value
 * @param instrs Instructions to convert
 */
function optimizeStoreOps(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isConstant(ins, index) &&
            isAdd(ins, index + 1) &&
            (isLocalGet(ins, index + 2) || isGlobalGet(ins, index + 2)) &&
            (isStore(ins, index + 3))) {
            const offset = ins[index].value;
            const storeOp = ins[index + 3];
            storeOp.offset = Number(offset);
            ins.splice(index, 2);
            return true;
        }
        return false;
    });
}
/**
 * Changes "const", "local_tee", and "local_get" to two "const" instructions
 * @param instrs Instructions to convert
 */
function optimizeConstantDuplication(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isConstant(ins, index) &&
            isLocalTee(ins, index + 1) &&
            isLocalGet(ins, index + 2)) {
            const constantToDupl = ins[index];
            const localTee = ins[index + 1];
            const localGet = ins[index + 2];
            if (localTee.id === localGet.id) {
                ins[index + 1] = FunctionBuilder_1.constVal(constantToDupl.valueType, constantToDupl.value);
                ins.splice(index + 2, 1);
                return true;
            }
        }
        return false;
    });
}
/**
 * Reduces "const" and "br_if"
 * @param instrs Instructions to convert
 */
function reduceIntegerCasts(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isConstant(ins, index) &&
            isAnd(ins, index + 1) &&
            isStore(ins, index + 2)) {
            const constVal = ins[index];
            const store = ins[index + 2];
            if ((constVal.value === 0xffff && store.bits === wa_nodes_1.WaBitSpec.Bit16) ||
                (constVal.value === 0xff && store.bits === wa_nodes_1.WaBitSpec.Bit8)) {
                ins.splice(index, 2);
                return true;
            }
        }
        return false;
    });
}
/**
 * Optimizes empty and branch-only loops
 * @param instrs Instructions to optimize
 */
function optimizeEmptyLoop(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        // --- Check for loops
        if (isLoop(ins, index)) {
            const loop = ins[index];
            const body = loop.body;
            if (body.length === 0) {
                // --- Empty loop
                ins.splice(index, 1);
                return true;
            }
            if (body.length === 1 &&
                body[0].type === "Branch" &&
                body[0].label !== loop.id) {
                // --- Branch-only loop
                ins[index] = FunctionBuilder_1.branch(body[0].label);
                return true;
            }
            if (body.length === 1 &&
                body[0].type === "BranchIf" &&
                body[0].label !== loop.id) {
                // --- Branch-only loop
                ins[index] = FunctionBuilder_1.branchIf(body[0].label);
                return true;
            }
        }
        return false;
    });
}
/**
 * Optimizes empty and branch-only loops
 * @param instrs Instructions to optimize
 */
function optimizeEmptyBlock(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        // --- Check for block
        if (isBlock(ins, index)) {
            const block = ins[index];
            const body = block.body;
            if (body.length === 0) {
                // --- Empty loop
                ins.splice(index, 1);
                return true;
            }
            if (body.length === 1 &&
                body[0].type === "Branch" &&
                body[0].label === block.id) {
                // --- Branch-only block
                ins.splice(index, 1);
                return true;
            }
            if (body.length === 1 &&
                body[0].type === "BranchIf" &&
                body[0].label === block.id) {
                // --- Branch-only block
                ins.splice(index, 1);
                return true;
            }
        }
        return false;
    });
}
/**
 * Peels a loop that does not have a branch to itself
 * @param instrs Instructions to optimize
 */
function peelLoop(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isLoop(ins, index)) {
            const loop = ins[index];
            if (!wat_helpers_1.findInstruction(loop.body, (item) => (item.type === "Branch" || item.type === "BranchIf") &&
                item.label === loop.id)) {
                ins.splice(index, 1, ...loop.body);
                return true;
            }
        }
        return false;
    });
}
/**
 * Peels a block that has only branches to itself
 * @param instrs Instructions to optimize
 */
function peelBlock(instrs) {
    return wat_helpers_1.instructionsActionLoop(instrs, (ins, index) => {
        if (isBlock(ins, index)) {
            const block = ins[index];
            const reducedBody = [];
            let hasBranchOut = false;
            block.body.forEach((item) => {
                if (item.type === "Branch") {
                    hasBranchOut || (hasBranchOut = item.label !== block.id);
                }
                else {
                    reducedBody.push(item);
                    if (item.type === "Loop" || item.type === "Block") {
                        hasBranchOut || (hasBranchOut = wat_helpers_1.findInstruction(item.body, (it) => (it.type === "Branch" || it.type === "BranchIf") &&
                            it.label === block.id));
                    }
                    else if (item.type === "If") {
                        hasBranchOut || (hasBranchOut = wat_helpers_1.findInstruction(item.consequtive, (it) => (it.type === "Branch" || it.type === "BranchIf") &&
                            it.label === block.id));
                        if (item.alternate) {
                            hasBranchOut || (hasBranchOut = wat_helpers_1.findInstruction(item.alternate, (it) => (it.type === "Branch" || it.type === "BranchIf") &&
                                it.label === block.id));
                        }
                    }
                }
            });
            if (!hasBranchOut &&
                !block.body.some((item) => item.type === "BranchIf")) {
                ins.splice(index, 1, ...reducedBody);
                return true;
            }
        }
        return false;
    });
}
/**
 * Reduces the unary operation at the specified index
 * @param instrs Instructions
 * @param index Constant value index (followed by the unary op)
 * @returns true, if the operation has been reduced
 */
function reduceUnary(instrs, index) {
    const operand = instrs[index];
    const operandValue = operand.value;
    const waType = operand.valueType;
    const op = instrs[index + 1];
    let newInstr;
    switch (op.type) {
        case "Extend32":
            if (waType === wa_nodes_1.WaType.i32) {
                newInstr = FunctionBuilder_1.constVal(wa_nodes_1.WaType.i64, operandValue);
            }
            break;
        case "Demote64":
            if (waType === wa_nodes_1.WaType.f64) {
                newInstr = FunctionBuilder_1.constVal(wa_nodes_1.WaType.f32, operandValue);
            }
            break;
    }
    if (newInstr) {
        instrs[index] = newInstr;
        instrs.splice(index + 1, 1);
        return true;
    }
    return false;
}
/**
 * Reduces the binary operation at the specified index
 * @param instrs Instructions
 * @param index Constant value index (followed by the second operand and the binary op)
 * @returns true, if the operation has been reduced
 */
function reduceBinary(instrs, index) {
    const leftOp = instrs[index];
    const left = leftOp.value;
    const waType = leftOp.valueType;
    const right = instrs[index + 1].value;
    const op = instrs[index + 2];
    let value = null;
    switch (op.type) {
        case "Mul":
            value =
                typeof left === "number" && typeof right === "number"
                    ? left * right
                    : BigInt(left) * BigInt(right);
            break;
        case "Add":
            value =
                typeof left === "number" && typeof right === "number"
                    ? left + right
                    : BigInt(left) + BigInt(right);
            break;
        case "And":
            value =
                typeof left === "number" && typeof right === "number"
                    ? left & right
                    : BigInt(left) & BigInt(right);
            break;
        case "Or":
            value =
                typeof left === "number" && typeof right === "number"
                    ? left | right
                    : BigInt(left) | BigInt(right);
            break;
        case "Xor":
            value =
                typeof left === "number" && typeof right === "number"
                    ? left ^ right
                    : BigInt(left) ^ BigInt(right);
            break;
        case "Shl":
            value =
                typeof left === "number" && typeof right === "number"
                    ? left << right
                    : BigInt(left) << BigInt(right);
            break;
        case "Shr":
            if (op.signed) {
                value =
                    typeof left === "number" && typeof right === "number"
                        ? left >> right
                        : BigInt(left) >> BigInt(right);
            }
            else {
                value =
                    typeof left === "number" && typeof right === "number"
                        ? left >>> right
                        : Number(left) >>> Number(right);
            }
            break;
    }
    if (value !== null) {
        instrs[index] = FunctionBuilder_1.constVal(waType, value);
        instrs.splice(index + 1, 2);
        return true;
    }
    return false;
}
/**
 * Reduces the binary operation at the specified index
 * @param instrs Instructions
 * @param index Constant value index (const, binary, const, binary)
 * @returns true, if the operation has been reduced
 */
function reduceCascadedBinary(instrs, index) {
    const leftOp = instrs[index];
    const left = leftOp.value;
    const waType = leftOp.valueType;
    const right = instrs[index + 2].value;
    const opsType = instrs[index + 1].type + instrs[index + 3].type;
    let value = null;
    let updatedOp;
    switch (opsType) {
        case "AddAdd":
            value =
                typeof left === "number" && typeof right === "number"
                    ? left + right
                    : BigInt(left) + BigInt(right);
            updatedOp = FunctionBuilder_1.add(waType);
            break;
        case "SubSub":
            value =
                typeof left === "number" && typeof right === "number"
                    ? left + right
                    : BigInt(left) + BigInt(right);
            updatedOp = FunctionBuilder_1.sub(waType);
            break;
    }
    if (value !== null) {
        instrs[index] = FunctionBuilder_1.constVal(waType, value);
        instrs[index + 1] = updatedOp;
        instrs.splice(index + 2, 2);
        return true;
    }
    return false;
}
/**
 * Reduces the second constant of a binary operation
 * @param instrs Instructions
 * @param index Constant value index (const, binary)
 * @returns true, if the operation has been reduced
 */
function reduceSecondConstOfBinary(instrs, index) {
    const operand = instrs[index];
    const opType = instrs[index + 1].type;
    switch (opType) {
        case "Add":
        case "Sub":
            if (operand.value === 0) {
                instrs.splice(index, 2);
                return true;
            }
            break;
        case "Mul":
        case "Div":
            if (operand.value === 1) {
                instrs.splice(index, 2);
                return true;
            }
            break;
    }
    return false;
}
/**
 * Tests if the specified instruction is a constant
 * @param index Instruction index in the function body
 */
function isConstant(instrs, index) {
    return (index >= 0 &&
        index < instrs.length &&
        instrs[index].type === "ConstVal");
}
/**
 * Tests if the specified instruction is a unary operation
 * @param index Instruction index in the function body
 */
function isUnary(instrs, index) {
    return (index >= 0 &&
        index < instrs.length &&
        instructionTraits[instrs[index].type] === InstructionType.Unary);
}
/**
 * Tests if the specified instruction is a binary operation
 * @param index Instruction index in the function body
 */
function isBinary(instrs, index) {
    return (index >= 0 &&
        index < instrs.length &&
        instructionTraits[instrs[index].type] === InstructionType.Binary);
}
/**
 * Tests if the specified instruction is a constant
 * @param index Instruction index in the function body
 */
function isAdd(instrs, index) {
    return (index >= 0 &&
        index < instrs.length &&
        instrs[index].type === "Add");
}
/**
 * Tests if the specified instruction is "eqz"
 * @param index Instruction index in the function body
 */
function isEqz(instrs, index) {
    return index >= 0 && index < instrs.length && instrs[index].type === "Eqz";
}
/**
 * Tests if the specified instruction is an "if" operation
 * @param index Instruction index in the function body
 */
function isIf(instrs, index) {
    return index >= 0 && index < instrs.length && instrs[index].type === "If";
}
/**
 * Tests if the specified instruction is a "br" operation
 * @param index Instruction index in the function body
 */
function isBranch(instrs, index) {
    return index >= 0 && index < instrs.length && instrs[index].type === "Branch";
}
/**
 * Tests if the specified instruction is a "br" operation
 * @param index Instruction index in the function body
 */
function isBranchIf(instrs, index) {
    return (index >= 0 && index < instrs.length && instrs[index].type === "BranchIf");
}
/**
 * Tests if the specified instruction is a "local_get" operation
 * @param index Instruction index in the function body
 */
function isLocalGet(instrs, index) {
    return (index >= 0 && index < instrs.length && instrs[index].type === "LocalGet");
}
/**
 * Tests if the specified instruction is a "local_set" operation
 * @param index Instruction index in the function body
 */
function isLocalSet(instrs, index) {
    return (index >= 0 && index < instrs.length && instrs[index].type === "LocalSet");
}
/**
 * Tests if the specified instruction is a "local_tee" operation
 * @param index Instruction index in the function body
 */
function isLocalTee(instrs, index) {
    return (index >= 0 && index < instrs.length && instrs[index].type === "LocalTee");
}
/**
 * Tests if the specified instruction is a "global_get" operation
 * @param index Instruction index in the function body
 */
function isGlobalGet(instrs, index) {
    return (index >= 0 && index < instrs.length && instrs[index].type === "GlobalGet");
}
/**
 * Tests if the specified instruction is a "block" operation
 * @param index Instruction index in the function body
 */
function isBlock(instrs, index) {
    return index >= 0 && index < instrs.length && instrs[index].type === "Block";
}
/**
 * Tests if the specified instruction is a "loop" operation
 * @param index Instruction index in the function body
 */
function isLoop(instrs, index) {
    return index >= 0 && index < instrs.length && instrs[index].type === "Loop";
}
/**
 * Tests if the specified instruction is a "store" operation
 * @param index Instruction index in the function body
 */
function isStore(instrs, index) {
    return index >= 0 && index < instrs.length && instrs[index].type === "Store";
}
/**
 * Tests if the specified instruction is a "store" operation
 * @param index Instruction index in the function body
 */
function isLoad(instrs, index) {
    return index >= 0 && index < instrs.length && instrs[index].type === "Load";
}
/**
 * Tests if the specified instruction is an "and" operation
 * @param index Instruction index in the function body
 */
function isAnd(instrs, index) {
    return index >= 0 && index < instrs.length && instrs[index].type === "And";
}
/**
 * Type of a particular instruction
 */
var InstructionType;
(function (InstructionType) {
    InstructionType[InstructionType["None"] = 0] = "None";
    InstructionType[InstructionType["Const"] = 1] = "Const";
    InstructionType[InstructionType["Unary"] = 2] = "Unary";
    InstructionType[InstructionType["Binary"] = 3] = "Binary";
})(InstructionType || (InstructionType = {}));
const instructionTraits = {
    Abs: InstructionType.Unary,
    Add: InstructionType.Binary,
    And: InstructionType.Binary,
    Block: InstructionType.None,
    Branch: InstructionType.None,
    BranchIf: InstructionType.None,
    BranchTable: InstructionType.None,
    Call: InstructionType.None,
    CallIndirect: InstructionType.None,
    Ceil: InstructionType.None,
    Clz: InstructionType.None,
    Comment: InstructionType.None,
    ConstVal: InstructionType.Const,
    Convert32: InstructionType.None,
    Convert64: InstructionType.None,
    CopySign: InstructionType.None,
    Ctz: InstructionType.None,
    Demote64: InstructionType.Unary,
    Div: InstructionType.Binary,
    Drop: InstructionType.None,
    Eq: InstructionType.Binary,
    Eqz: InstructionType.None,
    Extend32: InstructionType.Unary,
    Floor: InstructionType.None,
    Ge: InstructionType.Binary,
    GlobalGet: InstructionType.None,
    GlobalSet: InstructionType.None,
    Gt: InstructionType.Binary,
    If: InstructionType.None,
    Le: InstructionType.Binary,
    Load: InstructionType.None,
    LocalGet: InstructionType.None,
    LocalSet: InstructionType.None,
    LocalTee: InstructionType.None,
    Loop: InstructionType.None,
    Lt: InstructionType.Binary,
    Max: InstructionType.Binary,
    Min: InstructionType.Binary,
    MemoryGrow: InstructionType.None,
    MemorySize: InstructionType.None,
    Mul: InstructionType.Binary,
    Ne: InstructionType.Binary,
    Nearest: InstructionType.None,
    Neg: InstructionType.None,
    Nop: InstructionType.None,
    Or: InstructionType.Binary,
    PopCnt: InstructionType.None,
    Promote32: InstructionType.None,
    ReinterpretF32: InstructionType.None,
    ReinterpretF64: InstructionType.None,
    ReinterpretI32: InstructionType.None,
    ReinterpretI64: InstructionType.None,
    Rem: InstructionType.Binary,
    Return: InstructionType.None,
    Rotl: InstructionType.None,
    Rotr: InstructionType.None,
    Select: InstructionType.None,
    SeparatorLine: InstructionType.None,
    Shl: InstructionType.Binary,
    Shr: InstructionType.Binary,
    Sqrt: InstructionType.None,
    Store: InstructionType.None,
    Sub: InstructionType.Binary,
    Trunc: InstructionType.None,
    Trunc32: InstructionType.None,
    Trunc64: InstructionType.None,
    Unreachable: InstructionType.None,
    Wrap64: InstructionType.None,
    Xor: InstructionType.Binary,
};


/***/ }),

/***/ "./src/core/InputStream.ts":
/*!*********************************!*\
  !*** ./src/core/InputStream.ts ***!
  \*********************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.InputStream = void 0;
/**
 * This class represents an input stream that takes characters from a
 * source code string
 */
class InputStream {
    /**
     * Creates a stream that uses the specified source code
     * @param chunkOrSourceCode The input to use with this stream
     */
    constructor(chunkOrSource) {
        // --- Current source string position
        this._pos = 0;
        if (typeof chunkOrSource === "string") {
            this._fileIndex = 0;
            this._posOffset = 0;
            this._line = 1;
            this._column = 0;
            this._source = chunkOrSource;
        }
        else {
            this._fileIndex = chunkOrSource.fileIndex;
            this._posOffset = chunkOrSource.pos;
            this._line = chunkOrSource.line;
            this._column = chunkOrSource.col;
            this._source = chunkOrSource.sourceCode;
        }
    }
    /**
     * Gets the source string of this stream
     */
    get source() {
        return this._source;
    }
    /**
     * Gets the specified part of the source code
     * @param start Start position
     * @param end End position
     */
    getSourceSpan(start, end) {
        return this._source.substring(start, end);
    }
    /**
     * Gets the file index of the stream
     */
    get fileIndex() {
        return this._fileIndex;
    }
    /**
     * Gets the current position in the stream. Starts from 0.
     */
    get position() {
        return this._pos + this._posOffset;
    }
    /**
     * Gets the current line number. Starts from 1.
     */
    get line() {
        return this._line;
    }
    /**
     * Gets the current column number. Starts from0.
     */
    get column() {
        return this._column;
    }
    /**
     * Peeks the next character in the stream
     * @returns null, if EOF; otherwise the current source code character
     */
    peek() {
        return this._pos > this._source.length - 1 ? null : this._source[this._pos];
    }
    /**
     * Gets the next character from the stream
     */
    get() {
        // --- Check for EOF
        if (this._pos >= this._source.length) {
            return null;
        }
        // --- Get the char, and keep track of position
        const ch = this._source[this._pos++];
        if (ch === "\n") {
            this._line++;
            this._column = 0;
        }
        else {
            this._column++;
        }
        return ch;
    }
}
exports.InputStream = InputStream;


/***/ }),

/***/ "./src/core/MultiChunkInputStream.ts":
/*!*******************************************!*\
  !*** ./src/core/MultiChunkInputStream.ts ***!
  \*******************************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.MultiChunkInputStream = void 0;
/**
 * Implements an input stream that uses multiple source code chunks
 */
class MultiChunkInputStream {
    /**
     * Creates a stream that uses the specified source code chunks
     * @param chunks Chunks to use
     */
    constructor(chunks) {
        // --- The current chunk to read from
        this._nextChunkIndex = 0;
        // --- Current source string position
        this._pos = 0;
        // --- The current file index
        this._fileIndex = 0;
        // --- Position offset
        this._posOffset = 0;
        // --- Current line number
        this._line = 0;
        // --- Current column number
        this._column = 0;
        if (typeof chunks === "string") {
            chunks = [
                {
                    sourceCode: chunks,
                    fileIndex: 0,
                    pos: 0,
                    line: 1,
                    col: 0,
                },
            ];
        }
        this._chunks = chunks;
        this._nextChunkIndex = 0;
        this._source = "";
        this.moveToNext();
    }
    /**
     * Gets the file index of the stream
     */
    get fileIndex() {
        return this._fileIndex;
    }
    /**
     * Gets the current position in the stream. Starts from 0.
     */
    get position() {
        return this._pos + this._posOffset;
    }
    /**
     * Gets the current line number. Starts from 1.
     */
    get line() {
        return this._line;
    }
    /**
     * Gets the current column number. Starts from0.
     */
    get column() {
        return this._column;
    }
    /**
     * Peeks the next character in the stream
     * @returns null, if EOF; otherwise the current source code character
     */
    peek() {
        return this._pos > this._source.length - 1 ? null : this._source[this._pos];
    }
    /**
     * Gets the next character from the stream
     */
    get() {
        // --- Check for EOF
        if (this.moveToNext()) {
            return null;
        }
        // --- Get the char, and keep track of position
        const ch = this._source[this._pos++];
        if (ch === "\n") {
            this._line++;
            this._column = 0;
        }
        else {
            this._column++;
        }
        this.moveToNext();
        return ch;
    }
    /**
     * Moves to the next chunk
     * @returns True, if end-of-chunks; otherwise, false
     */
    moveToNext() {
        if (this._pos >= this._source.length) {
            // --- Move to the next chunk
            if (this._nextChunkIndex >= this._chunks.length) {
                // --- No more chunks
                return true;
            }
            // --- Load the next chunk
            const chunk = this._chunks[this._nextChunkIndex++];
            this._source = chunk.sourceCode;
            this._fileIndex = chunk.fileIndex;
            this._pos = 0;
            this._posOffset = chunk.pos;
            this._line = chunk.line;
            this._column = chunk.col;
        }
        return false;
    }
}
exports.MultiChunkInputStream = MultiChunkInputStream;


/***/ }),

/***/ "./src/core/errors.ts":
/*!****************************!*\
  !*** ./src/core/errors.ts ***!
  \****************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.errorMessages = exports.ParserError = void 0;
/**
 * The common root class of all parser error objects
 */
class ParserError extends Error {
    /**
     * Instantiates a custom parser error object
     * @param message Error message
     * @param code Optional code to identify the message
     */
    constructor(message, code) {
        super(message);
        this.code = code;
        // --- Set the prototype explicitly.
        Object.setPrototypeOf(this, ParserError.prototype);
    }
}
exports.ParserError = ParserError;
/**
 * The error messages of error codes
 */
exports.errorMessages = {
    // --- Preprocessor errors
    P001: "Unknown source code token found during preprocessing: {0}",
    P002: "Cannot find preprocessor directive",
    P003: "Unknown preprocessor directive: {0}",
    P004: "Symbol identifier expected",
    P005: "Invalid preprocessor directive completion",
    P006: "Unexpected token: {0}",
    P007: "An expression expected",
    P008: "Missing ')'",
    P009: "Cannot use #elseif without a correspondig #if",
    P010: "Cannot use #elseif after #else",
    P011: "Cannot use #else without a correspondig #if",
    P012: "The current #if already has an #else",
    P013: "Cannot use #endif without a correspondig #if",
    P014: "Missing #endif",
    P015: "String literal expected",
    P016: "#include error: {0}",
    // --- WAT# syntax issues
    W001: "A comma expected",
    W002: "An expression expected",
    W003: "Unexpected token: {0}",
    W004: "An identifier expected",
    W005: "An intrinsic type specifier expected but {0} received.",
    W006: "Missing semicolon closing the declaration",
    W007: "Expected '='",
    W008: "Type specification expected",
    W009: "'{' expected",
    W010: "'}' expected",
    W011: "'[' expected",
    W012: "']' expected",
    W013: "Data declaration accepts only integral types",
    W014: "Import function declaration must have a result type or 'void'",
    W015: "Import functions can have only intrinsic result and parameter types",
    W016: "'(' expected",
    W017: "')' expected",
    W018: "Unexpected token in variable/function declaration",
    W019: "Function parameters can be intrinsic types or pointers only",
    W020: "Function result types can be intrinsic types, pointers, or 'void'",
    W021: "Missing function parameter name",
    W022: "'while' expected",
    W023: "A statement expected",
    W024: "An assignment operator expected",
    W025: "'*' or an identifier expected",
    // --- WAT# semantic issues
    W100: "Duplicated declaration identifier: '{0}'",
    W101: "Unknown declaration: '{0}'",
    W102: "'{0}' is not a type declaration",
    W103: "Cannot resolve '{0}' because of circular declaration reference",
    W104: "Invalid construct in constant expressions",
    W105: "The '&' and '*' operators cannot be used in constant expressions",
    W106: "The '~' operators cannot be used with a float value",
    W107: "Error when evaluating an expression: {0}",
    W108: "'{0}' is not a const declaration",
    W109: "A table can hold only function identifiers",
    W110: "The break statement can only be used in a loop",
    W111: "The continue statement can only be used in a loop",
    W112: "Invalid left value",
    // --- WAT# Compiler issues
    W140: "Local parameter/variable '{0}' has already been declared",
    W141: "Cannot cast the right side of the assignment to the target type",
    W142: "Cannot resolve name '{0}' to a local or a global variable declaration",
    W143: "Only variables with intrinsic and pointer types can be used",
    W144: "The {0} operator works only with instrinsic types",
    W145: "The {0} operator works only with integer types",
    W146: "Only variables have addresses, globals and locals do not",
    W147: "The left operand of the member access operator must be a struct",
    W148: "Unknown struct member: {0}",
    W149: "The item access operator can be used only with arrays",
    W150: "The {0} built-in function works only with floating-point types",
    W151: "The {0} built-in function works only with integer types",
    W152: "Non-pointer expressions cannot be dereferenced",
    W153: "Cannot resolve name '{0}' to a function declaration",
    W154: "The function has {0} argument(s) but it is called with {1}",
    W155: "Cannot convert floating-point arguments to integer function parameters implicitly",
    W156: "A function with void result type should return no value",
    W157: "A function with non-void result type should return a value",
    W158: "Function signature of '{0}' does not match table signature",
    W159: "A table function invocation requires a dispatcher expression",
    W160: "Only a table function invocation can have a dispatcher expression",
    W161: "Left values can be only intrinsic types or pointers",
    W162: "You can use only variable names for aliases",
    W163: "Variable {0} has not got an address yet, it cannot be an alias",
    W164: "Invalid pointer operation",
    W165: "Pointer operations '+' and '-' require an integer right operand",
    W166: "Copy assignment needs works only with array and struct types",
    W167: "Copy assignment requires an integer expression on its right side",
    W168: "Copy assignment does not support copying more than 256 bytes",
    W169: "Data declaration identifier cannot be used on the left side of an assignment",
    W170: "Expression cannot be an operand of item, member, or dereference access",
    W171: "Relational operators with a pointer and the left require a pointer on the right",
};


/***/ }),

/***/ "./src/core/tokens.ts":
/*!****************************!*\
  !*** ./src/core/tokens.ts ***!
  \****************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.isRestrictedInString = exports.isHexadecimalDigit = exports.isDecimalDigit = exports.isBinaryDigit = exports.isIdContinuation = exports.isIdStart = exports.isWs = exports.isEof = exports.TokenType = void 0;
/**
 * This enumeration defines the token types
 */
var TokenType;
(function (TokenType) {
    TokenType[TokenType["Eof"] = -1] = "Eof";
    TokenType[TokenType["Ws"] = -2] = "Ws";
    TokenType[TokenType["BlockComment"] = -3] = "BlockComment";
    TokenType[TokenType["EolComment"] = -4] = "EolComment";
    TokenType[TokenType["Unknown"] = 0] = "Unknown";
    // --- Only the preprocessor uses these tokens
    TokenType[TokenType["SourceChunk"] = 1] = "SourceChunk";
    TokenType[TokenType["PreprocDirective"] = 2] = "PreprocDirective";
    TokenType[TokenType["PPIdentifier"] = 3] = "PPIdentifier";
    TokenType[TokenType["PPStringLiteral"] = 4] = "PPStringLiteral";
    TokenType[TokenType["NewLine"] = 5] = "NewLine";
    TokenType[TokenType["OrOp"] = 6] = "OrOp";
    TokenType[TokenType["XorOp"] = 7] = "XorOp";
    TokenType[TokenType["AndOp"] = 8] = "AndOp";
    TokenType[TokenType["NotOp"] = 9] = "NotOp";
    // --- Both the preprocessor and WA# uses these tokens
    TokenType[TokenType["LParent"] = 10] = "LParent";
    TokenType[TokenType["RParent"] = 11] = "RParent";
    // --- WA#-specific tokens
    TokenType[TokenType["Identifier"] = 12] = "Identifier";
    TokenType[TokenType["Divide"] = 13] = "Divide";
    TokenType[TokenType["DivideAsgn"] = 14] = "DivideAsgn";
    TokenType[TokenType["Asterisk"] = 15] = "Asterisk";
    TokenType[TokenType["MultiplyAsgn"] = 16] = "MultiplyAsgn";
    TokenType[TokenType["Remainder"] = 17] = "Remainder";
    TokenType[TokenType["RemainderAsgn"] = 18] = "RemainderAsgn";
    TokenType[TokenType["Plus"] = 19] = "Plus";
    TokenType[TokenType["AddAsgn"] = 20] = "AddAsgn";
    TokenType[TokenType["Minus"] = 21] = "Minus";
    TokenType[TokenType["SubtractAsgn"] = 22] = "SubtractAsgn";
    TokenType[TokenType["Xor"] = 23] = "Xor";
    TokenType[TokenType["XorAsgn"] = 24] = "XorAsgn";
    TokenType[TokenType["Or"] = 25] = "Or";
    TokenType[TokenType["OrAsgn"] = 26] = "OrAsgn";
    TokenType[TokenType["Ampersand"] = 27] = "Ampersand";
    TokenType[TokenType["AndAsgn"] = 28] = "AndAsgn";
    TokenType[TokenType["CopyAsgn"] = 29] = "CopyAsgn";
    TokenType[TokenType["Semicolon"] = 30] = "Semicolon";
    TokenType[TokenType["Comma"] = 31] = "Comma";
    TokenType[TokenType["Colon"] = 32] = "Colon";
    TokenType[TokenType["LSquare"] = 33] = "LSquare";
    TokenType[TokenType["RSquare"] = 34] = "RSquare";
    TokenType[TokenType["QuestionMark"] = 35] = "QuestionMark";
    TokenType[TokenType["BinaryNot"] = 36] = "BinaryNot";
    TokenType[TokenType["LBrace"] = 37] = "LBrace";
    TokenType[TokenType["RBrace"] = 38] = "RBrace";
    TokenType[TokenType["Asgn"] = 39] = "Asgn";
    TokenType[TokenType["Equal"] = 40] = "Equal";
    TokenType[TokenType["Not"] = 41] = "Not";
    TokenType[TokenType["NotEqual"] = 42] = "NotEqual";
    TokenType[TokenType["LessThan"] = 43] = "LessThan";
    TokenType[TokenType["LessThanOrEqual"] = 44] = "LessThanOrEqual";
    TokenType[TokenType["ShiftLeft"] = 45] = "ShiftLeft";
    TokenType[TokenType["ShiftLeftAsgn"] = 46] = "ShiftLeftAsgn";
    TokenType[TokenType["GreaterThan"] = 47] = "GreaterThan";
    TokenType[TokenType["GreaterThanOrEqual"] = 48] = "GreaterThanOrEqual";
    TokenType[TokenType["ShiftRight"] = 49] = "ShiftRight";
    TokenType[TokenType["ShiftRightAsgn"] = 50] = "ShiftRightAsgn";
    TokenType[TokenType["SignedShiftRight"] = 51] = "SignedShiftRight";
    TokenType[TokenType["SignedShiftRightAsgn"] = 52] = "SignedShiftRightAsgn";
    TokenType[TokenType["Dot"] = 53] = "Dot";
    TokenType[TokenType["Sizeof"] = 54] = "Sizeof";
    TokenType[TokenType["Bool"] = 55] = "Bool";
    TokenType[TokenType["I8"] = 56] = "I8";
    TokenType[TokenType["U8"] = 57] = "U8";
    TokenType[TokenType["I16"] = 58] = "I16";
    TokenType[TokenType["U16"] = 59] = "U16";
    TokenType[TokenType["I32"] = 60] = "I32";
    TokenType[TokenType["U32"] = 61] = "U32";
    TokenType[TokenType["I64"] = 62] = "I64";
    TokenType[TokenType["U64"] = 63] = "U64";
    TokenType[TokenType["F32"] = 64] = "F32";
    TokenType[TokenType["F64"] = 65] = "F64";
    TokenType[TokenType["Void"] = 66] = "Void";
    TokenType[TokenType["DecimalLiteral"] = 67] = "DecimalLiteral";
    TokenType[TokenType["HexadecimalLiteral"] = 68] = "HexadecimalLiteral";
    TokenType[TokenType["BinaryLiteral"] = 69] = "BinaryLiteral";
    TokenType[TokenType["RealLiteral"] = 70] = "RealLiteral";
    TokenType[TokenType["StringLiteral"] = 71] = "StringLiteral";
    TokenType[TokenType["Infinity"] = 72] = "Infinity";
    TokenType[TokenType["NaN"] = 73] = "NaN";
    TokenType[TokenType["True"] = 74] = "True";
    TokenType[TokenType["False"] = 75] = "False";
    TokenType[TokenType["Inline"] = 76] = "Inline";
    TokenType[TokenType["Export"] = 77] = "Export";
    TokenType[TokenType["Type"] = 78] = "Type";
    TokenType[TokenType["Struct"] = 79] = "Struct";
    TokenType[TokenType["Table"] = 80] = "Table";
    TokenType[TokenType["Data"] = 81] = "Data";
    TokenType[TokenType["Import"] = 82] = "Import";
    TokenType[TokenType["Clz"] = 83] = "Clz";
    TokenType[TokenType["Ctz"] = 84] = "Ctz";
    TokenType[TokenType["PopCnt"] = 85] = "PopCnt";
    TokenType[TokenType["Abs"] = 86] = "Abs";
    TokenType[TokenType["Ceil"] = 87] = "Ceil";
    TokenType[TokenType["Floor"] = 88] = "Floor";
    TokenType[TokenType["Trunc"] = 89] = "Trunc";
    TokenType[TokenType["Nearest"] = 90] = "Nearest";
    TokenType[TokenType["Sqrt"] = 91] = "Sqrt";
    TokenType[TokenType["Min"] = 92] = "Min";
    TokenType[TokenType["Max"] = 93] = "Max";
    TokenType[TokenType["Neg"] = 94] = "Neg";
    TokenType[TokenType["CopySign"] = 95] = "CopySign";
    TokenType[TokenType["Const"] = 96] = "Const";
    TokenType[TokenType["Global"] = 97] = "Global";
    TokenType[TokenType["Local"] = 98] = "Local";
    TokenType[TokenType["If"] = 99] = "If";
    TokenType[TokenType["Else"] = 100] = "Else";
    TokenType[TokenType["Do"] = 101] = "Do";
    TokenType[TokenType["While"] = 102] = "While";
    TokenType[TokenType["Break"] = 103] = "Break";
    TokenType[TokenType["Continue"] = 104] = "Continue";
    TokenType[TokenType["Return"] = 105] = "Return";
})(TokenType = exports.TokenType || (exports.TokenType = {}));
/**
 * Tests if a token id EOF
 * @param t Token instance
 */
function isEof(t) {
    return t.type === TokenType.Eof;
}
exports.isEof = isEof;
/**
 * Tests if a token is whitespace
 * @param t Token instance
 */
function isWs(t) {
    return t.type <= TokenType.Ws;
}
exports.isWs = isWs;
/**
 * Tests if a character is an identifier start character
 * @param ch Character to test
 */
function isIdStart(ch) {
    return ((ch >= "a" && ch <= "z") ||
        (ch >= "A" && ch <= "Z") ||
        ch === "_" ||
        ch === "$");
}
exports.isIdStart = isIdStart;
/**
 * Tests if a character is an identifier continuation character
 * @param ch Character to test
 */
function isIdContinuation(ch) {
    return ((ch >= "a" && ch <= "z") ||
        (ch >= "A" && ch <= "Z") ||
        (ch >= "0" && ch <= "9") ||
        ch === "_" ||
        ch === "$");
}
exports.isIdContinuation = isIdContinuation;
/**
 * Tests if a character is a binary digit
 * @param ch Character to test
 */
function isBinaryDigit(ch) {
    return ch === "0" || ch === "1";
}
exports.isBinaryDigit = isBinaryDigit;
/**
 * Tests if a character is a decimal digit
 * @param ch Character to test
 */
function isDecimalDigit(ch) {
    return ch >= "0" && ch <= "9";
}
exports.isDecimalDigit = isDecimalDigit;
/**
 * Tests if a character is a hexadecimal digit
 * @param ch Character to test
 */
function isHexadecimalDigit(ch) {
    return ((ch >= "0" && ch <= "9") ||
        (ch >= "A" && ch <= "F") ||
        (ch >= "a" && ch <= "f"));
}
exports.isHexadecimalDigit = isHexadecimalDigit;
/**
 * Tests if a character is restricted in a string
 * @param ch Character to test
 */
function isRestrictedInString(ch) {
    return (ch === "\r" ||
        ch === "\n" ||
        ch === "\u0085" ||
        ch === "\u2028" ||
        ch === "\u2029");
}
exports.isRestrictedInString = isRestrictedInString;


/***/ }),

/***/ "./src/preprocessor/PreprocessorExpressionLexer.ts":
/*!*********************************************************!*\
  !*** ./src/preprocessor/PreprocessorExpressionLexer.ts ***!
  \*********************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.PreprocessorExpressionLexer = void 0;
const tokens_1 = __webpack_require__(/*! ../core/tokens */ "./src/core/tokens.ts");
/**
 * This class implements the tokenizer (lexer) of WA# preprocessor expresion handler
 */
class PreprocessorExpressionLexer {
    /**
     * Initializes the tokenizer with the input stream
     * @param input Input source code stream
     */
    constructor(input) {
        this.input = input;
        // --- Already fetched tokens
        this._ahead = [];
        // --- Prefetched character (from the next token)
        this._prefetched = null;
        // --- Prefetched character position (from the next token)
        this._prefetchedPos = null;
        // --- Prefetched character column (from the next token)
        this._prefetchedColumn = null;
    }
    /**
     * Fethches the next token without advancing to its position
     * @param ws If true, retrieve whitespaces too
     */
    peek(ws = false) {
        return this.ahead(0, ws);
    }
    /**
     *
     * @param n Number of token positions to read ahead
     * @param ws If true, retrieve whitespaces too
     */
    ahead(n = 1, ws = false) {
        if (n > 16) {
            throw new Error("Cannot look ahead more than 16 tokens");
        }
        // --- Prefetch missing tokens
        while (this._ahead.length <= n) {
            const token = this.fetch();
            if (tokens_1.isEof(token)) {
                return token;
            }
            if (ws || (!ws && !tokens_1.isWs(token))) {
                this._ahead.push(token);
            }
        }
        return this._ahead[n];
    }
    /**
     * Fethces the next token and advances the stream position
     * @param ws If true, retrieve whitespaces too
     */
    get(ws = false) {
        if (this._ahead.length > 0) {
            const token = this._ahead.shift();
            if (!token) {
                throw new Error("Token expected");
            }
            return token;
        }
        while (true) {
            const token = this.fetch();
            if (tokens_1.isEof(token) || ws || (!ws && !tokens_1.isWs(token))) {
                return token;
            }
        }
    }
    /**
     * Resets the parser; drops read-ahead tokens.
     */
    reset() {
        this._ahead = [];
    }
    /**
     * Fetches the next token from the input stream
     */
    fetch() {
        const lexer = this;
        const input = this.input;
        const startPos = this._prefetchedPos || input.position;
        const startLine = input.line;
        const startColumn = this._prefetchedColumn || input.column;
        let text = "";
        let tokenType = tokens_1.TokenType.Eof;
        let lastEndPos = input.position;
        let lastEndColumn = input.column;
        let ch = null;
        let phase = LexerPhase.Start;
        while (true) {
            // --- Get the next character
            ch = fetchNextChar();
            // --- In case of EOF, return the current token data
            if (ch === null) {
                return makeToken();
            }
            // --- Set the intial token type to unknown for the other characters
            if (tokenType === tokens_1.TokenType.Eof) {
                tokenType = tokens_1.TokenType.Unknown;
            }
            // --- Follow the lexer state machine
            switch (phase) {
                // ====================================================================
                // We are at the beginning of a new line
                case LexerPhase.Start:
                    // --- We start a source code chunk
                    switch (ch) {
                        // --- Go on with whitespaces
                        case " ":
                        case "\t":
                        case "\r":
                            tokenType = tokens_1.TokenType.Ws;
                            phase = LexerPhase.InWhiteSpace;
                            break;
                        // --- End of the expression
                        case "\n":
                            return completeToken(tokens_1.TokenType.NewLine);
                        // --- Single character tokens
                        case "(":
                            return completeToken(tokens_1.TokenType.LParent);
                        case ")":
                            return completeToken(tokens_1.TokenType.RParent);
                        case "!":
                            return completeToken(tokens_1.TokenType.NotOp);
                        case "^":
                            return completeToken(tokens_1.TokenType.XorOp);
                        case "|":
                            return completeToken(tokens_1.TokenType.OrOp);
                        case "&":
                            return completeToken(tokens_1.TokenType.AndOp);
                        // --- Identifier start
                        default:
                            if ((ch >= "a" && ch <= "z") ||
                                (ch >= "A" && ch <= "Z") ||
                                ch === "_") {
                                // --- This is a valid identifier start character
                                phase = LexerPhase.IdTail;
                                tokenType = tokens_1.TokenType.PPIdentifier;
                                break;
                            }
                            // --- Unknown token
                            return makeToken();
                    }
                    break;
                // --- Collect all whitespace characters
                case LexerPhase.InWhiteSpace:
                    switch (ch) {
                        case " ":
                        case "\t":
                        case "\r":
                            break;
                        default:
                            return makeToken();
                    }
                    break;
                // --- Wait for the completion of an ID
                case LexerPhase.IdTail:
                    if (!tokens_1.isIdContinuation(ch)) {
                        return makeToken();
                    }
                    break;
                // ====================================================================
                // --- We cannot continue
                default:
                    return makeToken();
            }
            // --- Append the char to the current text
            appendTokenChar();
            // --- Go on with parsing the next character
        }
        /**
         * Appends the last character to the token, and manages positions
         */
        function appendTokenChar() {
            text += ch;
            lexer._prefetched = null;
            lexer._prefetchedPos = null;
            lexer._prefetchedColumn = null;
            lastEndPos = input.position;
            lastEndColumn = input.position;
        }
        /**
         * Fetches the next character from the input stream
         */
        function fetchNextChar() {
            let ch;
            if (!lexer._prefetched) {
                lexer._prefetchedPos = input.position;
                lexer._prefetchedColumn = input.column;
                lexer._prefetched = input.get();
            }
            return lexer._prefetched;
        }
        /**
         * Packs the specified type of token to send back
         * @param type
         */
        function makeToken() {
            return {
                text,
                type: tokenType,
                location: {
                    fileIndex: input.fileIndex,
                    startPosition: startPos,
                    endPosition: lastEndPos,
                    startLine,
                    endLine: input.line,
                    startColumn,
                    endColumn: lastEndColumn,
                },
            };
        }
        /**
         * Add the last character to the token and return it
         */
        function completeToken(suggestedType) {
            appendTokenChar();
            // --- Send back the token
            if (suggestedType !== undefined) {
                tokenType = suggestedType;
            }
            return makeToken();
        }
    }
}
exports.PreprocessorExpressionLexer = PreprocessorExpressionLexer;
/**
 * Represents the current lexer phase
 */
var LexerPhase;
(function (LexerPhase) {
    LexerPhase[LexerPhase["Start"] = 0] = "Start";
    LexerPhase[LexerPhase["InWhiteSpace"] = 1] = "InWhiteSpace";
    LexerPhase[LexerPhase["IdTail"] = 2] = "IdTail";
})(LexerPhase || (LexerPhase = {}));


/***/ }),

/***/ "./src/preprocessor/PreprocessorLexer.ts":
/*!***********************************************!*\
  !*** ./src/preprocessor/PreprocessorLexer.ts ***!
  \***********************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.PreprocessorLexer = void 0;
const tokens_1 = __webpack_require__(/*! ../core/tokens */ "./src/core/tokens.ts");
/**
 * This class implements the tokenizer (lexer) of the WA# preprocessor directives
 */
class PreprocessorLexer {
    /**
     * Initializes the tokenizer with the input stream
     * @param input Input source code stream
     */
    constructor(input) {
        this.input = input;
        // --- Already fetched tokens
        this._ahead = [];
        // --- Prefetched character (from the next token)
        this._prefetched = null;
        // --- Prefetched character position (from the next token)
        this._prefetchedPos = null;
        // --- Prefetched character column (from the next token)
        this._prefetchedColumn = null;
    }
    /**
     * Fethches the next token without advancing to its position
     * @param ws If true, retrieve whitespaces too
     */
    peek(ws = false) {
        return this.ahead(0, ws);
    }
    /**
     *
     * @param n Number of token positions to read ahead
     * @param ws If true, retrieve whitespaces too
     */
    ahead(n = 1, ws = false) {
        if (n > 16) {
            throw new Error("Cannot look ahead more than 16 tokens");
        }
        // --- Prefetch missing tokens
        while (this._ahead.length <= n) {
            const token = this.fetch();
            if (tokens_1.isEof(token)) {
                return token;
            }
            if (ws || (!ws && !tokens_1.isWs(token))) {
                this._ahead.push(token);
            }
        }
        return this._ahead[n];
    }
    /**
     * Fethces the next token and advances the stream position
     * @param ws If true, retrieve whitespaces too
     */
    get(ws = false) {
        if (this._ahead.length > 0) {
            const token = this._ahead.shift();
            if (!token) {
                throw new Error("Token expected");
            }
            return token;
        }
        while (true) {
            const token = this.fetch();
            if (tokens_1.isEof(token) || ws || (!ws && !tokens_1.isWs(token))) {
                return token;
            }
        }
    }
    /**
     * Fetches the next token from the input stream
     */
    fetch() {
        const lexer = this;
        const input = this.input;
        const startPos = this._prefetchedPos || input.position;
        const startLine = input.line;
        const startColumn = this._prefetchedColumn || input.column;
        let text = "";
        let tokenType = tokens_1.TokenType.Eof;
        let lastEndPos = input.position;
        let lastEndColumn = input.column;
        let ch = null;
        let phase = LexerPhase.LineStart;
        while (true) {
            // --- Get the next character
            ch = fetchNextChar();
            // --- In case of EOF, return the current token data
            if (ch === null) {
                return makeToken();
            }
            // --- Set the intial token type to unknown for the other characters
            if (tokenType === tokens_1.TokenType.Eof) {
                tokenType = tokens_1.TokenType.Unknown;
            }
            // --- Follow the lexer state machine
            switch (phase) {
                // ====================================================================
                // We are at the beginning of a new line
                case LexerPhase.LineStart:
                    // --- We start a source code chunk
                    tokenType = tokens_1.TokenType.SourceChunk;
                    switch (ch) {
                        // --- Go on with whitespaces
                        case " ":
                        case "\t":
                        case "\r":
                            phase = LexerPhase.InWhiteSpace;
                            break;
                        // --- We stay in this phase with a new line
                        case "\n":
                            break;
                        // --- May be the beginning of a comment
                        case "/":
                            phase = LexerPhase.CommentStart;
                            break;
                        // --- May be the beginning of a preprocessor directive
                        case "#":
                            phase = LexerPhase.PreprocStart;
                            break;
                        // --- This line may not be a preprocessor directive
                        default:
                            phase = LexerPhase.NoPreproc;
                            break;
                    }
                    break;
                // --- Collect all whitespace characters
                case LexerPhase.InWhiteSpace:
                    switch (ch) {
                        case " ":
                        case "\t":
                        case "\r":
                            break;
                        // --- We stay in this phase with a new line
                        case "\n":
                            phase = LexerPhase.LineStart;
                            break;
                        // --- May be the beginning of a comment
                        case "/":
                            phase = LexerPhase.CommentStart;
                            break;
                        // --- May be the beginning of a preprocessor directive
                        case "#":
                            phase = LexerPhase.PreprocStart;
                            break;
                        default:
                            phase = LexerPhase.NoPreproc;
                            break;
                    }
                    break;
                // --- May be the beginning of a comment
                case LexerPhase.CommentStart:
                    switch (ch) {
                        // --- Starts an end-of-line comment
                        case "/":
                            phase = LexerPhase.InlineCommentTrail;
                            break;
                        // --- Starts a block comment
                        case "*":
                            phase = LexerPhase.BlockCommentTrail1;
                            break;
                    }
                    break;
                // --- Wait for an "*" that may complete a block comment
                case LexerPhase.BlockCommentTrail1:
                    if (ch === "*") {
                        phase = LexerPhase.BlockCommentTrail2;
                    }
                    break;
                // --- Wait for a "/" that may complete a block comment
                case LexerPhase.BlockCommentTrail2:
                    phase =
                        ch === "/" ? LexerPhase.NoPreproc : LexerPhase.BlockCommentTrail1;
                    break;
                case LexerPhase.InlineCommentTrail:
                    if (ch === "\n") {
                        phase = LexerPhase.LineStart;
                    }
                    break;
                // --- May be the start of a preprocessor directive
                case LexerPhase.PreprocStart:
                    if ((ch >= "a" && ch <= "z") ||
                        (ch >= "A" && ch <= "Z") ||
                        ch === "_") {
                        phase = LexerPhase.PreprocIdStart;
                        tokenType = tokens_1.TokenType.PreprocDirective;
                    }
                    else {
                        phase = LexerPhase.NoPreproc;
                    }
                    break;
                case LexerPhase.PreprocIdStart:
                    if (!tokens_1.isIdContinuation(ch)) {
                        tokenType = tokens_1.TokenType.PreprocDirective;
                        return makeToken();
                    }
                // --- Whatever comes, it cannot be a preprocessor directive
                case LexerPhase.NoPreproc:
                    switch (ch) {
                        // --- Maybe a comment
                        case "/":
                            phase = LexerPhase.CommentStart;
                            break;
                        // --- A new line, which may contain a directive
                        case "\n":
                            phase = LexerPhase.LineStart;
                    }
                    break;
                // ====================================================================
                // --- We cannot continue
                default:
                    return makeToken();
            }
            // --- Append the char to the current text
            appendTokenChar();
            // --- Go on with parsing the next character
        }
        /**
         * Appends the last character to the token, and manages positions
         */
        function appendTokenChar() {
            text += ch;
            lexer._prefetched = null;
            lexer._prefetchedPos = null;
            lexer._prefetchedColumn = null;
            lastEndPos = input.position;
            lastEndColumn = input.position;
        }
        /**
         * Fetches the next character from the input stream
         */
        function fetchNextChar() {
            let ch;
            if (!lexer._prefetched) {
                lexer._prefetchedPos = input.position;
                lexer._prefetchedColumn = input.column;
                lexer._prefetched = input.get();
            }
            return lexer._prefetched;
        }
        /**
         * Packs the specified type of token to send back
         * @param type
         */
        function makeToken() {
            return {
                text,
                type: tokenType,
                location: {
                    fileIndex: input.fileIndex,
                    startPosition: startPos,
                    endPosition: lastEndPos,
                    startLine,
                    endLine: input.line,
                    startColumn,
                    endColumn: lastEndColumn,
                },
            };
        }
    }
}
exports.PreprocessorLexer = PreprocessorLexer;
/**
 * Represents the current lexer phase
 */
var LexerPhase;
(function (LexerPhase) {
    LexerPhase[LexerPhase["LineStart"] = 0] = "LineStart";
    LexerPhase[LexerPhase["InWhiteSpace"] = 1] = "InWhiteSpace";
    LexerPhase[LexerPhase["NoPreproc"] = 2] = "NoPreproc";
    LexerPhase[LexerPhase["CommentStart"] = 3] = "CommentStart";
    LexerPhase[LexerPhase["BlockCommentTrail1"] = 4] = "BlockCommentTrail1";
    LexerPhase[LexerPhase["BlockCommentTrail2"] = 5] = "BlockCommentTrail2";
    LexerPhase[LexerPhase["InlineCommentTrail"] = 6] = "InlineCommentTrail";
    LexerPhase[LexerPhase["PreprocStart"] = 7] = "PreprocStart";
    LexerPhase[LexerPhase["PreprocIdStart"] = 8] = "PreprocIdStart";
})(LexerPhase || (LexerPhase = {}));


/***/ }),

/***/ "./src/preprocessor/PreprocessorParser.ts":
/*!************************************************!*\
  !*** ./src/preprocessor/PreprocessorParser.ts ***!
  \************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.PreprocessorParser = void 0;
const errors_1 = __webpack_require__(/*! ../core/errors */ "./src/core/errors.ts");
const InputStream_1 = __webpack_require__(/*! ../core/InputStream */ "./src/core/InputStream.ts");
const preprocessor_string_1 = __webpack_require__(/*! ./preprocessor-string */ "./src/preprocessor/preprocessor-string.ts");
const PreprocessorExpressionLexer_1 = __webpack_require__(/*! ./PreprocessorExpressionLexer */ "./src/preprocessor/PreprocessorExpressionLexer.ts");
const PreprocessorLexer_1 = __webpack_require__(/*! ./PreprocessorLexer */ "./src/preprocessor/PreprocessorLexer.ts");
const PreprocessorStringLiteralLexer_1 = __webpack_require__(/*! ./PreprocessorStringLiteralLexer */ "./src/preprocessor/PreprocessorStringLiteralLexer.ts");
const tokens_1 = __webpack_require__(/*! ../core/tokens */ "./src/core/tokens.ts");
/**
 * This class implements the WA# preprocessor
 */
class PreprocessorParser {
    /**
     * Instantiates a parser for the preprocessor
     * @param source Source code to parse
     * @param fileIndex File index information
     * @param includeHandler Optional handler managin include files
     */
    constructor(source, fileIndex = 0, includeHandler) {
        this.source = source;
        this.fileIndex = fileIndex;
        this.includeHandler = includeHandler;
        // --- Keep track of error messages
        this._parseErrors = [];
        // --- Preprocessor symbols already defined
        this._preprocessorSymbols = {};
        // --- Stack of #if .. #elseif .. #else .. #endif constructs
        this._ifStatusStack = [];
        this._inputStream = new InputStream_1.InputStream({
            fileIndex,
            sourceCode: source,
            pos: 0,
            line: 1,
            col: 0,
        });
        this._lexer = new PreprocessorLexer_1.PreprocessorLexer(this._inputStream);
        this._exprLexer = new PreprocessorExpressionLexer_1.PreprocessorExpressionLexer(this._inputStream);
        this._stringLexer = new PreprocessorStringLiteralLexer_1.PreprocessorStringLiteralLexer(this._inputStream);
    }
    /**
     * Tests if the specified symbol is defined in the context
     * @param symbolId Symbol identifier
     */
    idSymbolDefined(symbolId) {
        return this._preprocessorSymbols[symbolId];
    }
    /**
     * Gets the container of defined preprocessor symbols
     */
    get preprocessorSymbols() {
        return this._preprocessorSymbols;
    }
    /**
     * The errors raised during the parse phase
     */
    get errors() {
        return this._parseErrors;
    }
    /**
     * Indicates if there were any errors during the parse phase
     */
    get hasErrors() {
        return this._parseErrors.length > 0;
    }
    /**
     * Preprocesses the source code and retrives the source code chunks
     * found
     */
    preprocessSource() {
        const sourceChunks = [];
        let token;
        while ((token = this._lexer.get()).type !== tokens_1.TokenType.Eof) {
            // --- Obtain the next chunk of source code
            if (token.type === tokens_1.TokenType.SourceChunk) {
                const trimmedText = token.text.trim();
                if (trimmedText.length > 0) {
                    const loc = token.location;
                    sourceChunks.push({
                        sourceCode: trimmedText,
                        fileIndex: loc.fileIndex,
                        pos: loc.startPosition,
                        line: loc.startLine,
                        col: loc.startColumn,
                    });
                }
            }
            else if (token.type === tokens_1.TokenType.PreprocDirective) {
                this.parseDirective(sourceChunks, token);
            }
            else {
                // --- Any other token type must be an error
                this.reportError("P001", token, tokens_1.TokenType[token.type]);
                return null;
            }
        }
        if (this._ifStatusStack.length > 0) {
            this.reportError("P014");
        }
        return sourceChunks;
    }
    /**
     * Parses an expression from the input stream
     */
    parseExpression() {
        const token = this._exprLexer.peek();
        if (token.type === tokens_1.TokenType.Eof) {
            this.reportError("P007");
            return null;
        }
        return this.parseOrExpression();
    }
    /**
     * Parses a string literal from the input stream
     */
    parseStringLiteral() {
        const token = this._stringLexer.peek();
        this._stringLexer.reset();
        if (token.type !== tokens_1.TokenType.PPStringLiteral) {
            this.reportError("P015");
            return null;
        }
        return token.text;
    }
    /**
     * Parses and evaluares the expression
     */
    evalExpression() {
        const parser = this;
        const expr = this.parseExpression();
        this._exprLexer.reset();
        return evalInner(expr);
        function evalInner(expr) {
            switch (expr.type) {
                case "Id":
                    return !!parser._preprocessorSymbols[expr.name];
                case "NotExpr":
                    return !evalInner(expr.operand);
                case "BinaryExpr":
                    const left = evalInner(expr.leftOperand);
                    const right = evalInner(expr.rightOperand);
                    switch (expr.operator) {
                        case "^":
                            return left !== right;
                        case "|":
                            return left || right;
                        case "&":
                            return left && right;
                    }
            }
        }
    }
    // ==========================================================================
    // Helpers
    /**
     * Parse a preprocessor directive
     * @param directive
     */
    parseDirective(sourceChunks, token) {
        // --- The lexer found a directive
        const ppPos = token.text.lastIndexOf("#");
        if (ppPos < 0) {
            // --- Cannot extract directive name, it must be an error
            this.reportError("P002");
            return;
        }
        // --- By default, add source chunks
        let addSourceChunk = true;
        // --- Parse the directive found
        const directive = token.text.substr(ppPos);
        switch (directive) {
            case "#define":
                addSourceChunk = this.processDefine();
                break;
            case "#undef":
                addSourceChunk = this.processUndef();
                break;
            case "#if":
                addSourceChunk = this.processIf();
                break;
            case "#elseif":
                addSourceChunk = this.processElseIf();
                break;
            case "#else":
                addSourceChunk = this.processElse();
                break;
            case "#endif":
                addSourceChunk = this.processEndIf();
                break;
            case "#include":
                this.processInclude(sourceChunks, token);
                addSourceChunk = false;
                break;
            default:
                this.reportError("P003", null, directive);
                return null;
        }
        if (addSourceChunk) {
            if (ppPos > 0) {
                const trimmedText = token.text.substr(0, ppPos).trim();
                if (trimmedText.length > 0) {
                    const loc = token.location;
                    sourceChunks.push({
                        sourceCode: trimmedText,
                        fileIndex: loc.fileIndex,
                        pos: loc.startPosition,
                        line: loc.startLine,
                        col: loc.startColumn,
                    });
                }
            }
        }
    }
    /**
     * Processes the #define directive
     */
    processDefine() {
        const idToken = this._exprLexer.get();
        if (idToken.type !== tokens_1.TokenType.PPIdentifier) {
            this.reportError("P004");
            return;
        }
        this.expectEnd();
        // --- Should skip this directive?
        const ifStatus = this._ifStatusStack[this._ifStatusStack.length - 1];
        if (ifStatus && (ifStatus.skip || !ifStatus.satisfied)) {
            return false;
        }
        // --- Define the symbol
        this._preprocessorSymbols[idToken.text] = true;
        // --- Ensure that a previously satisifed condition branch is injected
        if (ifStatus === null || ifStatus === void 0 ? void 0 : ifStatus.satisfied) {
            ifStatus.skip = true;
            if (ifStatus.injected) {
                return false;
            }
            else {
                ifStatus.injected = true;
            }
        }
        // --- Add the chunk
        return true;
    }
    /**
     * Processes the #under directive
     */
    processUndef() {
        const idToken = this._exprLexer.get();
        if (idToken.type !== tokens_1.TokenType.PPIdentifier) {
            this.reportError("P004");
            return;
        }
        this.expectEnd();
        // --- Should skip this directive?
        const ifStatus = this._ifStatusStack[this._ifStatusStack.length - 1];
        if (ifStatus && (ifStatus.skip || !ifStatus.satisfied)) {
            return false;
        }
        // --- Define the symbol
        delete this._preprocessorSymbols[idToken.text];
        // --- Ensure that a previously satisifed condition branch is injected
        if (ifStatus === null || ifStatus === void 0 ? void 0 : ifStatus.satisfied) {
            ifStatus.skip = true;
            if (ifStatus.injected) {
                return false;
            }
            else {
                ifStatus.injected = true;
            }
        }
        // --- Add the chunk
        return true;
    }
    /**
     * Processes the #if directive
     */
    processIf() {
        if (this._ifStatusStack.length === 0) {
            const satisfied = this.evalExpression();
            this._ifStatusStack.push({
                satisfied,
                elseReached: false,
                injected: false,
                skip: false,
            });
            return true;
        }
        else {
            const ifStatus = this._ifStatusStack[this._ifStatusStack.length - 1];
            const satisfied = this.evalExpression();
            const skip = ifStatus.skip || !ifStatus.satisfied;
            this._ifStatusStack.push({
                satisfied: skip ? false : satisfied,
                elseReached: false,
                injected: false,
                skip,
            });
            return !skip;
        }
    }
    /**
     * Processes the #elseif directive
     */
    processElseIf() {
        // --- Check if #elseif is valid here
        if (this._ifStatusStack.length === 0) {
            this.reportError("P009");
            return;
        }
        const ifStatus = this._ifStatusStack[this._ifStatusStack.length - 1];
        if (ifStatus.elseReached) {
            this.reportError("P010");
            return;
        }
        // --- Evaluate the new condition
        const satisfied = this.evalExpression();
        // --- Should skip this branch?
        if (ifStatus.skip) {
            return false;
        }
        // --- Ensure that a previously satisifed condition branch is injected
        if (ifStatus.satisfied) {
            ifStatus.skip = true;
            if (ifStatus.injected) {
                return false;
            }
            else {
                ifStatus.injected = true;
                return true;
            }
        }
        // --- Store the result of evaluation
        ifStatus.satisfied = satisfied;
        // --- Do not inject the source chunk before the #elseif
        return false;
    }
    /**
     * Processes the #else directive
     */
    processElse() {
        // --- Check if #else is valid here
        if (this._ifStatusStack.length === 0) {
            this.reportError("P011");
            return;
        }
        const ifStatus = this._ifStatusStack[this._ifStatusStack.length - 1];
        if (ifStatus.elseReached) {
            this.reportError("P012");
            return;
        }
        // --- Should skip this branch?
        if (ifStatus.skip) {
            return false;
        }
        // --- Ensure that a previously satisifed condition branch is injected
        if (ifStatus.satisfied) {
            ifStatus.skip = true;
            if (ifStatus.injected) {
                return false;
            }
            else {
                ifStatus.injected = true;
                return true;
            }
        }
        // --- The #else branch satisfies the condition
        ifStatus.satisfied = true;
        // --- Do not inject the source chunk before the #else
        return false;
    }
    /**
     * Processes the #endif directive
     */
    processEndIf() {
        // --- Check if #endif is valid here
        if (this._ifStatusStack.length === 0) {
            this.reportError("P013");
            return;
        }
        // --- Remove the #if frame
        const ifStatus = this._ifStatusStack.pop();
        // --- Ensure that a previously satisfied condition branch is injected
        return ifStatus.satisfied && !ifStatus.injected;
    }
    /**
     * Processes the include directive
     */
    processInclude(sourceChunks, token) {
        const ppString = this.parseStringLiteral();
        this.expectEnd();
        // --- Check if this include should be processed
        const ifStatus = this._ifStatusStack[this._ifStatusStack.length - 1];
        // --- Should skip this branch?
        if (ifStatus && (ifStatus.skip || !ifStatus.satisfied)) {
            return;
        }
        // --- Ensure that the webchunk before #include is processed
        const ppPos = token.text.lastIndexOf("#");
        if (ppPos < 0) {
            // --- Cannot extract directive name, it must be an error
            this.reportError("P002");
            return;
        }
        if (ppPos > 0) {
            const trimmedText = token.text.substr(0, ppPos).trim();
            if (trimmedText.length > 0) {
                const loc = token.location;
                sourceChunks.push({
                    sourceCode: trimmedText,
                    fileIndex: loc.fileIndex,
                    pos: loc.startPosition,
                    line: loc.startLine,
                    col: loc.startColumn,
                });
            }
        }
        // --- Process the include file
        const filename = preprocessor_string_1.convertPPStringLiteralToRawString(ppString);
        if (this.includeHandler) {
            let sourceInfo;
            try {
                sourceInfo = this.includeHandler(this.fileIndex, filename);
            }
            catch (err) {
                this.reportError("P016", null, err.toString());
            }
            const includeParser = new PreprocessorParser(sourceInfo.source, sourceInfo.fileIndex, this.includeHandler);
            includeParser._preprocessorSymbols = this._preprocessorSymbols;
            const chunks = includeParser.preprocessSource();
            sourceChunks.push(...chunks);
        }
    }
    /**
     * Checks if a preprocessor
     */
    expectEnd() {
        const completionToken = this._exprLexer.get();
        if (completionToken.type !== tokens_1.TokenType.Eof &&
            completionToken.type != tokens_1.TokenType.NewLine) {
            this.reportError("P005");
        }
    }
    /**
     * Parses a binary OR expression
     * orExpr
     *   : xorExpr ( "|" xorExpr )?
     */
    parseOrExpression() {
        let leftExpr = this.parseXorExpression();
        if (!leftExpr) {
            return null;
        }
        while (this.skipToken(tokens_1.TokenType.OrOp)) {
            const rightExpr = this.parseXorExpression();
            if (!rightExpr) {
                this.reportError("P007");
                return null;
            }
            leftExpr = {
                type: "BinaryExpr",
                operator: "|",
                leftOperand: leftExpr,
                rightOperand: rightExpr,
            };
        }
        return leftExpr;
    }
    /**
     * Evaluates a binary XOR expression
     * xorExpr
     *   : andExpr ( "^" andExpr )?
     */
    parseXorExpression() {
        let leftExpr = this.parseAndExpression();
        if (!leftExpr) {
            return null;
        }
        while (this.skipToken(tokens_1.TokenType.XorOp)) {
            const rightExpr = this.parseAndExpression();
            if (!rightExpr) {
                this.reportError("P007");
                return null;
            }
            leftExpr = {
                type: "BinaryExpr",
                operator: "^",
                leftOperand: leftExpr,
                rightOperand: rightExpr,
            };
        }
        return leftExpr;
    }
    /**
     * Evaluates a binary AND expression
     * andExpr
     *   : primaryExpr ( "&" primaryExpr )?
     */
    parseAndExpression() {
        let leftExpr = this.parsePrimaryExpression();
        if (!leftExpr) {
            return null;
        }
        while (this.skipToken(tokens_1.TokenType.AndOp)) {
            const rightExpr = this.parsePrimaryExpression();
            if (!rightExpr) {
                this.reportError("P007");
                return null;
            }
            leftExpr = {
                type: "BinaryExpr",
                operator: "&",
                leftOperand: leftExpr,
                rightOperand: rightExpr,
            };
        }
        return leftExpr;
    }
    /**
     * Evaluates a primary expression
     */
    parsePrimaryExpression() {
        let next = this._exprLexer.peek();
        // --- Test for parenthesized expression
        if (next.type === tokens_1.TokenType.LParent) {
            return this.parseParenthesizedExpression();
        }
        // --- Test for NOT expression
        if (next.type === tokens_1.TokenType.NotOp) {
            this._exprLexer.get();
            const expr = this.parseExpression();
            return {
                type: "NotExpr",
                operand: expr,
            };
        }
        // --- Test for identifier
        if (next.type !== tokens_1.TokenType.PPIdentifier) {
            this.reportError("P004");
            return null;
        }
        // --- Rip off the identifier
        this._exprLexer.get();
        return {
            type: "Id",
            name: next.text,
        };
    }
    /**
     * Evaluates a parenthesized expression
     * parenthesizedExpr
     *   : "(" expr ")"
     *   ;
     */
    parseParenthesizedExpression() {
        if (this.skipToken(tokens_1.TokenType.LParent)) {
            const expr = this.parseExpression();
            if (!expr) {
                return null;
            }
            this.expectToken(tokens_1.TokenType.RParent, "P008");
            return expr;
        }
        return null;
    }
    /**
     * Tests the type of the next token
     * @param type Expected token type
     */
    expectToken(type, errorCode, allowEof) {
        const next = this._exprLexer.peek();
        if (next.type === type || (allowEof && next.type === tokens_1.TokenType.Eof)) {
            // --- Skip the expected token
            this._exprLexer.get();
            return;
        }
        this.reportError(errorCode !== null && errorCode !== void 0 ? errorCode : "P006", next, next.text);
    }
    /**
     * Skips the next token if the type is the specified one
     * @param type Token type to check
     */
    skipToken(type) {
        const next = this._exprLexer.peek();
        if (next.type === type) {
            this._exprLexer.get();
            return next;
        }
        return null;
    }
    /**
     * Skips the next token if the type is the specified one
     * @param type Token type to check
     */
    skipTokens(...types) {
        const next = this._exprLexer.peek();
        for (const type of types) {
            if (next.type === type) {
                this._exprLexer.get();
                return next;
            }
        }
        return null;
    }
    /**
     * Reports the specified error
     * @param errorCode Error code
     * @param token Token that represents the error's position
     * @param options Error message options
     */
    reportError(errorCode, token, ...options) {
        var _a;
        let errorText = (_a = errors_1.errorMessages[errorCode]) !== null && _a !== void 0 ? _a : "Unkonwn error";
        if (options) {
            options.forEach((o, idx) => (errorText = replace(errorText, `{${idx}}`, options[idx].toString())));
        }
        if (!token) {
            token = this._lexer.peek();
        }
        this._parseErrors.push({
            code: errorCode,
            text: errorText,
            line: token.location.startLine,
            column: token.location.startColumn,
            position: token.location.startPosition,
        });
        throw new errors_1.ParserError(errorText, errorCode);
        function replace(input, placeholder, replacement) {
            do {
                input = input.replace(placeholder, replacement);
            } while (input.includes(placeholder));
            return input;
        }
    }
}
exports.PreprocessorParser = PreprocessorParser;


/***/ }),

/***/ "./src/preprocessor/PreprocessorStringLiteralLexer.ts":
/*!************************************************************!*\
  !*** ./src/preprocessor/PreprocessorStringLiteralLexer.ts ***!
  \************************************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.PreprocessorStringLiteralLexer = void 0;
const tokens_1 = __webpack_require__(/*! ../core/tokens */ "./src/core/tokens.ts");
/**
 * This class implements the tokenizer (lexer) of the WA# string literal handler
 */
class PreprocessorStringLiteralLexer {
    /**
     * Initializes the tokenizer with the input stream
     * @param input Input source code stream
     */
    constructor(input) {
        this.input = input;
        // --- Already fetched tokens
        this._ahead = [];
        // --- Prefetched character (from the next token)
        this._prefetched = null;
        // --- Prefetched character position (from the next token)
        this._prefetchedPos = null;
        // --- Prefetched character column (from the next token)
        this._prefetchedColumn = null;
    }
    /**
     * Fethches the next token without advancing to its position
     * @param ws If true, retrieve whitespaces too
     */
    peek(ws = false) {
        return this.ahead(0, ws);
    }
    /**
     *
     * @param n Number of token positions to read ahead
     * @param ws If true, retrieve whitespaces too
     */
    ahead(n = 1, ws = false) {
        if (n > 16) {
            throw new Error("Cannot look ahead more than 16 tokens");
        }
        // --- Prefetch missing tokens
        while (this._ahead.length <= n) {
            const token = this.fetch();
            if (tokens_1.isEof(token)) {
                return token;
            }
            if (ws || (!ws && !tokens_1.isWs(token))) {
                this._ahead.push(token);
            }
        }
        return this._ahead[n];
    }
    /**
     * Fethces the next token and advances the stream position
     * @param ws If true, retrieve whitespaces too
     */
    get(ws = false) {
        if (this._ahead.length > 0) {
            const token = this._ahead.shift();
            if (!token) {
                throw new Error("Token expected");
            }
            return token;
        }
        while (true) {
            const token = this.fetch();
            if (tokens_1.isEof(token) || ws || (!ws && !tokens_1.isWs(token))) {
                return token;
            }
        }
    }
    /**
     * Resets the parser; drops read-ahead tokens.
     */
    reset() {
        this._ahead = [];
    }
    /**
     * Fetches the next token from the input stream
     */
    fetch() {
        const lexer = this;
        const input = this.input;
        const startPos = this._prefetchedPos || input.position;
        const startLine = input.line;
        const startColumn = this._prefetchedColumn || input.column;
        let text = "";
        let tokenType = tokens_1.TokenType.Eof;
        let lastEndPos = input.position;
        let lastEndColumn = input.column;
        let ch = null;
        let phase = LexerPhase.Start;
        while (true) {
            // --- Get the next character
            ch = fetchNextChar();
            // --- In case of EOF, return the current token data
            if (ch === null) {
                return makeToken();
            }
            // --- Set the intial token type to unknown for the other characters
            if (tokenType === tokens_1.TokenType.Eof) {
                tokenType = tokens_1.TokenType.Unknown;
            }
            // --- Follow the lexer state machine
            switch (phase) {
                // ====================================================================
                // We are at the beginning of a new line
                case LexerPhase.Start:
                    // --- We start a source code chunk
                    tokenType = tokens_1.TokenType.SourceChunk;
                    switch (ch) {
                        // --- Go on with whitespaces
                        case " ":
                        case "\t":
                        case "\r":
                            phase = LexerPhase.InWhiteSpace;
                            break;
                        // --- The start of a string literal
                        case '"':
                            phase = LexerPhase.String;
                            break;
                        // --- End of the expression
                        case "\n":
                            return completeToken(tokens_1.TokenType.NewLine);
                    }
                    break;
                // --- Collect all whitespace characters
                case LexerPhase.InWhiteSpace:
                    switch (ch) {
                        case " ":
                        case "\t":
                        case "\r":
                            break;
                        // --- We stay in this phase with a new line
                        case "\n":
                            phase = LexerPhase.Start;
                            break;
                        // --- The start of a string literal
                        case '"':
                            phase = LexerPhase.String;
                            break;
                        // --- It cannot be a string literal
                        default:
                            return makeToken();
                    }
                    break;
                // --- Next character of the string
                case LexerPhase.String:
                    if (ch === '"') {
                        return completeToken(tokens_1.TokenType.PPStringLiteral);
                    }
                    else if (tokens_1.isRestrictedInString(ch)) {
                        return completeToken(tokens_1.TokenType.Unknown);
                    }
                    else if (ch === "\\") {
                        phase = LexerPhase.StringBackSlash;
                        tokenType = tokens_1.TokenType.Unknown;
                    }
                    break;
                // Start of string character escape
                case LexerPhase.StringBackSlash:
                    switch (ch) {
                        case "b":
                        case "f":
                        case "n":
                        case "r":
                        case "t":
                        case "v":
                        case "0":
                        case "'":
                        case '"':
                        case "\\":
                            phase = LexerPhase.String;
                            break;
                        default:
                            if (ch === "x") {
                                phase = LexerPhase.StringHexa1;
                            }
                            else {
                                phase = LexerPhase.String;
                            }
                    }
                    break;
                // First hexadecimal digit of string character escape
                case LexerPhase.StringHexa1:
                    if (tokens_1.isHexadecimalDigit(ch)) {
                        phase = LexerPhase.StringHexa2;
                    }
                    else {
                        return completeToken(tokens_1.TokenType.Unknown);
                    }
                    break;
                // Second hexadecimal digit of character escape
                case LexerPhase.StringHexa2:
                    if (tokens_1.isHexadecimalDigit(ch)) {
                        phase = LexerPhase.String;
                    }
                    else {
                        return completeToken(tokens_1.TokenType.Unknown);
                    }
                    break;
                // ====================================================================
                // --- We cannot continue
                default:
                    return makeToken();
            }
            // --- Append the char to the current text
            appendTokenChar();
            // --- Go on with parsing the next character
        }
        /**
         * Appends the last character to the token, and manages positions
         */
        function appendTokenChar() {
            text += ch;
            lexer._prefetched = null;
            lexer._prefetchedPos = null;
            lexer._prefetchedColumn = null;
            lastEndPos = input.position;
            lastEndColumn = input.position;
        }
        /**
         * Fetches the next character from the input stream
         */
        function fetchNextChar() {
            let ch;
            if (!lexer._prefetched) {
                lexer._prefetchedPos = input.position;
                lexer._prefetchedColumn = input.column;
                lexer._prefetched = input.get();
            }
            return lexer._prefetched;
        }
        /**
         * Packs the specified type of token to send back
         * @param type
         */
        function makeToken() {
            return {
                text,
                type: tokenType,
                location: {
                    fileIndex: input.fileIndex,
                    startPosition: startPos,
                    endPosition: lastEndPos,
                    startLine,
                    endLine: input.line,
                    startColumn,
                    endColumn: lastEndColumn,
                },
            };
        }
        /**
         * Add the last character to the token and return it
         */
        function completeToken(suggestedType) {
            appendTokenChar();
            // --- Send back the token
            if (suggestedType !== undefined) {
                tokenType = suggestedType;
            }
            return makeToken();
        }
    }
}
exports.PreprocessorStringLiteralLexer = PreprocessorStringLiteralLexer;
/**
 * Represents the current lexer phase
 */
var LexerPhase;
(function (LexerPhase) {
    LexerPhase[LexerPhase["Start"] = 0] = "Start";
    LexerPhase[LexerPhase["InWhiteSpace"] = 1] = "InWhiteSpace";
    LexerPhase[LexerPhase["String"] = 2] = "String";
    LexerPhase[LexerPhase["StringBackSlash"] = 3] = "StringBackSlash";
    LexerPhase[LexerPhase["StringHexa1"] = 4] = "StringHexa1";
    LexerPhase[LexerPhase["StringHexa2"] = 5] = "StringHexa2";
})(LexerPhase || (LexerPhase = {}));


/***/ }),

/***/ "./src/preprocessor/preprocessor-string.ts":
/*!*************************************************!*\
  !*** ./src/preprocessor/preprocessor-string.ts ***!
  \*************************************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.convertPPStringLiteralToRawString = void 0;
/**
 * Converts a perprocesor string literal to intrinsic string
 * @param input ZX Spectrum string to convert
 */
function convertPPStringLiteralToRawString(input) {
    input = input.substr(1, input.length - 2);
    let result = "";
    let state = StrParseState.Normal;
    let collect = 0;
    for (const ch of input) {
        switch (state) {
            case StrParseState.Normal:
                if (ch === "\\") {
                    state = StrParseState.Backslash;
                }
                else {
                    result += ch;
                }
                break;
            case StrParseState.Backslash:
                state = StrParseState.Normal;
                switch (ch) {
                    case "b":
                        result += "\b";
                        break;
                    case "f":
                        result += "\f";
                        break;
                    case "n":
                        result += "\n";
                        break;
                    case "r":
                        result += "\r";
                        break;
                    case "t":
                        result += "\t";
                        break;
                    case "v":
                        result += "\v";
                        break;
                    case "0":
                        result += "\0";
                        break;
                    case "'":
                        result += "'";
                        break;
                    case '"':
                        result += '"';
                        break;
                    case "\\":
                        result += "\\";
                        break;
                    case "x":
                        state = StrParseState.X;
                        break;
                    default:
                        result += ch;
                        break;
                }
                break;
            case StrParseState.X:
                if ((ch >= "0" && ch <= "9") ||
                    (ch >= "a" && ch <= "f") ||
                    (ch >= "A" && ch <= "F")) {
                    collect = parseInt(ch, 16);
                    state = StrParseState.Xh;
                }
                else {
                    result += "x";
                    state = StrParseState.Normal;
                }
                break;
            case StrParseState.Xh:
                if ((ch >= "0" && ch <= "9") ||
                    (ch >= "a" && ch <= "f") ||
                    (ch >= "A" && ch <= "F")) {
                    collect = collect * 0x10 + parseInt(ch, 16);
                    result += String.fromCharCode(collect);
                    state = StrParseState.Normal;
                }
                else {
                    result += String.fromCharCode(collect);
                    result += ch;
                    state = StrParseState.Normal;
                }
                break;
        }
    }
    // --- Handle the final machine state
    switch (state) {
        case StrParseState.Backslash:
            result += "\\";
            break;
        case StrParseState.X:
            result += "x";
            break;
        case StrParseState.Xh:
            result += String.fromCharCode(collect);
            break;
    }
    return result;
}
exports.convertPPStringLiteralToRawString = convertPPStringLiteralToRawString;
/**
 * States of the string parsing
 */
var StrParseState;
(function (StrParseState) {
    StrParseState[StrParseState["Normal"] = 0] = "Normal";
    StrParseState[StrParseState["Backslash"] = 1] = "Backslash";
    StrParseState[StrParseState["X"] = 2] = "X";
    StrParseState[StrParseState["Xh"] = 3] = "Xh";
})(StrParseState || (StrParseState = {}));


/***/ }),

/***/ "./src/version.ts":
/*!************************!*\
  !*** ./src/version.ts ***!
  \************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.getVersion = void 0;
// Cannot be `import` as it's not under TS root dir
/* eslint-disable-next-line @typescript-eslint/no-var-requires */
const { version: VERSION } = __webpack_require__(/*! ../package.json */ "./package.json");
function getVersion() {
    return VERSION;
}
exports.getVersion = getVersion;


/***/ }),

/***/ "./src/wa-ast/FunctionBuilder.ts":
/*!***************************************!*\
  !*** ./src/wa-ast/FunctionBuilder.ts ***!
  \***************************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.demote64 = exports.convert64 = exports.convert32 = exports.trunc64 = exports.trunc32 = exports.extend32 = exports.wrap64 = exports.gt = exports.ge = exports.lt = exports.le = exports.ne = exports.eq = exports.eqz = exports.rotr = exports.rotl = exports.shr = exports.shl = exports.or = exports.xor = exports.and = exports.rem = exports.div = exports.mul = exports.sub = exports.add = exports.popcnt = exports.ctz = exports.clz = exports.memGrow = exports.memSize = exports.store = exports.load = exports.globalSet = exports.globalGet = exports.localTee = exports.localSet = exports.localGet = exports.select = exports.drop = exports.callIndirect = exports.call = exports.ret = exports.branchTable = exports.branchIf = exports.branch = exports.nop = exports.unreachable = exports.constVal = exports.FunctionBuilder = void 0;
exports.comment = exports.ifBlock = exports.loop = exports.block = exports.copysign = exports.max = exports.min = exports.sqrt = exports.nearest = exports.trunc = exports.floor = exports.ceil = exports.neg = exports.abs = exports.reinterpretI64 = exports.reinterpretI32 = exports.reinterpretF64 = exports.reinterpretF32 = exports.promote32 = void 0;
const wa_nodes_1 = __webpack_require__(/*! ./wa-nodes */ "./src/wa-ast/wa-nodes.ts");
/**
 * This class represents a function node that can be used as a builder to
 * inject function body instructions
 */
class FunctionBuilder {
    constructor(id, params, resultType, locals, body) {
        this.id = id;
        this.params = params;
        this.resultType = resultType;
        this.type = "Func";
        this.locals = locals !== null && locals !== void 0 ? locals : [];
        this.body = body !== null && body !== void 0 ? body : [];
    }
    /**
     * Adds a local node to the function locals
     * @param id
     * @param valueType Value type of operation
     */
    addLocal(id, valueType) {
        const newNode = {
            type: "Local",
            id,
            valueType,
        };
        this.locals.push(newNode);
        return newNode;
    }
}
exports.FunctionBuilder = FunctionBuilder;
/**
 * Factory method for a const WA instrcution
 * @param valueType Value type of operation
 * @param value Constant value
 */
function constVal(valueType, value) {
    return {
        type: "ConstVal",
        valueType,
        value,
    };
}
exports.constVal = constVal;
/**
 * Factory method for an unreachable WA instruction
 */
function unreachable() {
    return {
        type: "Unreachable",
    };
}
exports.unreachable = unreachable;
/**
 * Factory method for an nop WA instruction
 */
function nop() {
    return {
        type: "Nop",
    };
}
exports.nop = nop;
/**
 * Factory method for a br WA instruction
 * @param label Target label
 */
function branch(label) {
    return {
        type: "Branch",
        label,
    };
}
exports.branch = branch;
/**
 * Factory method for a br_if WA instruction
 * @param label Target label
 */
function branchIf(label) {
    return {
        type: "BranchIf",
        label,
    };
}
exports.branchIf = branchIf;
/**
 * Factory method for a br_table WA instruction
 * @param caseIds Branch IDs
 * @param defaultId Default branch ID
 */
function branchTable(caseIds, defaultId) {
    return {
        type: "BranchTable",
        caseIds,
        defaultId,
    };
}
exports.branchTable = branchTable;
/**
 * Factory method for a return WA instruction
 */
function ret() {
    return {
        type: "Return",
    };
}
exports.ret = ret;
/**
 * Factory method for a call WA instruction
 * @param id Function ID
 */
function call(id) {
    return {
        type: "Call",
        id,
    };
}
exports.call = call;
/**
 * Factory method for a call_indirect WA instruction
 * @param typeId Function type identifier
 */
function callIndirect(typeId) {
    return {
        type: "CallIndirect",
        typeId,
    };
}
exports.callIndirect = callIndirect;
/**
 * Factory method for a drop WA instruction
 */
function drop() {
    return {
        type: "Drop",
    };
}
exports.drop = drop;
/**
 * Factory method for a select WA instruction
 */
function select() {
    return {
        type: "Select",
    };
}
exports.select = select;
/**
 * Factory method for a local_get WA instruction
 * @param id Identifier
 */
function localGet(id) {
    return {
        type: "LocalGet",
        id,
    };
}
exports.localGet = localGet;
/**
 * Factory method for a local_set WA instruction
 * @param id Identifier
 */
function localSet(id) {
    return {
        type: "LocalSet",
        id,
    };
}
exports.localSet = localSet;
/**
 * Factory method for a local_tee WA instruction
 * @param id Identifier
 */
function localTee(id) {
    return {
        type: "LocalTee",
        id,
    };
}
exports.localTee = localTee;
/**
 * Factory method for a global_get WA instruction
 * @param id Identifier
 */
function globalGet(id) {
    return {
        type: "GlobalGet",
        id,
    };
}
exports.globalGet = globalGet;
/**
 * Factory method for a global_set WA instruction
 * @param id Identifier
 */
function globalSet(id) {
    return {
        type: "GlobalSet",
        id,
    };
}
exports.globalSet = globalSet;
/**
 * Factory method for a load WA instruction
 * @param valueType Value type of operation
 * @param bits Bit specification
 * @param offset Optional offset
 * @param align Optional alignment
 * @param signed Signed operation?
 */
function load(valueType, bits = wa_nodes_1.WaBitSpec.None, offset = 0, align = 0, signed = false) {
    // --- Check input consistency
    if ((valueType === wa_nodes_1.WaType.f32 || valueType === wa_nodes_1.WaType.f64) &&
        bits !== wa_nodes_1.WaBitSpec.None) {
        throw new Error("Cannot use floating-point load instruction with bit-width specification");
    }
    else if (valueType === wa_nodes_1.WaType.i32 && bits === wa_nodes_1.WaBitSpec.Bit32) {
        bits = wa_nodes_1.WaBitSpec.None;
    }
    return {
        type: "Load",
        valueType,
        bits,
        offset,
        align,
        signed,
    };
}
exports.load = load;
/**
 * Factory method for a load WA instruction
 * @param valueType Value type of operation
 * @param bits Bit specification
 * @param offset Optional offset
 * @param align Optional alignment
 */
function store(valueType, bits = wa_nodes_1.WaBitSpec.None, offset = 0, align = 0) {
    // --- Check input consistency
    if ((valueType === wa_nodes_1.WaType.f32 || valueType === wa_nodes_1.WaType.f64) &&
        bits !== wa_nodes_1.WaBitSpec.None) {
        throw new Error("Cannot use floating-point store instruction with bit-width specification");
    }
    else if (valueType === wa_nodes_1.WaType.i32 && bits === wa_nodes_1.WaBitSpec.Bit32) {
        bits = wa_nodes_1.WaBitSpec.None;
    }
    return {
        type: "Store",
        valueType,
        bits,
        offset,
        align,
    };
}
exports.store = store;
/**
 * Factory method for a memory.size WA instruction
 */
function memSize() {
    return {
        type: "MemorySize",
    };
}
exports.memSize = memSize;
/**
 * Factory method for a memory.grow WA instruction
 */
function memGrow() {
    return {
        type: "MemoryGrow",
    };
}
exports.memGrow = memGrow;
/**
 * Factory method for a clz WA instruction
 * @param valueType Value type of operation
 */
function clz(valueType) {
    checkInteger(valueType, "clz");
    return {
        type: "Clz",
        valueType,
    };
}
exports.clz = clz;
/**
 * Factory method for a ctz WA instruction
 * @param valueType Value type of operation
 */
function ctz(valueType) {
    checkInteger(valueType, "ctz");
    return {
        type: "Ctz",
        valueType,
    };
}
exports.ctz = ctz;
/**
 * Factory method for a popcnt WA instruction
 * @param valueType Value type of operation
 */
function popcnt(valueType) {
    checkInteger(valueType, "popcnt");
    return {
        type: "PopCnt",
        valueType,
    };
}
exports.popcnt = popcnt;
/**
 * Factory method for an add WA instruction
 * @param valueType Value type of operation
 */
function add(valueType) {
    return {
        type: "Add",
        valueType,
    };
}
exports.add = add;
/**
 * Factory method for a sub WA instruction
 * @param valueType Value type of operation
 */
function sub(valueType) {
    return {
        type: "Sub",
        valueType,
    };
}
exports.sub = sub;
/**
 * Factory method for a mul WA instruction
 * @param valueType Value type of operation
 */
function mul(valueType) {
    return {
        type: "Mul",
        valueType,
    };
}
exports.mul = mul;
/**
 * Factory method for a div WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function div(valueType, signed) {
    return {
        type: "Div",
        valueType,
        signed,
    };
}
exports.div = div;
/**
 * Factory method for a rem WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function rem(valueType, signed) {
    return {
        type: "Rem",
        valueType,
        signed,
    };
}
exports.rem = rem;
/**
 * Factory method for an and WA instruction
 * @param valueType Value type of operation
 */
function and(valueType) {
    checkInteger(valueType, "and");
    return {
        type: "And",
        valueType,
    };
}
exports.and = and;
/**
 * Factory method for an xor WA instruction
 * @param valueType Value type of operation
 */
function xor(valueType) {
    checkInteger(valueType, "xor");
    return {
        type: "Xor",
        valueType,
    };
}
exports.xor = xor;
/**
 * Factory method for an or WA instruction
 * @param valueType Value type of operation
 */
function or(valueType) {
    checkInteger(valueType, "or");
    return {
        type: "Or",
        valueType,
    };
}
exports.or = or;
/**
 * Factory method for a shl WA instruction
 * @param valueType Value type of operation
 */
function shl(valueType) {
    checkInteger(valueType, "shl");
    return {
        type: "Shl",
        valueType,
    };
}
exports.shl = shl;
/**
 * Factory method for a shr WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function shr(valueType, signed) {
    checkInteger(valueType, "shr");
    return {
        type: "Shr",
        valueType,
        signed,
    };
}
exports.shr = shr;
/**
 * Factory method for a rotl WA instruction
 * @param valueType Value type of operation
 */
function rotl(valueType) {
    checkInteger(valueType, "rotl");
    return {
        type: "Rotl",
        valueType,
    };
}
exports.rotl = rotl;
/**
 * Factory method for a rotr WA instruction
 * @param valueType Value type of operation
 */
function rotr(valueType) {
    checkInteger(valueType, "rotr");
    return {
        type: "Rotr",
        valueType,
    };
}
exports.rotr = rotr;
/**
 * Factory method for an eqz WA instruction
 * @param valueType Value type of operation
 */
function eqz(valueType) {
    checkInteger(valueType, "eqz");
    return {
        type: "Eqz",
        valueType,
    };
}
exports.eqz = eqz;
/**
 * Factory method for an eq WA instruction
 * @param valueType Value type of operation
 */
function eq(valueType) {
    checkInteger(valueType, "eq");
    return {
        type: "Eq",
        valueType,
    };
}
exports.eq = eq;
/**
 * Factory method for an ne WA instruction
 * @param valueType Value type of operation
 */
function ne(valueType) {
    checkInteger(valueType, "ne");
    return {
        type: "Ne",
        valueType,
    };
}
exports.ne = ne;
/**
 * Factory method for a le WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function le(valueType, signed) {
    return {
        type: "Le",
        valueType,
        signed,
    };
}
exports.le = le;
/**
 * Factory method for a lt WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function lt(valueType, signed) {
    return {
        type: "Lt",
        valueType,
        signed,
    };
}
exports.lt = lt;
/**
 * Factory method for a ge WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function ge(valueType, signed) {
    return {
        type: "Ge",
        valueType,
        signed,
    };
}
exports.ge = ge;
/**
 * Factory method for a ge WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function gt(valueType, signed) {
    return {
        type: "Gt",
        valueType,
        signed,
    };
}
exports.gt = gt;
/**
 * Factory method for an i32.wrap_i64 WA instruction
 */
function wrap64() {
    return {
        type: "Wrap64",
    };
}
exports.wrap64 = wrap64;
/**
 * Factory method for an extend32 WA instruction
 * @param signed Signed operation?
 */
function extend32(signed) {
    return {
        type: "Extend32",
        signed,
    };
}
exports.extend32 = extend32;
/**
 * Factory method for an i32.trunc WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function trunc32(valueType, signed) {
    checkFloat(valueType, "trunc32");
    return {
        type: "Trunc32",
        valueType,
        signed,
    };
}
exports.trunc32 = trunc32;
/**
 * Factory method for an i64.trunc WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function trunc64(valueType, signed) {
    checkFloat(valueType, "trunc64");
    return {
        type: "Trunc64",
        valueType,
        signed,
    };
}
exports.trunc64 = trunc64;
/**
 * Factory method for an f32.convert WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function convert32(valueType, signed) {
    checkInteger(valueType, "convert32");
    return {
        type: "Convert32",
        valueType,
        signed,
    };
}
exports.convert32 = convert32;
/**
 * Factory method for an f64.convert WA instruction
 * @param valueType Value type of operation
 * @param signed Signed operation?
 */
function convert64(valueType, signed) {
    checkInteger(valueType, "convert64");
    return {
        type: "Convert64",
        valueType,
        signed,
    };
}
exports.convert64 = convert64;
/**
 * Factory method for an f64.demote WA instruction
 */
function demote64() {
    return {
        type: "Demote64",
    };
}
exports.demote64 = demote64;
/**
 * Factory method for an f32.promote WA instruction
 */
function promote32() {
    return {
        type: "Promote32",
    };
}
exports.promote32 = promote32;
/**
 * Factory method for an i32.reinterpret WA instruction
 */
function reinterpretF32() {
    return {
        type: "ReinterpretF32",
    };
}
exports.reinterpretF32 = reinterpretF32;
/**
 * Factory method for an i64.reinterpret WA instruction
 */
function reinterpretF64() {
    return {
        type: "ReinterpretF64",
    };
}
exports.reinterpretF64 = reinterpretF64;
/**
 * Factory method for an f32.reinterpret WA instruction
 */
function reinterpretI32() {
    return {
        type: "ReinterpretI32",
    };
}
exports.reinterpretI32 = reinterpretI32;
/**
 * Factory method for an f64.reinterpret WA instruction
 */
function reinterpretI64() {
    return {
        type: "ReinterpretI64",
    };
}
exports.reinterpretI64 = reinterpretI64;
/**
 * Factory method for an abs WA instruction
 * @param valueType Value type of operation
 */
function abs(valueType) {
    checkFloat(valueType, "abs");
    return {
        type: "Abs",
        valueType,
    };
}
exports.abs = abs;
/**
 * Factory method for a neg WA instruction
 * @param valueType Value type of operation
 */
function neg(valueType) {
    checkFloat(valueType, "neg");
    return {
        type: "Neg",
        valueType,
    };
}
exports.neg = neg;
/**
 * Factory method for a ceil WA instruction
 * @param valueType Value type of operation
 */
function ceil(valueType) {
    checkFloat(valueType, "ceil");
    return {
        type: "Ceil",
        valueType,
    };
}
exports.ceil = ceil;
/**
 * Factory method for a floor WA instruction
 * @param valueType Value type of operation
 */
function floor(valueType) {
    checkFloat(valueType, "floor");
    return {
        type: "Floor",
        valueType,
    };
}
exports.floor = floor;
/**
 * Factory method for a trunc WA instruction
 * @param valueType Value type of operation
 */
function trunc(valueType) {
    checkFloat(valueType, "trunc");
    return {
        type: "Trunc",
        valueType,
    };
}
exports.trunc = trunc;
/**
 * Factory method for a nearest WA instruction
 * @param valueType Value type of operation
 */
function nearest(valueType) {
    checkFloat(valueType, "nearest");
    return {
        type: "Nearest",
        valueType,
    };
}
exports.nearest = nearest;
/**
 * Factory method for an sqrt WA instruction
 * @param valueType Value type of operation
 */
function sqrt(valueType) {
    checkFloat(valueType, "sqrt");
    return {
        type: "Sqrt",
        valueType,
    };
}
exports.sqrt = sqrt;
/**
 * Factory method for a min WA instruction
 * @param valueType Value type of operation
 */
function min(valueType) {
    checkFloat(valueType, "min");
    return {
        type: "Min",
        valueType,
    };
}
exports.min = min;
/**
 * Factory method for a max WA instruction
 * @param valueType Value type of operation
 */
function max(valueType) {
    checkFloat(valueType, "max");
    return {
        type: "Max",
        valueType,
    };
}
exports.max = max;
/**
 * Factory method for a copysign WA instruction
 * @param valueType Value type of operation
 */
function copysign(valueType) {
    checkFloat(valueType, "copysign");
    return {
        type: "CopySign",
        valueType,
    };
}
exports.copysign = copysign;
/**
 * Factory method for a block WA instruction
 * @param id Block identifier
 * @param body Block body
 * @param resultType Type of optional block result
 */
function block(id, body, resultType) {
    return {
        type: "Block",
        id,
        body,
        resultType,
    };
}
exports.block = block;
/**
 * Factory method for a loop WA instruction
 * @param id Loop identifier
 * @param body Loop body
 * @param resultType Type of optional loop result
 */
function loop(id, body, resultType) {
    return {
        type: "Loop",
        id,
        body,
        resultType,
    };
}
exports.loop = loop;
/**
 * Factory method for an if WA instruction
 * @param consequtive Consequtive branch body
 * @param alternate Alternate branch body
 * @param resultType Type of optional result
 */
function ifBlock(consequtive, alternate, resultType) {
    return {
        type: "If",
        consequtive,
        alternate,
        resultType,
    };
}
exports.ifBlock = ifBlock;
/**
 * Factory method for a WA comment
 * @param text Comment text
 * @param isBlock Block comment?
 */
function comment(text, isBlock) {
    return {
        type: "Comment",
        text,
        isBlock,
    };
}
exports.comment = comment;
// ============================================================================
// Helpers
function checkInteger(type, op) {
    if (type !== wa_nodes_1.WaType.i32 && type !== wa_nodes_1.WaType.i64) {
        throw new Error(`Only i32 and i64 accepted for ${op}`);
    }
}
function checkFloat(type, op) {
    if (type !== wa_nodes_1.WaType.f32 && type !== wa_nodes_1.WaType.f64) {
        throw new Error(`Only f32 and f64 accepted for ${op}`);
    }
}


/***/ }),

/***/ "./src/wa-ast/WaTree.ts":
/*!******************************!*\
  !*** ./src/wa-ast/WaTree.ts ***!
  \******************************/
/***/ ((__unused_webpack_module, exports, __webpack_require__) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.WaTree = void 0;
const FunctionBuilder_1 = __webpack_require__(/*! ./FunctionBuilder */ "./src/wa-ast/FunctionBuilder.ts");
const wa_nodes_1 = __webpack_require__(/*! ./wa-nodes */ "./src/wa-ast/wa-nodes.ts");
/**
 * Represents a WebAssembly Text Format Source Tree
 */
class WaTree {
    /**
     * Initializes the tree
     */
    constructor() {
        // --- Indentation to use
        this._indentSpaces = 2;
        this._module = {
            type: "Module",
            memory: {
                export: {
                    name: "memory",
                },
                limit: 10,
            },
        };
    }
    /**
     * Gets the root module of the tree
     */
    get module() {
        return this._module;
    }
    /**
     * Sets the module's memory specification
     * @param limit Memory limit (in 64K pages)
     * @param exp Optional memory export name
     */
    setMemorySpecification(limit, exp) {
        this._module.memory.limit = limit;
        if (exp) {
            this._module.memory.export.name = exp;
        }
    }
    /**
     * Sets the module's table specification
     * @param limit Table size
     * @param id Optional table name
     */
    setTable(limit, id) {
        this._module.table = {
            type: "Table",
            id: id !== null && id !== void 0 ? id : "$table",
            limit,
        };
    }
    // ==========================================================================
    // Module emission methods
    /**
     * Adds a function builder to this tree
     * @param builder
     */
    addFunc(builder) {
        this._module.fields.push(builder);
    }
    /**
     * Injects a function import node into the tree
     * @param id Function identifier
     * @param name1 First import nametag
     * @param name2 Second import nametag
     * @param params Function parameters
     * @param resultType Function result type
     */
    functionImport(id, name1, name2, params, resultType) {
        this.ensureFields();
        const newNode = {
            type: "FuncImport",
            id,
            import: { name1, name2 },
            params,
            resultType,
        };
        this._module.fields.push(newNode);
        return newNode;
    }
    /**
     * Injects a function import node into the tree
     * @param id Function identifier
     * @param name1 First import nametag
     * @param name2 Second import nametag
     * @param params Function parameters
     * @param resultType Function result type
     */
    functionExport(id, name) {
        this.ensureFields();
        const newNode = {
            type: "FuncExport",
            id,
            export: { name },
        };
        this._module.fields.push(newNode);
        return newNode;
    }
    /**
     * Injects a mutable global declaration into the tree
     * @param id Global identifier
     * @param valueType Declaration value type
     * @param initialValue Initial value
     * @param exportId Optional export name
     */
    global(id, valueType, initialValue, exportId) {
        this.ensureFields();
        const newNode = {
            type: "Global",
            id,
            valueType,
            initialValue,
            exportSpec: exportId ? { name: exportId } : undefined,
        };
        this._module.fields.push(newNode);
        return newNode;
    }
    /**
     * Injects a type definition into the tree
     * @param id Type identifier
     * @param params Function parameters
     * @param resultType Optional result type
     */
    typeDef(id, params, resultType) {
        this.ensureFields();
        const newNode = {
            type: "TypeDef",
            id,
            params,
            resultType,
        };
        this._module.fields.push(newNode);
        return newNode;
    }
    /**
     * Injects an element into the tree
     * @param index Element index
     * @param ids Element IDs
     */
    element(index, ids) {
        this.ensureFields();
        const newNode = {
            type: "Element",
            index,
            ids,
        };
        this._module.fields.push(newNode);
        return newNode;
    }
    /**
     * Injects a data tag into the tree
     * @param index Element index
     * @param ids Element IDs
     */
    data(address, bytes) {
        this.ensureFields();
        const newNode = {
            type: "Data",
            address,
            bytes,
        };
        this._module.fields.push(newNode);
        return newNode;
    }
    /**
     * Injects a function into the tree
     * @param id Function identifier
     * @param params Function parameters
     * @param resultType Optional result type
     * @param body Instructions of the function body
     */
    func(id, params, resultType, locals, body) {
        this.ensureFields();
        const newNode = new FunctionBuilder_1.FunctionBuilder(id, params, resultType, locals, body);
        return newNode;
    }
    /**
     * Injects a comment into the tree
     * @param id Type identifier
     * @param params Function parameters
     * @param resultType Optional result type
     */
    comment(text, isBlock) {
        this.ensureFields();
        const newNode = {
            type: "Comment",
            text,
            isBlock,
        };
        this._module.fields.push(newNode);
        return newNode;
    }
    /**
     * Injects a separator line into the tree
     */
    separatorLine() {
        this.ensureFields();
        const newNode = {
            type: "SeparatorLine",
        };
        this._module.fields.push(newNode);
    }
    /**
     * Ensures
     */
    ensureFields() {
        if (!this._module.fields) {
            this._module.fields = [];
        }
    }
    // ==========================================================================
    // Rendering methods
    /**
     * Renders the WebAssembly tree
     */
    render() {
        return this.renderNode(this._module);
    }
    /**
     * Renders the specified node
     * @param node Node to render
     * @param indent Rendering indentation depth
     */
    renderNode(node, indent = 0) {
        const indentation = "".padStart(indent * this._indentSpaces, " ");
        return `${indentation}${this.renderInternal(node, indent)}`;
    }
    /**
     * Renders a Function node
     * @param node Node to render
     * @param indent Rendering indentation depth
     */
    renderFunctionNode(node, indent = 0) {
        const indentation = "".padStart(indent * this._indentSpaces, " ");
        const body = this.renderFunctionBody(node, indent + 1);
        return `${indentation}(func ${node.id}${node.params.length > 0 || node.resultType ? " " : ""}${this.renderFuncSignature(node.params, node.resultType)}\n${body}${body.length > 0 ? "\n" : ""}${indentation})`;
    }
    /**
     * Renders an instruction node with its children
     * @param node Function body node
     * @param indent Rendering indentation depth
     */
    renderInstructionNode(node, indent = 0, parenthesized = false) {
        const indentation = "".padStart(indent * this._indentSpaces, " ");
        return `${indentation}${parenthesized ? "(" : ""}${this.renderPureBodyNode(node, indent)}${parenthesized ? ")" : ""}`;
    }
    // ==========================================================================
    // Helpers
    /**
     * Carries out the rendering
     * @param node Node to render
     * @param indent Rendering indentation depth
     */
    renderInternal(node, indent) {
        switch (node.type) {
            case "Module": {
                const moduleFields = this.renderModule(node, indent + 1);
                return `(module\n${moduleFields}\n)`;
            }
        }
        return "";
    }
    /**
     * Renders a module
     * @param module Module to tender
     */
    renderModule(module, indent) {
        var _a, _b;
        const indentation = "".padStart(indent * this._indentSpaces, " ");
        const imports = (module.fields
            ? module.fields.filter((fi) => fi.type === "FuncImport")
            : []);
        const otherFields = module.fields
            ? module.fields.filter((fi) => fi.type !== "FuncImport")
            : [];
        return `${
        // --- Module imports
        imports
            .map((field) => this.renderModuleImport(field, indent))
            .join("\n") + (imports.length > 0 ? "\n" : "")}${indentation
        // --- Memory specification
        }(memory (export \"${(_a = module.memory.export.name) !== null && _a !== void 0 ? _a : "memory"}\") ${(_b = module.memory.limit) !== null && _b !== void 0 ? _b : 10})${module.table ? "\n" : ""}${
        // --- Table specification
        module.table
            ? indentation +
                "(table " +
                module.table.id +
                " " +
                module.table.limit +
                " anyfunc)"
            : ""}${otherFields.length > 0 ? "\n" : ""}${
        // --- Other module fields
        otherFields
            .map((field) => this.renderModuleField(field, indent))
            .join("\n")}`;
    }
    renderModuleImport(field, indent) {
        const indentation = "".padStart(indent * this._indentSpaces, " ");
        return `${indentation}(func ${field.id} (import \"${field.import.name1}\" \"${field.import.name2}\")${field.params.length > 0 || field.resultType
            ? " " + this.renderFuncSignature(field.params, field.resultType)
            : ""})`;
    }
    /**
     * Renders a FuncImport node
     * @param node Node to render
     * @param indent Rendering indentation depth
     */
    renderModuleField(field, indent) {
        var _a;
        const indentation = "".padStart(indent * this._indentSpaces, " ");
        const gap = "".padStart(this._indentSpaces, " ");
        switch (field.type) {
            case "FuncExport":
                return `${indentation}(export "${field.export.name}" (func ${field.id}))`;
            case "Global":
                return `${indentation}(global ${field.id} ${field.exportSpec ? "(export " + field.exportSpec.name + ") " : ""}(mut ${wa_nodes_1.WaType[field.valueType]}) (${wa_nodes_1.WaType[field.valueType]}.const ${(_a = field.initialValue) !== null && _a !== void 0 ? _a : 0}))`;
            case "TypeDef":
                return `${indentation}(type ${field.id} (func${field.params.length > 0 || field.resultType ? " " : ""}${this.renderFuncSignature(field.params, field.resultType)}))`;
            case "Element":
                return `${indentation}(elem (i32.const ${field.index})\n${field.ids
                    .map((id) => indentation + gap + id)
                    .join("\n")}\n${indentation})`;
            case "Data":
                return `${indentation}(data (i32.const ${field.address}) "${field.bytes
                    .map((b) => "\\" + (b & 0xff).toString(16).padStart(2, "0"))
                    .join("")}")`;
            case "Func":
                return this.renderFunctionNode(field, indent);
            case "Comment":
                return (indentation +
                    (field.isBlock ? `(; ${field.text} ;)` : `;; ${field.text}`));
            case "SeparatorLine":
                return indentation;
        }
    }
    /**
     * Renders the signature of the specified function
     * @param params Function parameters
     * @param resultType Optional result types
     */
    renderFuncSignature(params, resultType) {
        return `${params
            .map((p) => "(param " + (p.id ? p.id + " " : "") + wa_nodes_1.WaType[p.type] + ")")
            .join(" ")}${params.length > 0 && resultType ? " " : ""}${resultType ? "(result " + wa_nodes_1.WaType[resultType] + ")" : ""}`;
    }
    /**
     * Renders the body of a function
     * @param func Function body instructions
     */
    renderFunctionBody(func, indent) {
        const indentation = "".padStart(indent * this._indentSpaces, " ");
        const locals = func.locals
            .map((l) => `${indentation}${this.renderLocal(l)}`)
            .join("\n");
        const body = func.body
            .map((inst) => this.renderInstructionNode(inst, indent))
            .join("\n");
        return `${locals}${locals.length > 0 && body.length > 0 ? "\n" : ""}${body}`;
    }
    /**
     * Renders the specified local declaration
     * @param local Local declaration
     */
    renderLocal(local) {
        return `(local ${local.id} ${wa_nodes_1.WaType[local.valueType]})`;
    }
    /**
     * Renders a function body node without its children
     * @param node
     */
    renderPureBodyNode(node, indent) {
        const indentation = "".padStart(indent * this._indentSpaces, " ");
        switch (node.type) {
            case "ConstVal":
                return `${wa_nodes_1.WaType[node.valueType]}.const ${node.value}`;
            case "Unreachable":
                return "unreachable";
            case "Nop":
                return "nop";
            case "Branch":
                return `br ${node.label}`;
            case "BranchIf":
                return `br_if ${node.label}`;
            case "BranchTable":
                return `br_table${node.caseIds.length > 0 ? " " : ""}${node.caseIds.join(" ")} ${node.defaultId}`;
            case "Return":
                return "return";
            case "Call":
                return `call ${node.id}`;
            case "CallIndirect":
                return `call_indirect (type ${node.typeId})`;
            case "Drop":
                return "drop";
            case "Select":
                return "select";
            case "LocalGet":
                return `get_local ${node.id}`;
            case "LocalSet":
                return `set_local ${node.id}`;
            case "LocalTee":
                return `tee_local ${node.id}`;
            case "GlobalGet":
                return `get_global ${node.id}`;
            case "GlobalSet":
                return `set_global ${node.id}`;
            case "Load":
                return `${wa_nodes_1.WaType[node.valueType]}.load${bitSpecToString(node.bits)}${node.bits !== wa_nodes_1.WaBitSpec.None ? (node.signed ? "_s" : "_u") : ""}${node.offset ? " offset=" + node.offset : ""}${node.align ? " align=" + node.align : ""}`;
            case "Store":
                return `${wa_nodes_1.WaType[node.valueType]}.store${bitSpecToString(node.bits)}${node.offset ? " offset=" + node.offset : ""}${node.align ? " align=" + node.align : ""}`;
            case "MemorySize":
                return "memory.size";
            case "MemoryGrow":
                return "memory.grow";
            case "Clz":
                return `${wa_nodes_1.WaType[node.valueType]}.clz`;
            case "Ctz":
                return `${wa_nodes_1.WaType[node.valueType]}.ctz`;
            case "PopCnt":
                return `${wa_nodes_1.WaType[node.valueType]}.popcnt`;
            case "Add":
                return `${wa_nodes_1.WaType[node.valueType]}.add`;
            case "Sub":
                return `${wa_nodes_1.WaType[node.valueType]}.sub`;
            case "Mul":
                return `${wa_nodes_1.WaType[node.valueType]}.mul`;
            case "Div":
                return `${wa_nodes_1.WaType[node.valueType]}.div${signedTag(node.valueType, node.signed)}`;
            case "Rem":
                return `${wa_nodes_1.WaType[node.valueType]}.rem${signedTag(node.valueType, node.signed)}`;
            case "And":
                return `${wa_nodes_1.WaType[node.valueType]}.and`;
            case "Xor":
                return `${wa_nodes_1.WaType[node.valueType]}.xor`;
            case "Or":
                return `${wa_nodes_1.WaType[node.valueType]}.or`;
            case "Shl":
                return `${wa_nodes_1.WaType[node.valueType]}.shl`;
            case "Shr":
                return `${wa_nodes_1.WaType[node.valueType]}.shr${signedTag(node.valueType, node.signed)}`;
            case "Rotl":
                return `${wa_nodes_1.WaType[node.valueType]}.rotl`;
            case "Rotr":
                return `${wa_nodes_1.WaType[node.valueType]}.rotr`;
            case "Eqz":
                return `${wa_nodes_1.WaType[node.valueType]}.eqz`;
            case "Eq":
                return `${wa_nodes_1.WaType[node.valueType]}.eq`;
            case "Ne":
                return `${wa_nodes_1.WaType[node.valueType]}.ne`;
            case "Le":
                return `${wa_nodes_1.WaType[node.valueType]}.le${signedTag(node.valueType, node.signed)}`;
            case "Lt":
                return `${wa_nodes_1.WaType[node.valueType]}.lt${signedTag(node.valueType, node.signed)}`;
            case "Ge":
                return `${wa_nodes_1.WaType[node.valueType]}.ge${signedTag(node.valueType, node.signed)}`;
            case "Gt":
                return `${wa_nodes_1.WaType[node.valueType]}.gt${signedTag(node.valueType, node.signed)}`;
            case "Wrap64":
                return "i32.wrap/i64";
            case "Extend32":
                return `i64.extend${signedTag(wa_nodes_1.WaType.i32, node.signed)}/i32`;
            case "Trunc32":
                return `i32.trunc${signedTag(wa_nodes_1.WaType.i32, node.signed)}/${wa_nodes_1.WaType[node.valueType]}`;
            case "Trunc64":
                return `i64.trunc${signedTag(wa_nodes_1.WaType.i32, node.signed)}/${wa_nodes_1.WaType[node.valueType]}`;
            case "Convert32":
                return `f32.convert${signedTag(wa_nodes_1.WaType.i32, node.signed)}/${wa_nodes_1.WaType[node.valueType]}`;
            case "Convert64":
                return `f64.convert${signedTag(wa_nodes_1.WaType.i32, node.signed)}/${wa_nodes_1.WaType[node.valueType]}`;
            case "Demote64":
                return "f32.demote/f64";
            case "Promote32":
                return "f64.promote/f32";
            case "ReinterpretF32":
                return "i32.reinterpret/f32";
            case "ReinterpretF64":
                return "i64.reinterpret/f64";
            case "ReinterpretI32":
                return "f32.reinterpret/i32";
            case "ReinterpretI64":
                return "f64.reinterpret/i64";
            case "Abs":
                return `${wa_nodes_1.WaType[node.valueType]}.abs`;
            case "Neg":
                return `${wa_nodes_1.WaType[node.valueType]}.neg`;
            case "Ceil":
                return `${wa_nodes_1.WaType[node.valueType]}.ceil`;
            case "Floor":
                return `${wa_nodes_1.WaType[node.valueType]}.floor`;
            case "Trunc":
                return `${wa_nodes_1.WaType[node.valueType]}.trunc`;
            case "Nearest":
                return `${wa_nodes_1.WaType[node.valueType]}.nearest`;
            case "Sqrt":
                return `${wa_nodes_1.WaType[node.valueType]}.sqrt`;
            case "Min":
                return `${wa_nodes_1.WaType[node.valueType]}.min`;
            case "Max":
                return `${wa_nodes_1.WaType[node.valueType]}.max`;
            case "CopySign":
                return `${wa_nodes_1.WaType[node.valueType]}.copysign`;
            case "Block":
                const body = node.body
                    .map((inst) => this.renderInstructionNode(inst, indent + 1))
                    .join("\n");
                return `block ${node.id}${node.resultType === undefined
                    ? ""
                    : " (result " + wa_nodes_1.WaType[node.resultType] + ")"}\n${body}${body.length > 0 ? "\n" : ""}${indentation}end`;
            case "Loop":
                const loop = node.body
                    .map((inst) => this.renderInstructionNode(inst, indent + 1))
                    .join("\n");
                return `loop ${node.id}${node.resultType === undefined
                    ? ""
                    : " (result " + wa_nodes_1.WaType[node.resultType] + ")"}\n${loop}${loop.length > 0 ? "\n" : ""}${indentation}end`;
            case "If":
                const consequtive = node.consequtive
                    ? node.consequtive
                        .map((inst) => this.renderInstructionNode(inst, indent + 1))
                        .join("\n")
                    : "";
                const alternate = node.alternate
                    ? node.alternate
                        .map((inst) => this.renderInstructionNode(inst, indent + 1))
                        .join("\n")
                    : "";
                return `if${node.resultType === undefined
                    ? ""
                    : " (result " + wa_nodes_1.WaType[node.resultType] + ")"}\n${consequtive}${consequtive.length > 0 ? "\n" : ""}${node.alternate ? `${indentation}else\n` : ""}${alternate}${alternate.length > 0 ? "\n" : ""}${indentation}end`;
            case "Comment":
                return node.isBlock ? `(; ${node.text} ;)` : `;; ${node.text}`;
            case "SeparatorLine":
                return indentation;
        }
        // --- Convert bit specification to string
        function bitSpecToString(bits) {
            switch (bits) {
                case wa_nodes_1.WaBitSpec.Bit8:
                    return "8";
                case wa_nodes_1.WaBitSpec.Bit16:
                    return "16";
                case wa_nodes_1.WaBitSpec.Bit32:
                    return "32";
                default:
                    return "";
            }
        }
        function signedTag(type, signed) {
            return type === wa_nodes_1.WaType.i32 || type === wa_nodes_1.WaType.i64
                ? signed
                    ? "_s"
                    : "_u"
                : "";
        }
    }
}
exports.WaTree = WaTree;


/***/ }),

/***/ "./src/wa-ast/wa-nodes.ts":
/*!********************************!*\
  !*** ./src/wa-ast/wa-nodes.ts ***!
  \********************************/
/***/ ((__unused_webpack_module, exports) => {


Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.WaBitSpec = exports.WaType = void 0;
// ============================================================================
// Fundamental WebAssembly AST types
/**
 * Types in WebAssembly
 */
var WaType;
(function (WaType) {
    WaType[WaType["None"] = 0] = "None";
    WaType[WaType["i32"] = 1] = "i32";
    WaType[WaType["i64"] = 2] = "i64";
    WaType[WaType["f32"] = 3] = "f32";
    WaType[WaType["f64"] = 4] = "f64";
})(WaType = exports.WaType || (exports.WaType = {}));
/**
 * Bit count specification for storage statements
 */
var WaBitSpec;
(function (WaBitSpec) {
    WaBitSpec[WaBitSpec["None"] = 0] = "None";
    WaBitSpec[WaBitSpec["Bit8"] = 1] = "Bit8";
    WaBitSpec[WaBitSpec["Bit16"] = 2] = "Bit16";
    WaBitSpec[WaBitSpec["Bit32"] = 3] = "Bit32";
})(WaBitSpec = exports.WaBitSpec || (exports.WaBitSpec = {}));


/***/ }),

/***/ "./package.json":
/*!**********************!*\
  !*** ./package.json ***!
  \**********************/
/***/ ((module) => {

module.exports = JSON.parse('{"name":"tpwatson","version":"0.1.3","description":"This utility is a prototype design and implementation of the WAT# programming language. WAT# is a lightweight programming language that transpiles its output to WebAssembly text format. Its aim is that you can easily add powerful WebAssembly code to your JavaScript code, which rely on the performance of native WA.","bin":{"tpwatson":"./bin/tpwatson.js"},"files":["dist"],"scripts":{"build":"npm run clean && webpack","build:prod":"npm run build -- --mode=production","clean":"rimraf ./index.*","test":"ts-mocha -p ./tsconfig.json ./test/**/*.test.ts","pack":"npm pack","link":"npm link --only=production"},"repository":{"type":"git","url":"git+https://github.com/Dotneteer/watson.git"},"keywords":["WebAssembly","WATson","watson","Watson","Watson transpiler"],"author":"Istvan Novak","license":"MIT","bugs":{"url":"https://github.com/Dotneteer/watsharp/issues"},"homepage":"https://github.com/Dotneteer/watsharp#readme","devDependencies":{"@types/expect":"^24.3.0","@types/mocha":"^8.2.0","@types/node":"^14.14.20","expect":"^26.6.2","mocha":"^8.2.1","rimraf":"^3.0.2","ts-loader":"^8.0.18","ts-mocha":"^8.0.0","typescript":"^4.2.3","webpack":"^5.28.0","webpack-cli":"^4.5.0"},"dependencies":{"import-local":"^3.0.2"}}');

/***/ }),

/***/ "fs":
/*!*********************!*\
  !*** external "fs" ***!
  \*********************/
/***/ ((module) => {

module.exports = require("fs");;

/***/ }),

/***/ "path":
/*!***********************!*\
  !*** external "path" ***!
  \***********************/
/***/ ((module) => {

module.exports = require("path");;

/***/ })

/******/ 	});
/************************************************************************/
/******/ 	// The module cache
/******/ 	var __webpack_module_cache__ = {};
/******/ 	
/******/ 	// The require function
/******/ 	function __webpack_require__(moduleId) {
/******/ 		// Check if module is in cache
/******/ 		var cachedModule = __webpack_module_cache__[moduleId];
/******/ 		if (cachedModule !== undefined) {
/******/ 			return cachedModule.exports;
/******/ 		}
/******/ 		// Create a new module (and put it into the cache)
/******/ 		var module = __webpack_module_cache__[moduleId] = {
/******/ 			// no module.id needed
/******/ 			// no module.loaded needed
/******/ 			exports: {}
/******/ 		};
/******/ 	
/******/ 		// Execute the module function
/******/ 		__webpack_modules__[moduleId](module, module.exports, __webpack_require__);
/******/ 	
/******/ 		// Return the exports of the module
/******/ 		return module.exports;
/******/ 	}
/******/ 	
/************************************************************************/
var __webpack_exports__ = {};
// This entry need to be wrapped in an IIFE because it need to be isolated against other modules in the chunk.
(() => {
var exports = __webpack_exports__;
/*!**********************!*\
  !*** ./src/index.ts ***!
  \**********************/

Object.defineProperty(exports, "__esModule", ({ value: true }));
exports.run = void 0;
const path = __webpack_require__(/*! path */ "path");
const fs = __webpack_require__(/*! fs */ "fs");
const version_1 = __webpack_require__(/*! ./version */ "./src/version.ts");
const WatSharpCompiler_1 = __webpack_require__(/*! ./compiler/WatSharpCompiler */ "./src/compiler/WatSharpCompiler.ts");
// The files used in this compilation
const filesIncluded = {};
/**
 * Runs the watsc command line tool
 */
function run() {
    console.log(`WATson Transpiler (v${version_1.getVersion()})`);
    // --- Input and output file information
    const inFile = process.argv[2];
    let outFile = process.argv[3];
    if (!inFile) {
        console.log("Error: Input file not specified");
        process.exit(1);
    }
    let fileIndex = 0;
    let source;
    try {
        source = fs.readFileSync(inFile, "utf8");
    }
    catch (err) {
        console.log(err);
        process.exit(1);
    }
    if (!outFile) {
        outFile = inFile.split(".").slice(0, -1).join(".") + ".wat";
    }
    if (inFile === outFile) {
        console.log("Error: The input file is the same as the output file");
        process.exit(1);
    }
    // --- Do the compilation
    filesIncluded[0] = inFile;
    console.log(`Compiling ${inFile} from WAT# to WAT...`);
    const compiler = new WatSharpCompiler_1.WatSharpCompiler(source, includeHandler);
    const outputWat = compiler.compile();
    if (compiler.hasErrors) {
        for (const errInfo of compiler.errors) {
            console.log(`${errInfo.code}: ${errInfo.text} (${errInfo.line}:${errInfo.column})`);
        }
        process.exit(1);
    }
    // --- Try to write the output file
    const dir = path.dirname(outFile);
    try {
        if (!fs.existsSync(dir)) {
            fs.mkdirSync(dir, { recursive: true });
        }
        fs.writeFileSync(outFile, outputWat);
        console.log(`WAT code written to ${outFile}`);
        process.exit(0);
    }
    catch (err) {
        console.log(err);
        process.exit(1);
    }
    function includeHandler(baseFileIndex, fileName) {
        const inputDir = path.dirname(path.resolve(filesIncluded[baseFileIndex]));
        fileName = path.isAbsolute(fileName) ? fileName : path.join(inputDir, fileName);
        let source;
        try {
            source = fs.readFileSync(fileName, "utf8");
        }
        catch (err) {
            console.log(err);
            process.exit(1);
        }
        fileIndex++;
        filesIncluded[fileIndex] = fileName;
        return {
            fileIndex,
            source,
        };
    }
}
exports.run = run;

})();

/******/ 	return __webpack_exports__;
/******/ })()
;
});
//# sourceMappingURL=index.js.map